@@ -1,18 +1,42 @@
+/*
+ * Copyright (c) 2014-2020 Yubico AB
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are
+ * met:
+ *
+ *   * Redistributions of source code must retain the above copyright
+ *     notice, this list of conditions and the following disclaimer.
+ *
+ *   * Redistributions in binary form must reproduce the above
+ *     copyright notice, this list of conditions and the following
+ *     disclaimer in the documentation and/or other materials provided
+ *     with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+ * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ *
+ */
+
+#include "internal.h"
 #ifdef _WIN32
 #include <windows.h>
-#ifdef _MSC_VER
-#define strcasecmp _stricmp
-#endif
+#include <wincrypt.h>
+#include <bcrypt.h>
+#include <strsafe.h>
 #else
 #include <ctype.h>
 #include <syslog.h>
-#endif
-
-/* the _WINDOWS define really means Windows native crypto-api/CNG */
-#ifdef _WINDOWS
-#include <wincrypt.h>
-#include <bcrypt.h>
-#else
 #include <openssl/des.h>
 #include <openssl/evp.h>
 #include <openssl/rand.h>
@@ -23,20 +47,14 @@
 #include <stdlib.h>
 #include <string.h>
 
-#ifdef _WIN32
-#include <strsafe.h> /* must be included after openssl headers */
-#endif
-
-#include "internal.h"
-
 /*
 ** Definitions
 */
 
 /* crypt defines */
 
-#ifdef _WINDOWS
-
+#ifdef _WIN32
+#define strcasecmp _stricmp
 #define STATUS_SUCCESS                   ((NTSTATUS)0x00000000L)
 
 struct des_key {
@@ -132,7 +150,7 @@ des_rc des_import_key(const int type, const unsigned char* keyraw, const size_t
   des_rc rc = DES_OK;
   size_t cb_expectedkey = DES_LEN_3DES;
 
-#ifdef _WINDOWS
+#ifdef _WIN32
 
   HCRYPTKEY hNullKey = 0;
   ALG_ID alg = 0;
@@ -242,7 +260,7 @@ des_rc des_import_key(const int type, const unsigned char* keyraw, const size_t
 #endif
 
 EXIT:
-#ifdef _WINDOWS
+#ifdef _WIN32
   if (pbSessionBlob) {
     yc_memzero(pbSessionBlob, cbSessionBlob);
     free(pbSessionBlob);
@@ -268,7 +286,7 @@ ERROR_EXIT:
 
 des_rc des_destroy_key(des_key* key) {
   if (key) {
-#ifdef _WINDOWS
+#ifdef _WIN32
     if (key->hKey) {
       CryptDestroyKey(key->hKey);
       key->hKey = 0;
@@ -288,14 +306,14 @@ des_rc des_destroy_key(des_key* key) {
 des_rc des_encrypt(des_key* key, const unsigned char* in, const size_t inlen, unsigned char* out, size_t* outlen) {
   des_rc rc = DES_OK;
 
-#ifdef _WINDOWS
+#ifdef _WIN32
   unsigned char buf[8] = { 0 };
   size_t buflen = sizeof(buf);
 #endif
 
   if (!key || !outlen || (*outlen < inlen) || !in || !out) { rc = DES_INVALID_PARAMETER; goto EXIT; }
 
-#ifdef _WINDOWS
+#ifdef _WIN32
 
   if (!key->hKey) { rc = DES_INVALID_PARAMETER; goto EXIT; }
 
@@ -308,11 +326,8 @@ des_rc des_encrypt(des_key* key, const unsigned char* in, const size_t inlen, un
 
 #else
 
-#pragma GCC diagnostic push
-#pragma GCC diagnostic ignored "-Wcast-qual"
   /* openssl returns void */
   DES_ecb3_encrypt((const_DES_cblock *)in, (DES_cblock*)out, &(key->ks1), &(key->ks2), &(key->ks3), 1);
-#pragma GCC diagnostic pop
 
 #endif
 
@@ -323,14 +338,14 @@ EXIT:
 des_rc des_decrypt(des_key* key, const unsigned char* in, const size_t inlen, unsigned char* out, size_t* outlen) {
   des_rc rc = DES_OK;
 
-#ifdef _WINDOWS
+#ifdef _WIN32
   unsigned char buf[8] = { 0 };
   size_t buflen = sizeof(buf);
 #endif
 
   if (!key || !outlen || (*outlen < inlen) || !in || !out) { rc = DES_INVALID_PARAMETER; goto EXIT; }
 
-#ifdef _WINDOWS
+#ifdef _WIN32
 
   if (!key->hKey) { rc = DES_INVALID_PARAMETER; goto EXIT; }
 
@@ -343,11 +358,8 @@ des_rc des_decrypt(des_key* key, const unsigned char* in, const size_t inlen, un
 
 #else
 
-#pragma GCC diagnostic push
-#pragma GCC diagnostic ignored "-Wcast-qual"
   /* openssl returns void */
   DES_ecb3_encrypt((const_DES_cblock*)in, (DES_cblock*)out, &(key->ks1), &(key->ks2), &(key->ks3), 0);
-#pragma GCC diagnostic pop
 
 #endif
 
@@ -356,7 +368,7 @@ EXIT:
 }
 
 bool yk_des_is_weak_key(const unsigned char *key, const size_t cb_key) {
-#ifdef _WINDOWS
+#ifdef _WIN32
   bool rv = false;
   /* defined weak keys, borrowed from openssl to be consistent across platforms */
   static const unsigned char weak_keys[][DES_LEN_DES] = {
@@ -415,17 +427,14 @@ bool yk_des_is_weak_key(const unsigned char *key, const size_t cb_key) {
 #else
   (void)cb_key; /* unused */
 
-#pragma GCC diagnostic push
-#pragma GCC diagnostic ignored "-Wcast-qual"
   return DES_is_weak_key((const_DES_cblock *)key);
-#pragma GCC diagnostic pop
 #endif
 }
 
 prng_rc _ykpiv_prng_generate(unsigned char *buffer, const size_t cb_req) {
   prng_rc rc = PRNG_OK;
 
-#ifdef _WINDOWS
+#ifdef _WIN32
   HCRYPTPROV hProv = 0;
 
   if (CryptAcquireContext(&hProv, NULL, MS_ENHANCED_PROV, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT)) {
@@ -452,7 +461,7 @@ prng_rc _ykpiv_prng_generate(unsigned char *buffer, const size_t cb_req) {
 pkcs5_rc pkcs5_pbkdf2_sha1(const uint8_t* password, const size_t cb_password, const uint8_t* salt, const size_t cb_salt, uint64_t iterations, const uint8_t* key, const size_t cb_key) {
   pkcs5_rc rc = PKCS5_OK;
 
-#ifdef _WINDOWS
+#ifdef _WIN32
   BCRYPT_ALG_HANDLE hAlg = 0;
 
   /* mingw64 defines the BCryptDeriveKeyPBKDF2 function, but its dll link library doesn't include the export.
@@ -479,11 +488,8 @@ pkcs5_rc pkcs5_pbkdf2_sha1(const uint8_t* password, const size_t cb_password, co
 
 #else
 
-#pragma GCC diagnostic push
-#pragma GCC diagnostic ignored "-Wcast-qual"
   if(PKCS5_PBKDF2_HMAC_SHA1((const char*)password, cb_password, salt, cb_salt, iterations, cb_key, (unsigned char*)key) <= 0)
     rc = PKCS5_GENERAL_ERROR;
-#pragma GCC diagnostic pop
 
 #endif
 
@@ -530,7 +536,7 @@ setting_bool_t _get_bool_config(const char *sz_setting) {
 
 #else
   /* read from config file*/
-  char sz_line[256];
+  char sz_line[256] = { 0 };
   char *psz_name = 0;
   char *psz_value = 0;
   char sz_name[256] = { 0 };
@@ -570,23 +576,8 @@ setting_bool_t _get_bool_env(const char *sz_setting) {
   setting_bool_t setting = { false, SETTING_SOURCE_DEFAULT };
   char *psz_value = NULL;
   char sz_name[256] = { 0 };
-
   snprintf(sz_name, sizeof(sz_name) - 1, "%s%s", _ENV_PREFIX, sz_setting);
-
-  /* MINGW does not implement getenv_s, only _wgetenv_s */
-#ifdef _MSC_VER
-  size_t cb_value = 0;
-  char sz_value[100] = { 0 };
-
-  if ((getenv_s(&cb_value, sz_value, sizeof(sz_value) - 1, sz_name) == 0) && (cb_value > 0)) {
-    psz_value = sz_value;
-  }
-
-#else
   psz_value = getenv(sz_name);
-
-#endif
-
   if (psz_value) {
     setting.source = SETTING_SOURCE_USER;
     setting.value = (!strcmp(psz_value, "1") || !strcasecmp(psz_value, "true"));
@@ -614,7 +605,7 @@ setting_bool_t setting_get_bool(const char *sz_setting, bool def) {
 /* logging */
 
 void yc_log_event(uint32_t id, yc_log_level_t level, const char * sz_format, ...) {
-  char rgsz_message[4096];
+  char rgsz_message[4096] = {0};
   va_list vl;
 
 #ifdef _WIN32
