@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2016 Yubico AB
+ * Copyright (c) 2014-2020 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -31,12 +31,13 @@
 #ifndef YKPIV_INTERNAL_H
 #define YKPIV_INTERNAL_H
 
+#include "ykpiv-config.h"
 #include "ykpiv.h"
 
 #include <stdbool.h>
 
-#if BACKEND_PCSC
-#if defined HAVE_PCSC_WINSCARD_H
+#ifdef BACKEND_PCSC
+#ifdef HAVE_PCSC_WINSCARD_H
 # include <PCSC/wintypes.h>
 # include <PCSC/winscard.h>
 #else
@@ -157,6 +158,7 @@ struct ykpiv_state {
   SCARDCONTEXT context;
   SCARDHANDLE card;
   int  verbose;
+  int tries;
   char *pin;
   uint8_t *mgm_key;
   ykpiv_allocator allocator;
@@ -192,9 +194,9 @@ ykpiv_rc _ykpiv_begin_transaction(ykpiv_state *state);
 ykpiv_rc _ykpiv_end_transaction(ykpiv_state *state);
 ykpiv_rc _ykpiv_ensure_application_selected(ykpiv_state *state);
 ykpiv_rc _ykpiv_select_application(ykpiv_state *state);
-unsigned int _ykpiv_set_length(unsigned char *buffer, size_t length);
-unsigned int _ykpiv_get_length(const unsigned char *buffer, size_t *len);
-bool _ykpiv_has_valid_length(const unsigned char* buffer, size_t len);
+size_t _ykpiv_get_length_size(size_t length);
+size_t _ykpiv_set_length(unsigned char *buffer, size_t length);
+size_t _ykpiv_get_length(const unsigned char *buffer, const unsigned char* end, size_t *len);
 
 void* _ykpiv_alloc(ykpiv_state *state, size_t size);
 void* _ykpiv_realloc(ykpiv_state *state, void *address, size_t size);
@@ -242,7 +244,7 @@ void yc_log_event(uint32_t id, yc_log_level_t level, const char *sz_format, ...)
 #ifdef _WIN32
 #include <windows.h>
 #define yc_memzero SecureZeroMemory
-#elif defined(BSD)
+#elif defined(HAVE_EXPLICIT_BZERO)
 #include <strings.h>
 #define yc_memzero explicit_bzero
 #elif defined(__linux__)
