@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2016 Yubico AB
+ * Copyright (c) 2014-2020 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -30,7 +30,8 @@
 
 #include "ykpiv.h"
 #include "internal.h"
-#include "../../tool/openssl-compat.h"
+#include "../../common/openssl-compat.h"
+#include "test-config.h"
 
 #include <stddef.h>
 #include <stdio.h>
@@ -39,7 +40,7 @@
 
 #include <check.h>
 
-#ifdef __MINGW32__
+#ifdef _WIN32
 #define dprintf(fd, ...) fprintf(stdout, __VA_ARGS__)
 #endif
 
@@ -100,8 +101,8 @@ void teardown(void) {
 START_TEST(test_devicemodel) {
   ykpiv_rc res;
   ykpiv_devmodel model;
-  char version[256];
-  char reader_buf[2048];
+  char version[256] = {0};
+  char reader_buf[2048] = {0};
   size_t num_readers = sizeof(reader_buf);
 
   res = ykpiv_get_version(g_state, version, sizeof(version));
@@ -149,7 +150,7 @@ END_TEST
 
 START_TEST(test_list_readers) {
   ykpiv_rc res;
-  char reader_buf[2048];
+  char reader_buf[2048] = {0};
   size_t num_readers = sizeof(reader_buf);
   char *reader_ptr;
   res = ykpiv_list_readers(g_state, reader_buf, &num_readers);
@@ -263,36 +264,6 @@ static const char *certificate_pem =
   "cbJb2PEXibNEQG234riKxm7x7vNDLL79Jwtc\n"
   "-----END CERTIFICATE-----\n";
 
-static bool set_component(unsigned char *in_ptr, const BIGNUM *bn, int element_len) {
-  int real_len = BN_num_bytes(bn);
-
-  if(real_len > element_len) {
-    return false;
-  }
-  memset(in_ptr, 0, (size_t)(element_len - real_len));
-  in_ptr += element_len - real_len;
-  BN_bn2bin(bn, in_ptr);
-
-  return true;
-}
-
-static bool prepare_rsa_signature(const unsigned char *in, unsigned int in_len, unsigned char *out, unsigned int *out_len, int nid) {
-  X509_SIG *digestInfo;
-  X509_ALGOR *algor;
-  ASN1_OCTET_STRING *digest;
-  unsigned char data[1024];
-
-  memcpy(data, in, in_len);
-
-  digestInfo = X509_SIG_new();
-  X509_SIG_getm(digestInfo, &algor, &digest);
-  algor->algorithm = OBJ_nid2obj(nid);
-  X509_ALGOR_set0(algor, OBJ_nid2obj(nid), V_ASN1_NULL, NULL);
-  ASN1_STRING_set(digest, data, in_len);
-  *out_len = (unsigned int)i2d_X509_SIG(digestInfo, &out);
-  X509_SIG_free(digestInfo);
-  return true;
-}
 
 static void import_key(unsigned char slot, unsigned char pin_policy) {
   ykpiv_rc res;
@@ -302,12 +273,12 @@ static void import_key(unsigned char slot, unsigned char pin_policy) {
     EVP_PKEY *private_key = NULL;
     BIO *bio = NULL;
     RSA *rsa_private_key = NULL;
-    unsigned char e[4];
-    unsigned char p[128];
-    unsigned char q[128];
-    unsigned char dmp1[128];
-    unsigned char dmq1[128];
-    unsigned char iqmp[128];
+    unsigned char e[4] = {0};
+    unsigned char p[128] = {0};
+    unsigned char q[128] = {0};
+    unsigned char dmp1[128] = {0};
+    unsigned char dmq1[128] = {0};
+    unsigned char iqmp[128] = {0};
     int element_len = 128;
     const BIGNUM *bn_e, *bn_p, *bn_q, *bn_dmp1, *bn_dmq1, *bn_iqmp;
 
@@ -361,9 +332,9 @@ static void import_key(unsigned char slot, unsigned char pin_policy) {
     BIO *bio = NULL;
     X509 *cert = NULL;
     EVP_PKEY *pub_key = NULL;
-    unsigned char secret[32];
-    unsigned char secret2[32];
-    unsigned char data[256];
+    unsigned char secret[32] = {0};
+    unsigned char secret2[32] = {0};
+    unsigned char data[256] = {0};
     int len;
     size_t len2 = sizeof(data);
     RSA *rsa = NULL;
@@ -406,11 +377,11 @@ START_TEST(test_import_key) {
     const EVP_MD *md = EVP_sha256();
     EVP_MD_CTX *mdctx;
 
-    unsigned char signature[1024];
-    unsigned char encoded[1024];
-    unsigned char data[1024];
-    unsigned char signinput[1024];
-    unsigned char rand[128];
+    unsigned char signature[1024] = {0};
+    unsigned char encoded[1024] = {0};
+    unsigned char data[1024] = {0};
+    unsigned char signinput[1024] = {0};
+    unsigned char rand[128] = {0};
 
     size_t sig_len = sizeof(signature);
     size_t padlen = 256;
@@ -447,7 +418,7 @@ START_TEST(test_import_key) {
 
   // Verify that imported key can not be attested
   {
-    unsigned char attest[2048];
+    unsigned char attest[2048] = {0};
     size_t attest_len = sizeof(attest);
     ykpiv_devmodel model;
     model = ykpiv_util_devicemodel(g_state);
@@ -486,11 +457,11 @@ START_TEST(test_pin_policy_always) {
     const EVP_MD *md = EVP_sha256();
     EVP_MD_CTX *mdctx;
 
-    unsigned char signature[1024];
-    unsigned char encoded[1024];
-    unsigned char data[1024];
-    unsigned char signinput[1024];
-    unsigned char rand[128];
+    unsigned char signature[1024] = {0};
+    unsigned char encoded[1024] = {0};
+    unsigned char data[1024] = {0};
+    unsigned char signinput[1024] = {0};
+    unsigned char rand[128] = {0};
 
     size_t sig_len = sizeof(signature);
     size_t padlen = 256;
@@ -571,7 +542,7 @@ START_TEST(test_generate_key) {
   // Verify that imported key can be attested
   {
     ykpiv_devmodel model;
-    unsigned char attest[2048];
+    unsigned char attest[2048] = {0};
     size_t attest_len = sizeof(attest);
     model = ykpiv_util_devicemodel(g_state);
     res = ykpiv_attest(g_state, YKPIV_KEY_AUTHENTICATION, attest, &attest_len);
@@ -592,7 +563,7 @@ static void test_authenticate_helper() {
   const char *default_mgm_key = "010203040506070801020304050607080102030405060708";
   const char *mgm_key = "112233445566778811223344556677881122334455667788";
   const char *weak_mgm_key = "FEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFE";
-  unsigned char key[24];
+  unsigned char key[24] = {0};
   size_t key_len = sizeof(key);
 
   // Try new key, fail.
@@ -715,10 +686,10 @@ static int block_and_reset() {
   tries_until_blocked = 0;
   while (tries) {
     res = ykpiv_verify(g_state, "AAAAAA", &tries);
+    tries_until_blocked++;
     if (res == YKPIV_PIN_LOCKED)
       break;
     ck_assert_int_eq(res, YKPIV_WRONG_PIN);
-    tries_until_blocked++;
   }
 
   // Verify no PIN retries remaining
@@ -754,7 +725,7 @@ START_TEST(test_reset) {
   test_authenticate_helper();
   res = ykpiv_verify(g_state, "123456", &tries);
   ck_assert_int_eq(res, YKPIV_OK);
-  ck_assert_int_eq(tries, 0);
+  ck_assert_int_eq(tries, -1);
   res = ykpiv_set_pin_retries(g_state, 8, 3);
   ck_assert_int_eq(res, YKPIV_OK);
 
@@ -790,7 +761,7 @@ START_TEST(test_reset) {
   tries = 100;
   res = ykpiv_verify(g_state, "123456", &tries);
   ck_assert_int_eq(res, YKPIV_OK);
-  ck_assert_int_eq(tries, 0);
+  ck_assert_int_eq(tries, -1);
 
   // Verify back to 3 PIN retries remaining
   tries = 0;
@@ -892,7 +863,7 @@ END_TEST
 START_TEST(test_pin_cache) {
   ykpiv_rc res;
   ykpiv_state *local_state;
-  unsigned char data[256];
+  unsigned char data[256] = {0};
   unsigned char data_in[256] = {0};
   int len = sizeof(data);
   size_t len2 = sizeof(data);
@@ -950,10 +921,15 @@ END_TEST
 
 int destruction_confirmed(void) {
   char *confirmed = getenv("YKPIV_ENV_HWTESTS_CONFIRMED");
-  if (confirmed && confirmed[0] == '1')
+  if (confirmed && confirmed[0] == '1') {
+#ifdef _WIN32
     return 1;
-  // Use dprintf() to write directly to stdout, since automake eats the standard stdout/stderr pointers.
-  dprintf(0, "\n***\n*** Hardware tests skipped.  Run \"make hwcheck\".\n***\n\n");
+#else
+    return system("../../../tools/confirm.sh") == 0;
+#endif
+  }
+  // Use dprintf() to write directly to stdout, since cmake eats the standard stdout/stderr pointers.
+  dprintf(0, "\n***\n*** Hardware tests skipped.\n***\n\n");
   return 0;
 }
 
