@@ -1,5 +1,5 @@
  /*
- * Copyright (c) 2014-2016 Yubico AB
+ * Copyright (c) 2014-2020 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -96,10 +96,10 @@ static size_t _obj_size_max(ykpiv_state *state) {
 
 ykpiv_rc ykpiv_util_get_cardid(ykpiv_state *state, ykpiv_cardid *cardid) {
   ykpiv_rc res = YKPIV_OK;
-  uint8_t buf[CB_OBJ_MAX];
+  uint8_t buf[CB_OBJ_MAX] = {0};
   unsigned long len = sizeof(buf);
   uint8_t *p_temp = NULL;
-  size_t cb_temp = 0;
+  size_t offs, cb_temp = 0;
   uint8_t tag = 0;
 
   if (!cardid) return YKPIV_GENERIC_ERROR;
@@ -113,12 +113,13 @@ ykpiv_rc ykpiv_util_get_cardid(ykpiv_state *state, ykpiv_cardid *cardid) {
     while (p_temp < (buf + len)) {
       tag = *p_temp++;
 
-      if (!_ykpiv_has_valid_length(p_temp, (size_t)(buf + len - p_temp))) {
-        res = YKPIV_SIZE_ERROR;
+      offs = _ykpiv_get_length(p_temp, buf + len, &cb_temp);
+      if (!offs) {
+        res = YKPIV_PARSE_ERROR;
         goto Cleanup;
       }
 
-      p_temp += _ykpiv_get_length(p_temp, &cb_temp);
+      p_temp += offs;
 
       if (tag == TAG_CHUID_UUID) {
         /* found card uuid */
@@ -147,8 +148,8 @@ Cleanup:
 
 ykpiv_rc ykpiv_util_set_cardid(ykpiv_state *state, const ykpiv_cardid *cardid) {
   ykpiv_rc res = YKPIV_OK;
-  uint8_t id[YKPIV_CARDID_SIZE];
-  uint8_t buf[sizeof(CHUID_TMPL)];
+  uint8_t id[YKPIV_CARDID_SIZE] = {0};
+  uint8_t buf[sizeof(CHUID_TMPL)] = {0};
   size_t len = 0;
 
   if (!state) return YKPIV_GENERIC_ERROR;
@@ -179,7 +180,7 @@ Cleanup:
 
 ykpiv_rc ykpiv_util_get_cccid(ykpiv_state *state, ykpiv_cccid *ccc) {
   ykpiv_rc res = YKPIV_OK;
-  uint8_t buf[CB_OBJ_MAX];
+  uint8_t buf[CB_OBJ_MAX] = {0};
   unsigned long len = sizeof(buf);
 
   if (!ccc) return YKPIV_GENERIC_ERROR;
@@ -205,8 +206,8 @@ Cleanup:
 
 ykpiv_rc ykpiv_util_set_cccid(ykpiv_state *state, const ykpiv_cccid *ccc) {
   ykpiv_rc res = YKPIV_OK;
-  uint8_t id[YKPIV_CCCID_SIZE];
-  uint8_t buf[sizeof(CCC_TMPL)];
+  uint8_t id[YKPIV_CCCID_SIZE] = {0};
+  uint8_t buf[sizeof(CCC_TMPL)] = {0};
   size_t len = 0;
 
   if (!state) return YKPIV_GENERIC_ERROR;
@@ -247,7 +248,7 @@ ykpiv_rc ykpiv_util_list_keys(ykpiv_state *state, uint8_t *key_count, ykpiv_key
   uint8_t *pTemp = NULL;
   size_t cbData = 0;
   size_t offset = 0;
-  uint8_t buf[CB_BUF_MAX];
+  uint8_t buf[CB_BUF_MAX] = {0};
   size_t cbBuf = 0;
   size_t i = 0;
   size_t cbRealloc = 0;
@@ -363,7 +364,7 @@ ykpiv_rc ykpiv_util_free(ykpiv_state *state, void *data) {
 
 ykpiv_rc ykpiv_util_read_cert(ykpiv_state *state, uint8_t slot, uint8_t **data, size_t *data_len) {
   ykpiv_rc res = YKPIV_OK;
-  uint8_t buf[CB_BUF_MAX];
+  uint8_t buf[CB_BUF_MAX] = {0};
   size_t cbBuf = sizeof(buf);
 
   if ((NULL == data )|| (NULL == data_len)) return YKPIV_GENERIC_ERROR;
@@ -421,7 +422,7 @@ ykpiv_rc ykpiv_util_block_puk(ykpiv_state *state) {
   ykpiv_rc res = YKPIV_OK;
   uint8_t puk[] = { 0x30, 0x42, 0x41, 0x44, 0x46, 0x30, 0x30, 0x44 };
   int tries = -1;
-  uint8_t data[CB_BUF_MAX];
+  uint8_t data[CB_BUF_MAX] = {0};
   size_t  cb_data = sizeof(data);
   uint8_t *p_item = NULL;
   size_t  cb_item = 0;
@@ -479,9 +480,9 @@ Cleanup:
 
 ykpiv_rc ykpiv_util_read_mscmap(ykpiv_state *state, ykpiv_container **containers, size_t *n_containers) {
   ykpiv_rc res = YKPIV_OK;
-  uint8_t buf[CB_BUF_MAX];
+  uint8_t buf[CB_BUF_MAX] = {0};
   unsigned long cbBuf = sizeof(buf);
-  size_t len = 0;
+  size_t offs, len = 0;
   uint8_t *ptr = NULL;
 
   if ((NULL == containers) || (NULL == n_containers)) { res = YKPIV_GENERIC_ERROR; goto Cleanup; }
@@ -501,13 +502,12 @@ ykpiv_rc ykpiv_util_read_mscmap(ykpiv_state *state, ykpiv_container **containers
     }
 
     if (*ptr++ == TAG_MSCMAP) {
-      ptr += (unsigned long)_ykpiv_get_length(ptr, &len);
-
-      /* check that decoded length represents object contents */
-      if (len > (cbBuf - (size_t)(ptr - buf))) {
+      offs = _ykpiv_get_length(ptr, buf + cbBuf, &len);
+      if(!offs) {
         res = YKPIV_OK;
         goto Cleanup;
       }
+      ptr += offs;
 
       if (NULL == (*containers = _ykpiv_alloc(state, len))) {
         res = YKPIV_MEMORY_ERROR;
@@ -529,7 +529,7 @@ Cleanup:
 
 ykpiv_rc ykpiv_util_write_mscmap(ykpiv_state *state, ykpiv_container *containers, size_t n_containers) {
   ykpiv_rc res = YKPIV_OK;
-  uint8_t buf[CB_OBJ_MAX];
+  uint8_t buf[CB_OBJ_MAX] = {0};
   size_t offset = 0;
   size_t req_len = 0;
   size_t data_len = n_containers * sizeof(ykpiv_container);
@@ -579,9 +579,9 @@ Cleanup:
 
 ykpiv_rc ykpiv_util_read_msroots(ykpiv_state *state, uint8_t **data, size_t *data_len) {
   ykpiv_rc res = YKPIV_OK;
-  uint8_t buf[CB_BUF_MAX];
+  uint8_t buf[CB_BUF_MAX] = {0};
   unsigned long cbBuf = sizeof(buf);
-  size_t len = 0;
+  size_t offs, len = 0;
   uint8_t *ptr = NULL;
   int object_id = 0;
   uint8_t tag = 0;
@@ -626,13 +626,12 @@ ykpiv_rc ykpiv_util_read_msroots(ykpiv_state *state, uint8_t **data, size_t *dat
       goto Cleanup;
     }
 
-    ptr += _ykpiv_get_length(ptr, &len);
-
-    // check that decoded length represents object contents
-    if (len > (cbBuf - (size_t)(ptr - buf))) {
+    offs = _ykpiv_get_length(ptr, buf + cbBuf, &len);
+    if(!offs) {
       res = YKPIV_OK;
       goto Cleanup;
     }
+    ptr += offs;
 
     cbRealloc = len > (cbData - offset) ? len - (cbData - offset) : 0;
 
@@ -673,7 +672,7 @@ Cleanup:
 
 ykpiv_rc ykpiv_util_write_msroots(ykpiv_state *state, uint8_t *data, size_t data_len) {
   ykpiv_rc res = YKPIV_OK;
-  uint8_t buf[CB_OBJ_MAX];
+  uint8_t buf[CB_OBJ_MAX] = {0};
   size_t offset = 0;
   size_t data_offset = 0;
   size_t data_chunk = 0;
@@ -739,9 +738,9 @@ Cleanup:
 
 ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algorithm, uint8_t pin_policy, uint8_t touch_policy, uint8_t **modulus, size_t *modulus_len, uint8_t **exp, size_t *exp_len, uint8_t **point, size_t *point_len) {
   ykpiv_rc res = YKPIV_OK;
-  unsigned char in_data[11];
+  unsigned char in_data[11] = {0};
   unsigned char *in_ptr = in_data;
-  unsigned char data[1024];
+  unsigned char data[1024] = {0};
   unsigned char templ[] = { 0, YKPIV_INS_GENERATE_ASYMMETRIC, 0, 0 };
   unsigned long recv_len = sizeof(data);
   int sw;
@@ -751,6 +750,7 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
   size_t  cb_exp = 0;
   uint8_t *ptr_point = NULL;
   size_t  cb_point = 0;
+  size_t offs;
 
   setting_bool_t setting_roca = { 0 };
   const char sz_setting_roca[] = "Enable_Unsafe_Keygen_ROCA";
@@ -894,8 +894,8 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
   }
 
   if ((YKPIV_ALGO_RSA1024 == algorithm) || (YKPIV_ALGO_RSA2048 == algorithm)) {
-    unsigned char *data_ptr = data + 5;
-    size_t len = 0;
+    size_t len;
+    unsigned char *data_ptr = data + 2 + _ykpiv_get_length(data + 2, data + recv_len, &len);
 
     if (*data_ptr != TAG_RSA_MODULUS) {
       if (state->verbose) { fprintf(stderr, "Failed to parse public key structure (modulus).\n"); }
@@ -904,7 +904,13 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
     }
 
     data_ptr++;
-    data_ptr += _ykpiv_get_length(data_ptr, &len);
+    offs = _ykpiv_get_length(data_ptr, data + recv_len, &len);
+    if(!offs) {
+      if (state->verbose) { fprintf(stderr, "Failed to parse public key structure (modulus length).\n"); }
+      res = YKPIV_PARSE_ERROR;
+      goto Cleanup;
+    }
+    data_ptr += offs;
 
     cb_modulus = len;
     if (NULL == (ptr_modulus = _ykpiv_alloc(state, cb_modulus))) {
@@ -924,7 +930,13 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
     }
 
     data_ptr++;
-    data_ptr += _ykpiv_get_length(data_ptr, &len);
+    offs = _ykpiv_get_length(data_ptr, data + recv_len, &len);
+    if(!offs) {
+      if (state->verbose) { fprintf(stderr, "Failed to parse public key structure (public exponent length).\n"); }
+      res = YKPIV_PARSE_ERROR;
+      goto Cleanup;
+    }
+    data_ptr += offs;
 
     cb_exp = len;
     if (NULL == (ptr_exp = _ykpiv_alloc(state, cb_exp))) {
@@ -990,9 +1002,9 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
 
 Cleanup:
 
-  if (ptr_modulus) { _ykpiv_free(state, modulus); }
+  if (ptr_modulus) { _ykpiv_free(state, ptr_modulus); }
   if (ptr_exp) { _ykpiv_free(state, ptr_exp); }
-  if (ptr_point) { _ykpiv_free(state, ptr_exp); }
+  if (ptr_point) { _ykpiv_free(state, ptr_point); }
 
   _ykpiv_end_transaction(state);
   return res;
@@ -1010,7 +1022,6 @@ ykpiv_rc ykpiv_util_get_config(ykpiv_state *state, ykpiv_config *config) {
 
   // initialize default values
 
-  config->protected_data_available = false;
   config->puk_blocked = false;
   config->puk_noblock_on_upgrade = false;
   config->pin_last_changed = 0;
@@ -1030,6 +1041,7 @@ ykpiv_rc ykpiv_util_get_config(ykpiv_state *state, ykpiv_config *config) {
       if (config->mgm_type != YKPIV_CONFIG_MGM_MANUAL) {
         if (state->verbose) {
           fprintf(stderr, "conflicting types of mgm key administration configured\n");
+          config->mgm_type = YKPIV_CONFIG_MGM_INVALID;
         }
       }
       else {
@@ -1053,19 +1065,28 @@ ykpiv_rc ykpiv_util_get_config(ykpiv_state *state, ykpiv_config *config) {
   cb_data = sizeof(data);
 
   if (YKPIV_OK == _read_metadata(state, TAG_PROTECTED, data, &cb_data)) {
-    config->protected_data_available = true;
 
     if (YKPIV_OK == _get_metadata_item(data, cb_data, TAG_PROTECTED_FLAGS_1, &p_item, &cb_item)) {
       if (*p_item & PROTECTED_FLAGS_1_PUK_NOBLOCK) config->puk_noblock_on_upgrade = true;
     }
 
     if (YKPIV_OK == _get_metadata_item(data, cb_data, TAG_PROTECTED_MGM, &p_item, &cb_item)) {
-      if (config->mgm_type != YKPIV_CONFIG_MGM_PROTECTED) {
-        if (state->verbose) {
-          fprintf(stderr, "conflicting types of mgm key administration configured - protected mgm exists\n");
+      if(sizeof(config->mgm_key) == cb_item) {
+        memcpy(config->mgm_key, p_item, cb_item);
+        if (config->mgm_type != YKPIV_CONFIG_MGM_PROTECTED) {
+          if (state->verbose) fprintf(stderr, "conflicting types of mgm key administration configured - protected mgm exists\n");
+          config->mgm_type = YKPIV_CONFIG_MGM_PROTECTED;
         }
+      } else {
+        if (state->verbose) fprintf(stderr, "protected data contains mgm, but is the wrong size = %lu\n", (unsigned long)cb_item);
+        config->mgm_type = YKPIV_CONFIG_MGM_INVALID;
       }
-      config->mgm_type = YKPIV_CONFIG_MGM_PROTECTED; /* always favor protected mgm */
+    }
+  }
+  else {
+    if (config->mgm_type == YKPIV_CONFIG_MGM_PROTECTED) {
+      if (state->verbose) fprintf(stderr, "admin data indicates protected mgm present, but the object cannot be read\n");
+      config->mgm_type = YKPIV_CONFIG_MGM_INVALID;
     }
   }
 
@@ -1322,7 +1343,7 @@ Cleanup:
 
 ykpiv_rc ykpiv_util_reset(ykpiv_state *state) {
   unsigned char templ[] = {0, YKPIV_INS_RESET, 0, 0};
-  unsigned char data[0xff];
+  unsigned char data[0xff] = {0};
   unsigned long recv_len = sizeof(data);
   ykpiv_rc res;
   int sw;
@@ -1372,16 +1393,17 @@ uint32_t ykpiv_util_slot_object(uint8_t slot) {
 static ykpiv_rc _read_certificate(ykpiv_state *state, uint8_t slot, uint8_t *buf, size_t *buf_len) {
   ykpiv_rc res = YKPIV_OK;
   uint8_t *ptr = NULL;
+  unsigned long ul_len = (unsigned long)*buf_len;
   int object_id = (int)ykpiv_util_slot_object(slot);
-  size_t len = 0;
+  size_t offs, len = 0;
 
   if (-1 == object_id) return YKPIV_INVALID_OBJECT;
 
-  if (YKPIV_OK == (res = _ykpiv_fetch_object(state, object_id, buf, (unsigned long*)buf_len))) {
+  if (YKPIV_OK == (res = _ykpiv_fetch_object(state, object_id, buf, &ul_len))) {
     ptr = buf;
 
     // check that object contents are at least large enough to read the tag
-    if (*buf_len < CB_OBJ_TAG_MIN) {
+    if (ul_len < CB_OBJ_TAG_MIN) {
       *buf_len = 0;
       return YKPIV_OK;
     }
@@ -1389,13 +1411,12 @@ static ykpiv_rc _read_certificate(ykpiv_state *state, uint8_t slot, uint8_t *buf
     // check that first byte indicates "certificate" type
 
     if (*ptr++ == TAG_CERT) {
-      ptr += _ykpiv_get_length(ptr, &len);
-
-      // check that decoded length represents object contents
-      if (len > (*buf_len - (size_t)(ptr - buf))) {
+      offs = _ykpiv_get_length(ptr, buf + ul_len, &len);
+      if(!offs) {
         *buf_len = 0;
         return YKPIV_OK;
       }
+      ptr += offs;
 
       memmove(buf, ptr, len);
       *buf_len = len;
@@ -1409,7 +1430,7 @@ static ykpiv_rc _read_certificate(ykpiv_state *state, uint8_t slot, uint8_t *buf
 }
 
 static ykpiv_rc _write_certificate(ykpiv_state *state, uint8_t slot, uint8_t *data, size_t data_len, uint8_t certinfo) {
-  uint8_t buf[CB_OBJ_MAX];
+  uint8_t buf[CB_OBJ_MAX] = {0};
   int object_id = (int)ykpiv_util_slot_object(slot);
   size_t offset = 0;
   size_t req_len = 0;
@@ -1473,8 +1494,9 @@ static ykpiv_rc _write_certificate(ykpiv_state *state, uint8_t slot, uint8_t *da
 */
 static ykpiv_rc _get_metadata_item(uint8_t *data, size_t cb_data, uint8_t tag, uint8_t **pp_item, size_t *pcb_item) {
   uint8_t *p_temp = data;
-  size_t  cb_temp = 0;
+  size_t  offs, cb_temp = 0;
   uint8_t tag_temp = 0;
+  bool found = false;
 
   if (!data || !pp_item || !pcb_item) return YKPIV_GENERIC_ERROR;
 
@@ -1484,28 +1506,28 @@ static ykpiv_rc _get_metadata_item(uint8_t *data, size_t cb_data, uint8_t tag, u
   while (p_temp < (data + cb_data)) {
     tag_temp = *p_temp++;
 
-    if (!_ykpiv_has_valid_length(p_temp, (size_t)(data + cb_data - p_temp))) {
-      return YKPIV_SIZE_ERROR;
+    offs = _ykpiv_get_length(p_temp, data + cb_data, &cb_temp);
+    if (!offs) {
+      return YKPIV_PARSE_ERROR;
     }
 
-    p_temp += _ykpiv_get_length(p_temp, &cb_temp);
+    p_temp += offs;
 
     if (tag_temp == tag) {
       // found tag
+      found = true;
       break;
     }
 
     p_temp += cb_temp;
   }
 
-  // Make sure the item doesn't end after the buffer
-  if ((p_temp + cb_temp) <= (data + cb_data)) {
+  if (found) {
     *pp_item = p_temp;
     *pcb_item = cb_temp;
-    return YKPIV_OK;
   }
 
-  return YKPIV_GENERIC_ERROR;
+  return found ? YKPIV_OK : YKPIV_GENERIC_ERROR;
 }
 
 ykpiv_rc ykpiv_util_parse_metadata(uint8_t *data, size_t data_len, ykpiv_metadata *metadata) {
@@ -1546,18 +1568,6 @@ ykpiv_rc ykpiv_util_parse_metadata(uint8_t *data, size_t data_len, ykpiv_metadat
   return YKPIV_OK;
 }
 
-static int _get_length_size(size_t length) {
-  if (length < 0x80) {
-    return 1;
-  }
-  else if (length < 0xff) {
-    return 2;
-  }
-  else {
-    return 3;
-  }
-}
-
 /*
 ** _set_metadata_item
 **
@@ -1580,7 +1590,10 @@ static ykpiv_rc _set_metadata_item(uint8_t *data, size_t *pcb_data, size_t cb_da
 
   while (p_temp < (data + *pcb_data)) {
     tag_temp = *p_temp++;
-    cb_len = _ykpiv_get_length(p_temp, &cb_temp);
+    cb_len = _ykpiv_get_length(p_temp, data + *pcb_data, &cb_temp);
+    if(!cb_len) {
+        return YKPIV_PARSE_ERROR;
+    }
     p_temp += cb_len;
 
     if (tag_temp == tag) {
@@ -1595,23 +1608,17 @@ static ykpiv_rc _set_metadata_item(uint8_t *data, size_t *pcb_data, size_t cb_da
       /* length doesn't match, expand/shrink to fit */
       p_next = p_temp + cb_temp;
       cb_moved = (long)cb_item - (long)cb_temp +
-        ((long)(cb_item != 0 ? _get_length_size(cb_item) : -1 /* for tag, if deleting */) -
+        ((long)(cb_item != 0 ? (long)_ykpiv_get_length_size(cb_item) : -1l /* for tag, if deleting */) -
         (long)cb_len); /* accounts for different length encoding */
 
       /* length would cause buffer overflow, return error */
-#pragma GCC diagnostic push
-#pragma GCC diagnostic ignored "-Wsign-conversion"
       if ((size_t)(*pcb_data + cb_moved) > cb_data_max) {
         return YKPIV_GENERIC_ERROR;
       }
-#pragma GCC diagnostic pop
 
       /* move remaining data */
       memmove(p_next + cb_moved, p_next, *pcb_data - (size_t)(p_next - data));
-#pragma GCC diagnostic push
-#pragma GCC diagnostic ignored "-Wsign-conversion"
       *pcb_data += cb_moved;
-#pragma GCC diagnostic pop
 
       /* re-encode item and insert */
       if (cb_item != 0) {
@@ -1633,7 +1640,7 @@ static ykpiv_rc _set_metadata_item(uint8_t *data, size_t *pcb_data, size_t cb_da
 
   // we did not find an existing tag, append
   p_temp = data + *pcb_data;
-  cb_len = (size_t)_get_length_size(cb_item);
+  cb_len = _ykpiv_get_length_size(cb_item);
 
   // length would cause buffer overflow, return error
   if (*pcb_data + cb_len + cb_item > cb_data_max) {
@@ -1664,6 +1671,7 @@ static ykpiv_rc _read_metadata(ykpiv_state *state, uint8_t tag, uint8_t* data, s
   ykpiv_rc res = YKPIV_OK;
   uint8_t *p_temp = NULL;
   unsigned long cb_temp = 0;
+  size_t offs;
   int obj_id = 0;
 
   if (!data || !pcb_data || (CB_BUF_MAX > *pcb_data)) return YKPIV_GENERIC_ERROR;
@@ -1674,26 +1682,27 @@ static ykpiv_rc _read_metadata(ykpiv_state *state, uint8_t tag, uint8_t* data, s
   default: return YKPIV_INVALID_OBJECT;
   }
 
-  cb_temp = *pcb_data;
+  cb_temp = (unsigned long)*pcb_data;
   *pcb_data = 0;
 
   if (YKPIV_OK != (res = _ykpiv_fetch_object(state, obj_id, data, &cb_temp))) {
     return res;
   }
 
-  if (cb_temp < CB_OBJ_TAG_MIN) return YKPIV_GENERIC_ERROR;
+  if (cb_temp < CB_OBJ_TAG_MIN) return YKPIV_PARSE_ERROR;
 
   p_temp = data;
 
-  if (tag != *p_temp++) return YKPIV_GENERIC_ERROR;
+  if (tag != *p_temp++) return YKPIV_PARSE_ERROR;
 
-  p_temp += _ykpiv_get_length(p_temp, pcb_data);
-
-  if (*pcb_data > (cb_temp - (size_t)(p_temp - data))) {
+  offs = _ykpiv_get_length(p_temp, data + cb_temp, pcb_data);
+  if (!offs) {
     *pcb_data = 0;
-    return YKPIV_GENERIC_ERROR;
+    return YKPIV_PARSE_ERROR;
   }
 
+  p_temp += offs;
+
   memmove(data, p_temp, *pcb_data);
 
   return YKPIV_OK;
