@@ -1,5 +1,5 @@
  /*
- * Copyright (c) 2014-2016 Yubico AB
+ * Copyright (c) 2014-2020 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -124,7 +124,7 @@ static ykpiv_rc _cache_pin(ykpiv_state *state, const char *pin, size_t len);
 static ykpiv_rc _cache_mgm_key(ykpiv_state *state, unsigned const char *key);
 static ykpiv_rc _ykpiv_get_serial(ykpiv_state *state);
 static ykpiv_rc _ykpiv_get_version(ykpiv_state *state);
-static ykpiv_rc _ykpiv_verify(ykpiv_state *state, const char *pin, const size_t pin_len, int *tries);
+static ykpiv_rc _ykpiv_verify(ykpiv_state *state, const char *pin, const size_t pin_len);
 static ykpiv_rc _ykpiv_authenticate(ykpiv_state *state, unsigned const char *key);
 static ykpiv_rc _ykpiv_auth_deauthenticate(ykpiv_state *state);
 
@@ -175,7 +175,7 @@ void* _ykpiv_realloc(ykpiv_state *state, void *address, size_t size) {
 }
 
 void _ykpiv_free(ykpiv_state *state, void *data) {
-  if (!data || !state || (!(state->allocator.pfn_free))) return;
+  if (!data || !state || !(state->allocator.pfn_free)) return;
   state->allocator.pfn_free(state->allocator.alloc_data, data);
 }
 
@@ -186,7 +186,17 @@ static void dump_hex(const unsigned char *buf, unsigned int len) {
   }
 }
 
-unsigned int _ykpiv_set_length(unsigned char *buffer, size_t length) {
+size_t _ykpiv_get_length_size(size_t length) {
+  if(length < 0x80) {
+    return 1;
+  } else if(length < 0x100) {
+    return 2;
+  } else {
+    return 3;
+  }
+}
+
+size_t _ykpiv_set_length(unsigned char *buffer, size_t length) {
   if(length < 0x80) {
     *buffer++ = (unsigned char)length;
     return 1;
@@ -197,39 +207,27 @@ unsigned int _ykpiv_set_length(unsigned char *buffer, size_t length) {
   } else {
     *buffer++ = 0x82;
     *buffer++ = (length >> 8) & 0xff;
-    *buffer++ = (unsigned char)length & 0xff;
+    *buffer++ = length & 0xff;
     return 3;
   }
 }
 
-unsigned int _ykpiv_get_length(const unsigned char *buffer, size_t *len) {
-  if(buffer[0] < 0x81) {
+size_t _ykpiv_get_length(const unsigned char *buffer, const unsigned char* end, size_t *len) {
+  if(buffer + 1 <= end && buffer[0] < 0x80) {
     *len = buffer[0];
-    return 1;
-  } else if((*buffer & 0x7f) == 1) {
+    return buffer + 1 + *len <= end ? 1 : 0;
+  } else if(buffer + 2 <= end && buffer[0] == 0x81) {
     *len = buffer[1];
-    return 2;
-  } else if((*buffer & 0x7f) == 2) {
+    return buffer + 2 + *len <= end ? 2 : 0;
+  } else if(buffer + 3 <= end && buffer[0] == 0x82) {
     size_t tmp = buffer[1];
     *len = (tmp << 8) + buffer[2];
-    return 3;
+    return buffer + 3 + *len <= end ? 3 : 0;
   }
+  *len = 0;
   return 0;
 }
 
-bool _ykpiv_has_valid_length(const unsigned char* buffer, size_t len) {
-  if ((len > 0) && (*buffer < 0x81)) {
-    return true;
-  }
-  else if ((len > 1) && ((*buffer & 0x7f) == 1)) {
-    return true;
-  }
-  else if ((len > 2) && ((*buffer & 0x7f) == 2)) {
-    return true;
-  }
-  return false;
-}
-
 static unsigned char *set_object(int object_id, unsigned char *buffer) {
   *buffer++ = 0x5c;
   if(object_id == YKPIV_OBJ_DISCOVERY) {
@@ -310,8 +308,8 @@ ykpiv_rc ykpiv_disconnect(ykpiv_state *state) {
 }
 
 ykpiv_rc _ykpiv_select_application(ykpiv_state *state) {
-  APDU apdu;
-  unsigned char data[0xff];
+  APDU apdu = {0};
+  unsigned char data[0xff] = {0};
   uint32_t recv_len = sizeof(data);
   int sw;
   ykpiv_rc res = YKPIV_OK;
@@ -343,6 +341,10 @@ ykpiv_rc _ykpiv_select_application(ykpiv_state *state) {
    * can determine how to get the serial number, which for the NEO/Yk4
    * will result in another selection of the PIV applet. */
 
+  // This stores the number of PIN retries left in state
+  _ykpiv_verify(state, NULL, 0);
+  // WRONG_PIN or PIN_LOCKED is expected on successful query.
+
   res = _ykpiv_get_version(state);
   if (res != YKPIV_OK) {
     if (state->verbose) {
@@ -368,9 +370,9 @@ ykpiv_rc _ykpiv_ensure_application_selected(ykpiv_state *state) {
     return YKPIV_GENERIC_ERROR;
   }
 
-  res = _ykpiv_verify(state, NULL, 0, NULL);
+  res = _ykpiv_verify(state, NULL, 0);
 
-  if ((YKPIV_OK != res) && (YKPIV_WRONG_PIN != res)) {
+  if ((YKPIV_OK != res) && (YKPIV_WRONG_PIN != res) && (YKPIV_PIN_LOCKED != res)) {
     res = _ykpiv_select_application(state);
   }
   else {
@@ -398,9 +400,9 @@ static ykpiv_rc _ykpiv_connect(ykpiv_state *state, uintptr_t context, uintptr_t
 
   // if card handle has changed, determine if handle is valid (less efficient, but complete)
   if ((card != state->card)) {
-    char reader[CB_BUF_MAX];
+    char reader[CB_BUF_MAX] = {0};
     pcsc_word reader_len = sizeof(reader);
-    uint8_t atr[CB_ATR_MAX];
+    uint8_t atr[CB_ATR_MAX] = {0};
     pcsc_word atr_len = sizeof(atr);
 
     // Cannot set the reader len to NULL.  Confirmed in OSX 10.10, so we have to retrieve it even though we don't need it.
@@ -443,9 +445,62 @@ ykpiv_rc ykpiv_connect_with_external_card(ykpiv_state *state, uintptr_t context,
   return _ykpiv_connect(state, context, card);
 }
 
+ykpiv_rc ykpiv_validate(ykpiv_state *state, const char *wanted) {
+  if(state->card) {
+    if(state->verbose) {
+      fprintf(stderr, "Validate reader '%s'.\n", wanted);
+    }
+    char reader[CB_BUF_MAX] = {0};
+    pcsc_word reader_len = sizeof(reader);
+    uint8_t atr[CB_ATR_MAX] = {0};
+    pcsc_word atr_len = sizeof(atr);
+    LONG rc = SCardStatus(state->card, reader, &reader_len, NULL, NULL, atr, &atr_len);
+    if(rc != SCARD_S_SUCCESS) {
+      if(state->verbose) {
+        fprintf (stderr, "SCardStatus failed on reader '%s', rc=%lx\n", wanted, (long)rc);
+      }
+      rc = SCardDisconnect(state->card, SCARD_RESET_CARD);
+      if(rc != SCARD_S_SUCCESS) {
+        if(state->verbose) {
+          fprintf (stderr, "SCardDisconnect failed on reader '%s', rc=%lx\n", wanted, (long)rc);
+        }
+      }
+      state->card = 0;
+      state->serial = 0;
+      state->ver.major = 0;
+      state->ver.minor = 0;
+      state->ver.patch = 0;
+      _cache_pin(state, NULL, 0);
+      _cache_mgm_key(state, NULL);
+      return YKPIV_PCSC_ERROR;
+    }
+    if (strcmp(wanted, reader)) {
+      if(state->verbose) {
+        fprintf (stderr, "Disconnecting incorrect reader '%s' (wanted '%s'), rc=%lx\n", reader, wanted, (long)rc);
+      }
+      rc = SCardDisconnect(state->card, SCARD_RESET_CARD);
+      if(rc != SCARD_S_SUCCESS) {
+        if(state->verbose) {
+          fprintf (stderr, "SCardDisconnect failed on reader '%s' (wanted '%s'), rc=%lx\n", reader, wanted, (long)rc);
+        }
+      }
+      state->card = 0;
+      state->serial = 0;
+      state->ver.major = 0;
+      state->ver.minor = 0;
+      state->ver.patch = 0;
+      _cache_pin(state, NULL, 0);
+      _cache_mgm_key(state, NULL);
+      return YKPIV_GENERIC_ERROR;
+    }
+    return YKPIV_OK;
+  }
+  return YKPIV_GENERIC_ERROR;
+}
+
 ykpiv_rc ykpiv_connect(ykpiv_state *state, const char *wanted) {
   pcsc_word active_protocol;
-  char reader_buf[2048];
+  char reader_buf[2048] = {0};
   size_t num_readers = sizeof(reader_buf);
   LONG rc;
   char *reader_ptr;
@@ -634,11 +689,10 @@ ykpiv_rc _ykpiv_begin_transaction(ykpiv_state *state) {
         return res;
     }
     if (state->pin) {
-      int tries;
-      if((res = _ykpiv_verify(state, state->pin, strlen(state->pin), &tries)) != YKPIV_OK)
+      if((res = _ykpiv_verify(state, state->pin, strlen(state->pin))) != YKPIV_OK)
         return res;
       // De-authenticate always-authenticate keys by running an arbitrary command
-      unsigned char data[80];
+      unsigned char data[80] = {0};
       unsigned long recv_len = sizeof(data);
       if((res = _ykpiv_fetch_object(state, YKPIV_OBJ_DISCOVERY, data, &recv_len)) != YKPIV_OK)
         return res;
@@ -672,11 +726,11 @@ ykpiv_rc _ykpiv_transfer_data(ykpiv_state *state, const unsigned char *templ,
 
   do {
     size_t this_size = 0xff;
-    unsigned char data[261];
+    unsigned char data[261] = {0};
     uint32_t recv_len = sizeof(data);
-    APDU apdu;
+    APDU apdu = {0};
 
-    memset(apdu.raw, 0, sizeof(apdu.raw));
+    memset(apdu.raw, 0, sizeof(apdu));
     memcpy(apdu.raw, templ, 4);
     if(in_ptr + 0xff < in_data + in_len) {
       apdu.st.cla = 0x10;
@@ -710,15 +764,15 @@ ykpiv_rc _ykpiv_transfer_data(ykpiv_state *state, const unsigned char *templ,
     in_ptr += this_size;
   } while(in_ptr < in_data + in_len);
   while(*sw >> 8 == 0x61) {
-    APDU apdu;
-    unsigned char data[261];
+    APDU apdu = {0};
+    unsigned char data[261] = {0};
     uint32_t recv_len = sizeof(data);
 
     if(state->verbose > 2) {
       fprintf(stderr, "The card indicates there is %d bytes more data for us.\n", *sw & 0xff);
     }
 
-    memset(apdu.raw, 0, sizeof(apdu.raw));
+    memset(apdu.raw, 0, sizeof(apdu));
     apdu.st.ins = YKPIV_INS_GET_RESPONSE_APDU;
     res = _send_data(state, &apdu, data, &recv_len, sw);
     if(res != YKPIV_OK) {
@@ -804,9 +858,9 @@ Cleanup:
 }
 
 static ykpiv_rc _ykpiv_authenticate(ykpiv_state *state, unsigned const char *key) {
-  APDU apdu;
-  unsigned char data[261];
-  unsigned char challenge[8];
+  APDU apdu = {0};
+  unsigned char data[261] = {0};
+  unsigned char challenge[8] = {0};
   uint32_t recv_len = sizeof(data);
   int sw;
   ykpiv_rc res;
@@ -850,7 +904,7 @@ static ykpiv_rc _ykpiv_authenticate(ykpiv_state *state, unsigned const char *key
   /* send a response to the cards challenge and a challenge of our own. */
   {
     unsigned char *dataptr = apdu.st.data;
-    unsigned char response[8];
+    unsigned char response[8] = {0};
     out_len = sizeof(response);
     drc = des_decrypt(mgm_key, challenge, sizeof(challenge), response, &out_len);
 
@@ -893,7 +947,7 @@ static ykpiv_rc _ykpiv_authenticate(ykpiv_state *state, unsigned const char *key
 
   /* compare the response from the card with our challenge */
   {
-    unsigned char response[8];
+    unsigned char response[8] = {0};
     out_len = sizeof(response);
     drc = des_encrypt(mgm_key, challenge, sizeof(challenge), response, &out_len);
 
@@ -925,8 +979,8 @@ ykpiv_rc ykpiv_set_mgmkey(ykpiv_state *state, const unsigned char *new_key) {
 }
 
 ykpiv_rc ykpiv_set_mgmkey2(ykpiv_state *state, const unsigned char *new_key, const unsigned char touch) {
-  APDU apdu;
-  unsigned char data[261];
+  APDU apdu = {0};
+  unsigned char data[261] = {0};
   uint32_t recv_len = sizeof(data);
   int sw;
   ykpiv_rc res = YKPIV_OK;
@@ -1014,14 +1068,14 @@ static ykpiv_rc _general_authenticate(ykpiv_state *state,
     const unsigned char *sign_in, size_t in_len,
     unsigned char *out, size_t *out_len,
     unsigned char algorithm, unsigned char key, bool decipher) {
-  unsigned char indata[1024];
+  unsigned char indata[1024] = {0};
   unsigned char *dataptr = indata;
-  unsigned char data[1024];
+  unsigned char data[1024] = {0};
   unsigned char templ[] = {0, YKPIV_INS_AUTHENTICATE, algorithm, key};
   unsigned long recv_len = sizeof(data);
   size_t key_len = 0;
   int sw = 0;
-  size_t bytes;
+  size_t bytes, offs;
   size_t len = 0;
   ykpiv_rc res;
 
@@ -1053,22 +1107,16 @@ static ykpiv_rc _general_authenticate(ykpiv_state *state,
     default:
       return YKPIV_ALGORITHM_ERROR;
   }
-  
-  if(in_len < 0x80) {
-    bytes = 1;
-  } else if(in_len < 0xff) {
-    bytes = 2;
-  } else {
-    bytes = 3;
-  }
 
+  bytes = _ykpiv_get_length_size(in_len);
+  
   *dataptr++ = 0x7c;
   dataptr += _ykpiv_set_length(dataptr, in_len + bytes + 3);
   *dataptr++ = 0x82;
   *dataptr++ = 0x00;
   *dataptr++ = YKPIV_IS_EC(algorithm) && decipher ? 0x85 : 0x81;
   dataptr += _ykpiv_set_length(dataptr, in_len);
-  memcpy(dataptr, sign_in, (size_t)in_len);
+  memcpy(dataptr, sign_in, in_len);
   dataptr += in_len;
 
   if((res = _ykpiv_transfer_data(state, templ, indata, (long)(dataptr - indata), data, &recv_len, &sw)) != YKPIV_OK) {
@@ -1093,21 +1141,23 @@ static ykpiv_rc _general_authenticate(ykpiv_state *state,
     return YKPIV_PARSE_ERROR;
   }
   dataptr = data + 1;
-  dataptr += _ykpiv_get_length(dataptr, &len);
+  offs = _ykpiv_get_length(dataptr, data + recv_len, &len);
+  dataptr += offs;
   /* skip the 82 tag */
-  if(*dataptr != 0x82) {
+  if(!offs || *dataptr != 0x82) {
     if(state->verbose) {
       fprintf(stderr, "Failed parsing signature reply.\n");
     }
     return YKPIV_PARSE_ERROR;
   }
   dataptr++;
-  dataptr += _ykpiv_get_length(dataptr, &len);
-  if(len > *out_len) {
+  offs = _ykpiv_get_length(dataptr, data + recv_len, &len);
+  dataptr += offs;
+  if(!offs || len > *out_len) {
     if(state->verbose) {
       fprintf(stderr, "Wrong size on output buffer.\n");
     }
-    return YKPIV_SIZE_ERROR;
+    return YKPIV_PARSE_ERROR;
   }
   *out_len = len;
   memcpy(out, dataptr, len);
@@ -1154,8 +1204,8 @@ ykpiv_rc ykpiv_decipher_data(ykpiv_state *state, const unsigned char *in,
 }
 
 static ykpiv_rc _ykpiv_get_version(ykpiv_state *state) {
-  APDU apdu;
-  unsigned char data[261];
+  APDU apdu = {0};
+  unsigned char data[261] = {0};
   uint32_t recv_len = sizeof(data);
   int sw;
   ykpiv_rc res;
@@ -1200,7 +1250,7 @@ ykpiv_rc ykpiv_get_version(ykpiv_state *state, char *version, size_t len) {
 
   if ((res = _ykpiv_get_version(state)) >= YKPIV_OK) {
     int result = snprintf(version, len, "%d.%d.%d", state->ver.major, state->ver.minor, state->ver.patch);
-    if(result < 0 || result >= len) {
+    if(result < 0 || result >= (int)len) {
       res = YKPIV_SIZE_ERROR;
     }
   }
@@ -1214,11 +1264,10 @@ Cleanup:
 /* caller must make sure that this is wrapped in a transaction for synchronized operation */
 static ykpiv_rc _ykpiv_get_serial(ykpiv_state *state) {
   ykpiv_rc res = YKPIV_OK;
-  APDU apdu;
-  uint8_t data[0xff];
+  APDU apdu = {0};
+  uint8_t data[0xff] = {0};
   uint32_t recv_len = sizeof(data);
   int sw;
-  uint8_t *p_temp = NULL;
 
   if (!state) {
     return YKPIV_ARGUMENT_ERROR;
@@ -1231,7 +1280,7 @@ static ykpiv_rc _ykpiv_get_serial(ykpiv_state *state) {
 
   if (state->ver.major < 5) {
     /* get serial from neo/yk4 devices using the otp applet */
-    uint8_t temp[0xff];
+    uint8_t temp[0xff] = {0};
 
     recv_len = sizeof(temp);
     memset(apdu.raw, 0, sizeof(apdu));
@@ -1335,6 +1384,8 @@ Cleanup:
 ykpiv_rc ykpiv_get_serial(ykpiv_state *state, uint32_t *p_serial) {
   ykpiv_rc res = YKPIV_OK;
 
+  if (!state || !p_serial) return YKPIV_ARGUMENT_ERROR;
+
   if ((res = _ykpiv_begin_transaction(state)) != YKPIV_OK) return res;
   if ((res = _ykpiv_ensure_application_selected(state)) != YKPIV_OK) goto Cleanup;
 
@@ -1402,19 +1453,19 @@ static ykpiv_rc _cache_mgm_key(ykpiv_state *state, unsigned const char *key) {
 #endif
 }
 
-static ykpiv_rc _ykpiv_verify(ykpiv_state *state, const char *pin, const size_t pin_len, int *tries) {
-  APDU apdu;
-  unsigned char data[261];
+static ykpiv_rc _ykpiv_verify(ykpiv_state *state, const char *pin, const size_t pin_len) {
+  APDU apdu = {0};
+  unsigned char data[261] = {0};
   uint32_t recv_len = sizeof(data);
   int sw;
   ykpiv_rc res;
 
   if (pin_len > CB_PIN_MAX) {
-    if (tries) *tries = 0;
+    state->tries = -1;
     return YKPIV_SIZE_ERROR;
   }
 
-  memset(apdu.raw, 0, sizeof(apdu.raw));
+  memset(apdu.raw, 0, sizeof(apdu));
   apdu.st.ins = YKPIV_INS_VERIFY;
   apdu.st.p1 = 0x00;
   apdu.st.p2 = 0x80;
@@ -1430,29 +1481,32 @@ static ykpiv_rc _ykpiv_verify(ykpiv_state *state, const char *pin, const size_t
   yc_memzero(&apdu, sizeof(apdu));
 
   if (res != YKPIV_OK) {
-    if (tries) *tries = 0;
+    state->tries = -1;
     return res;
   }
   else if (sw == SW_SUCCESS) {
-    if (pin && pin_len) {
+    if (pin) {
       // Intentionally ignore errors.  If the PIN fails to save, it will only
       // be a problem if a reconnect is attempted.  Failure deferred until then.
       _cache_pin(state, pin, pin_len);
     }
-
-    if (tries) *tries = (sw & 0xf);
+    state->tries = -1;
     return YKPIV_OK;
   }
   else if ((sw >> 8) == 0x63) {
-    if (tries) *tries = (sw & 0xf);
+    if(pin)
+      _cache_pin(state, NULL, 0);
+    state->tries = (sw & 0xf);
     return YKPIV_WRONG_PIN;
   }
   else if (sw == SW_ERR_AUTH_BLOCKED) {
-    if (tries) *tries = 0;
-    return YKPIV_WRONG_PIN;
+    if(pin)
+      _cache_pin(state, NULL, 0);
+    state->tries = 0;
+    return YKPIV_PIN_LOCKED;
   }
   else {
-    if (tries) *tries = 0;
+    state->tries = -1;
     return YKPIV_GENERIC_ERROR;
   }
 }
@@ -1467,7 +1521,8 @@ ykpiv_rc ykpiv_verify_select(ykpiv_state *state, const char *pin, const size_t p
   if (force_select) {
     if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
   }
-  res = _ykpiv_verify(state, pin, pin_len, tries);
+  res = _ykpiv_verify(state, pin, pin_len);
+  if(tries) *tries = state->tries;
 Cleanup:
 
   _ykpiv_end_transaction(state);
@@ -1481,6 +1536,12 @@ ykpiv_rc ykpiv_get_pin_retries(ykpiv_state *state, int *tries) {
     return YKPIV_ARGUMENT_ERROR;
   }
 
+  // Just get the stored value if we get the magic flag
+  if(*tries == YKPIV_RETRIES_MAX) {
+    *tries = state->tries;
+    return YKPIV_OK;
+  }
+
   if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
 
   // Force a re-select to unverify, because once verified the spec dictates that
@@ -1491,17 +1552,18 @@ ykpiv_rc ykpiv_get_pin_retries(ykpiv_state *state, int *tries) {
   if (res != YKPIV_OK) goto Cleanup;
   res = _ykpiv_select_application(state);
   if (res != YKPIV_OK) goto Cleanup;
-  res = _ykpiv_verify(state, NULL, 0, tries);
-  // WRONG_PIN is expected on successful query.
+  res = _ykpiv_verify(state, NULL, 0);
+  // WRONG_PIN or PIN_LOCKED is expected on successful query.
+  *tries = state->tries;
 Cleanup:
   _ykpiv_end_transaction(state);
-  return res == YKPIV_WRONG_PIN ? YKPIV_OK : res;
+  return (res == YKPIV_WRONG_PIN || res == YKPIV_PIN_LOCKED) ? YKPIV_OK : res;
 }
 
 ykpiv_rc ykpiv_set_pin_retries(ykpiv_state *state, int pin_tries, int puk_tries) {
   ykpiv_rc res = YKPIV_OK;
   unsigned char templ[] = {0, YKPIV_INS_SET_PIN_RETRIES, 0, 0};
-  unsigned char data[0xff];
+  unsigned char data[0xff] = {0};
   unsigned long recv_len = sizeof(data);
   int sw = 0;
 
@@ -1541,8 +1603,8 @@ Cleanup:
 static ykpiv_rc _ykpiv_change_pin(ykpiv_state *state, int action, const char * current_pin, size_t current_pin_len, const char * new_pin, size_t new_pin_len, int *tries) {
   int sw;
   unsigned char templ[] = {0, YKPIV_INS_CHANGE_REFERENCE, 0, 0x80};
-  unsigned char indata[0x10];
-  unsigned char data[0xff];
+  unsigned char indata[0x10] = {0};
+  unsigned char data[0xff] = {0};
   unsigned long recv_len = sizeof(data);
   ykpiv_rc res;
   if (current_pin_len > CB_PIN_MAX) {
@@ -1648,7 +1710,7 @@ Cleanup:
 ykpiv_rc _ykpiv_fetch_object(ykpiv_state *state, int object_id,
     unsigned char *data, unsigned long *len) {
   int sw;
-  unsigned char indata[5];
+  unsigned char indata[5] = {0};
   unsigned char *inptr = indata;
   unsigned char templ[] = {0, YKPIV_INS_GET_DATA, 0x3f, 0xff};
   ykpiv_rc res;
@@ -1665,15 +1727,9 @@ ykpiv_rc _ykpiv_fetch_object(ykpiv_state *state, int object_id,
 
   if(sw == SW_SUCCESS) {
     size_t outlen = 0;
-    unsigned int offs = 0;
-    
-    if ((*len < 2) || !_ykpiv_has_valid_length(data + 1, *len - 1)) {
-      return YKPIV_SIZE_ERROR;
-    }
-
-    offs = _ykpiv_get_length(data + 1, &outlen);
-    if(offs == 0) {
-      return YKPIV_SIZE_ERROR;
+    size_t offs = _ykpiv_get_length(data + 1, data + *len, &outlen);    
+    if(!offs) {
+      return YKPIV_PARSE_ERROR;
     }
     if(outlen + offs + 1 != *len) {
       if(state->verbose) {
@@ -1717,7 +1773,7 @@ ykpiv_rc _ykpiv_save_object(
     int object_id,
     unsigned char *indata,
     size_t len) {
-  unsigned char data[CB_BUF_MAX];
+  unsigned char data[CB_BUF_MAX] = {0};
   unsigned char *dataptr = data;
   unsigned char templ[] = {0, YKPIV_INS_PUT_DATA, 0x3f, 0xff};
   int sw;
@@ -1733,7 +1789,8 @@ ykpiv_rc _ykpiv_save_object(
   }
   *dataptr++ = 0x53;
   dataptr += _ykpiv_set_length(dataptr, len);
-  memcpy(dataptr, indata, len);
+  if(indata)
+    memcpy(dataptr, indata, len);
   dataptr += len;
 
   if((res = _ykpiv_transfer_data(state, templ, data, (long)(dataptr - data), NULL, &outlen,
@@ -1761,15 +1818,15 @@ ykpiv_rc ykpiv_import_private_key(ykpiv_state *state, const unsigned char key, u
                                   const unsigned char *ec_data, unsigned char ec_data_len,
                                   const unsigned char pin_policy, const unsigned char touch_policy) {
 
-  unsigned char key_data[1024];
+  unsigned char key_data[1024] = {0};
   unsigned char *in_ptr = key_data;
   unsigned char templ[] = {0, YKPIV_INS_IMPORT_KEY, algorithm, key};
-  unsigned char data[256];
+  unsigned char data[256] = {0};
   unsigned long recv_len = sizeof(data);
   unsigned elem_len;
   int sw;
-  const unsigned char *params[5];
-  size_t lens[5];
+  const unsigned char *params[5] = {0};
+  size_t lens[5] = {0};
   size_t padding;
   unsigned char n_params;
   int i;
@@ -1829,13 +1886,10 @@ ykpiv_rc ykpiv_import_private_key(ykpiv_state *state, const unsigned char key, u
   }
   else if (algorithm == YKPIV_ALGO_ECCP256 || algorithm == YKPIV_ALGO_ECCP384) {
 
-#pragma GCC diagnostic push
-#pragma GCC diagnostic ignored "-Wtype-limits"
     if ((size_t)ec_data_len >= sizeof(key_data)) {
       /* This can never be true, but check to be explicit. */
       return YKPIV_SIZE_ERROR;
     }
-#pragma GCC diagnostic pop
 
     if (algorithm == YKPIV_ALGO_ECCP256)
       elem_len = 32;
@@ -2082,8 +2136,8 @@ ykpiv_rc ykpiv_auth_deauthenticate(ykpiv_state *state) {
 /* deauthenticates the user pin and mgm key */
 static ykpiv_rc _ykpiv_auth_deauthenticate(ykpiv_state *state) {
   ykpiv_rc res = YKPIV_OK;
-  APDU apdu;
-  unsigned char data[0xff];
+  APDU apdu = {0};
+  unsigned char data[0xff] = {0};
   uint32_t recv_len = sizeof(data);
   int sw;
 
