@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2015-2016 Yubico AB
+ * Copyright (c) 2015-2017,2020 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -34,12 +34,14 @@
 #include <stdio.h>
 #include <string.h>
 #include <assert.h>
-#include <unistd.h>
 
-#include "util.h"
+#include "../../common/util.h"
 
 #ifdef _WIN32
+#include <openssl/applink.c>
 #define pipe(fds) _pipe(fds,4096, 0)
+#else
+#include <unistd.h>
 #endif
 
 enum enum_format formats[] = {
@@ -50,8 +52,8 @@ enum enum_format formats[] = {
 
 static bool inout(enum enum_format format) {
   const unsigned char buf[] = {0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08};
-  unsigned char buf2[sizeof(buf)];
-  int pipefd[2];
+  unsigned char buf2[sizeof(buf)] = {0};
+  int pipefd[2] = {0};
   FILE *tmp1, *tmp2;
 
   if (pipe(pipefd) != 0)
