@@ -1,5 +1,5 @@
  /*
- * Copyright (c) 2014-2016 Yubico AB
+ * Copyright (c) 2014-2020 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -31,18 +31,21 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdbool.h>
+#include <limits.h>
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <unistd.h>
 
 #include "ykpiv.h"
 
 #ifdef _WIN32
 #include <windows.h>
+#include <openssl/applink.c>
+#else
+#include <unistd.h>
 #endif
 
-#include "openssl-compat.h"
+#include "../common/openssl-compat.h"
 #include <openssl/bn.h>
 #include <openssl/des.h>
 #include <openssl/pem.h>
@@ -52,7 +55,7 @@
 #include <openssl/x509v3.h>
 
 #include "cmdline.h"
-#include "util.h"
+#include "../common/util.h"
 
 #define MAX(a,b) (a) > (b) ? (a) : (b)
 
@@ -63,6 +66,8 @@
 
 #define KEY_LEN 24
 
+#define YKPIV_ATTESTATION_OID "1.3.6.1.4.1.41482.3"
+
 static enum file_mode key_file_mode(enum enum_key_format fmt, bool output) {
   if (fmt == key_format_arg_PEM) {
     if (output) {
@@ -90,7 +95,7 @@ static enum file_mode data_file_mode(enum enum_format fmt, bool output) {
 }
 
 static void print_version(ykpiv_state *state, const char *output_file_name) {
-  char version[7];
+  char version[7] = {0};
   FILE *output_file = open_file(output_file_name, OUTPUT_TEXT);
   if(!output_file) {
     return;
@@ -110,7 +115,7 @@ static void print_version(ykpiv_state *state, const char *output_file_name) {
 static bool sign_data(ykpiv_state *state, const unsigned char *in, size_t len, unsigned char *out,
     size_t *out_len, unsigned char algorithm, int key) {
 
-  unsigned char signinput[1024];
+  unsigned char signinput[1024] = {0};
   if(YKPIV_IS_RSA(algorithm)) {
     size_t padlen = algorithm == YKPIV_ALGO_RSA1024 ? 128 : 256;
     if(RSA_padding_add_PKCS1_type_1(signinput, padlen, in, len) == 0) {
@@ -171,7 +176,7 @@ int yk_ec_meth_sign(int type, const unsigned char *dgst, int dlen,
   return 0;
 }
 
-static int wrap_public_key(ykpiv_state *state, int algorithm, EVP_PKEY *public_key,
+static void wrap_public_key(ykpiv_state *state, int algorithm, EVP_PKEY *public_key,
     int key, const unsigned char *oid, size_t oid_len) {
   static struct internal_key int_key;
   int_key.state = state;
@@ -182,19 +187,28 @@ static int wrap_public_key(ykpiv_state *state, int algorithm, EVP_PKEY *public_k
   if(YKPIV_IS_RSA(algorithm)) {
     RSA_METHOD *meth = RSA_meth_dup(RSA_get_default_method());
     RSA *rsa = EVP_PKEY_get0_RSA(public_key);
-    RSA_meth_set0_app_data(meth, &int_key);
-    RSA_meth_set_sign(meth, yk_rsa_meth_sign);
-    RSA_set_method(rsa, meth);
+    if(RSA_meth_set0_app_data(meth, &int_key) != 1) {
+      fprintf(stderr, "Failed to set RSA data\n");
+    }
+    if(RSA_meth_set_sign(meth, yk_rsa_meth_sign) != 1) {
+      fprintf(stderr, "Failed to set RSA signature verification method\n");
+    }
+    if(RSA_set_method(rsa, meth) != 1) {
+      fprintf(stderr, "Failed to wrap RSA key\n");
+    }
   } else {
     EC_KEY *ec = EVP_PKEY_get0_EC_KEY(public_key);
     EC_KEY_METHOD *meth = EC_KEY_METHOD_new(EC_KEY_get_method(ec));
     if (ec_key_ex_data_idx == -1)
       ec_key_ex_data_idx = EC_KEY_get_ex_new_index(0, NULL, NULL, NULL, 0);
-    EC_KEY_set_ex_data(ec, ec_key_ex_data_idx, &int_key);
-    EC_KEY_METHOD_set_sign(meth, yk_ec_meth_sign, NULL, NULL); /* XXX ?? */
-    EC_KEY_set_method(ec, meth);
+    if(EC_KEY_set_ex_data(ec, ec_key_ex_data_idx, &int_key) != 1) {
+      fprintf(stderr, "Failed to set EC data\n");
+    }
+    EC_KEY_METHOD_set_sign(meth, yk_ec_meth_sign, NULL, NULL);
+    if(EC_KEY_set_method(ec, meth) != 1) {
+      fprintf(stderr, "Failed to wrap public EC key\n");
+    }
   }
-  return 0;
 }
 #endif
 
@@ -209,6 +223,7 @@ static bool generate_key(ykpiv_state *state, enum enum_slot slot,
   EVP_PKEY *public_key = NULL;
   RSA *rsa = NULL;
   EC_KEY *eckey = NULL;
+  EC_GROUP *group = NULL;
   EC_POINT *ecpoint = NULL;
   uint8_t *mod = NULL;
   uint8_t *exp = NULL;
@@ -224,6 +239,11 @@ static bool generate_key(ykpiv_state *state, enum enum_slot slot,
     return false;
   }
 
+  if(algorithm == algorithm_arg_RSA1024) {
+    fprintf(stderr, "\nWARNING. The use of RSA1024 is discouraged by the National Institute of Standards "
+                         "and Technology (NIST). See https://www.yubico.com/blog/comparing-asymmetric-encryption-algorithms\n\n");
+  }
+
   res = ykpiv_util_generate_key(state,
                                 (uint8_t)(key & 0xFF),
                                 get_piv_algorithm(algorithm),
@@ -257,10 +277,15 @@ static bool generate_key(ykpiv_state *state, enum enum_slot slot,
         goto generate_out;
       }
 
-      RSA_set0_key(rsa, bignum_n, bignum_e, NULL);
-      EVP_PKEY_set1_RSA(public_key, rsa);
+      if(RSA_set0_key(rsa, bignum_n, bignum_e, NULL) != 1) {
+        fprintf(stderr, "Failed to set RSA key\n");
+        goto generate_out;
+      }
+      if(EVP_PKEY_set1_RSA(public_key, rsa) != 1) {
+        fprintf(stderr, "Failed to set RSA public key\n");
+        goto generate_out;
+      }
     } else if(algorithm == algorithm_arg_ECCP256 || algorithm == algorithm_arg_ECCP384) {
-      EC_GROUP *group;
       int nid;
 
       if(algorithm == algorithm_arg_ECCP256) {
@@ -271,7 +296,10 @@ static bool generate_key(ykpiv_state *state, enum enum_slot slot,
       eckey = EC_KEY_new();
       group = EC_GROUP_new_by_curve_name(nid);
       EC_GROUP_set_asn1_flag(group, nid);
-      EC_KEY_set_group(eckey, group);
+      if(EC_KEY_set_group(eckey, group) != 1) {
+        fprintf(stderr, "Failed to set EC group.\n");
+        goto generate_out;
+      }
       ecpoint = EC_POINT_new(group);
 
       if(!EC_POINT_oct2point(group, ecpoint, point, point_len, NULL)) {
@@ -282,12 +310,19 @@ static bool generate_key(ykpiv_state *state, enum enum_slot slot,
         fprintf(stderr, "Failed to set the public key.\n");
         goto generate_out;
       }
-      EVP_PKEY_set1_EC_KEY(public_key, eckey);
+      if(EVP_PKEY_set1_EC_KEY(public_key, eckey) != 1) {
+        fprintf(stderr, "Failed to set EC public key.\n");
+        goto generate_out;
+      }
     } else {
       fprintf(stderr, "Wrong algorithm.\n");
     }
-    PEM_write_PUBKEY(output_file, public_key);
-    ret = true;
+    if(PEM_write_PUBKEY(output_file, public_key) == 1) {
+      ret = true;
+    } else {
+      fprintf(stderr, "Failed to write public key in PEM format\n");
+      goto generate_out;
+    }
   } else {
     fprintf(stderr, "Only PEM is supported as public_key output.\n");
     goto generate_out;
@@ -297,6 +332,9 @@ generate_out:
   if (output_file != stdout) {
     fclose(output_file);
   }
+  if (group) {
+    EC_GROUP_clear_free(group);
+  }
   if (ecpoint) {
     EC_POINT_free(ecpoint);
   }
@@ -373,7 +411,7 @@ static bool import_key(ykpiv_state *state, enum enum_key_format key_format,
       fprintf(stderr, "Failed to load PKCS12 from file.\n");
       goto import_out;
     }
-    if(PKCS12_parse(p12, password, &private_key, &cert, NULL) == 0) {
+    if(PKCS12_parse(p12, password, &private_key, &cert, NULL) == 0 || private_key == NULL) {
       fprintf(stderr, "Failed to parse PKCS12 structure. (wrong password?)\n");
       goto import_out;
     }
@@ -402,12 +440,12 @@ static bool import_key(ykpiv_state *state, enum enum_key_format key_format,
 
     if(YKPIV_IS_RSA(algorithm)) {
       RSA *rsa_private_key = EVP_PKEY_get1_RSA(private_key);
-      unsigned char e[4];
-      unsigned char p[128];
-      unsigned char q[128];
-      unsigned char dmp1[128];
-      unsigned char dmq1[128];
-      unsigned char iqmp[128];
+      unsigned char e[4] = {0};
+      unsigned char p[128] = {0};
+      unsigned char q[128] = {0};
+      unsigned char dmp1[128] = {0};
+      unsigned char dmq1[128] = {0};
+      unsigned char iqmp[128] = {0};
       const BIGNUM *bn_e, *bn_p, *bn_q, *bn_dmp1, *bn_dmq1, *bn_iqmp;
 
       int element_len = 128;
@@ -461,7 +499,7 @@ static bool import_key(ykpiv_state *state, enum enum_key_format key_format,
     else if(YKPIV_IS_EC(algorithm)) {
       EC_KEY *ec = EVP_PKEY_get1_EC_KEY(private_key);
       const BIGNUM *s = EC_KEY_get0_private_key(ec);
-      unsigned char s_ptr[48];
+      unsigned char s_ptr[48] = {0};
 
       int element_len = 32;
       if(algorithm == YKPIV_ALGO_ECCP384) {
@@ -557,6 +595,10 @@ static bool import_cert(ykpiv_state *state, enum enum_key_format cert_format,
       fprintf(stderr, "Failed checking input GZIP file.\n");
       goto import_cert_out;
     }
+    if (st.st_size > INT_MAX) {
+      fprintf(stderr, "Size of certificate file too large.\n");
+      goto import_cert_out;
+    }
     cert_len = st.st_size;
     compress = 0x01;
   } else {
@@ -569,11 +611,11 @@ static bool import_cert(ykpiv_state *state, enum enum_key_format cert_format,
   }
 
   {
-    unsigned char certdata[YKPIV_OBJ_MAX_SIZE];
+    unsigned char certdata[YKPIV_OBJ_MAX_SIZE] = {0};
     unsigned char *certptr = certdata;
     ykpiv_rc res;
 
-    if(cert_len > YKPIV_OBJ_MAX_SIZE) {
+    if(cert_len > YKPIV_OBJ_MAX_SIZE || cert_len < 0) {
       fprintf(stderr, "Length of certificate is more than can fit.\n");
       goto import_cert_out;
     }
@@ -584,7 +626,10 @@ static bool import_cert(ykpiv_state *state, enum enum_key_format cert_format,
         goto import_cert_out;
       }
     } else {
-      i2d_X509(cert, &certptr);
+      if(i2d_X509(cert, &certptr) < 0) {
+        fprintf(stderr, "Failed to encode X509 certificate\n");
+        goto import_cert_out;
+      }
     }
     if ((res = ykpiv_util_write_cert(state, get_slot_hex(slot), certdata, (size_t)cert_len, compress)) != YKPIV_OK) {
       fprintf(stderr, "Failed commands with device: %s\n", ykpiv_strerror(res));
@@ -612,7 +657,7 @@ import_cert_out:
 
 static bool set_cardid(ykpiv_state *state, int verbose, int type) {
   ykpiv_rc res;
-  unsigned char id[MAX(sizeof(ykpiv_cardid), sizeof(ykpiv_cccid))];
+  unsigned char id[MAX(sizeof(ykpiv_cardid), sizeof(ykpiv_cccid))] = {0};
 
   if(type == CHUID) {
     res = ykpiv_util_set_cardid(state, NULL);
@@ -634,9 +679,36 @@ static bool set_cardid(ykpiv_state *state, int verbose, int type) {
   return res == YKPIV_OK;
 }
 
+static int add_ext(STACK_OF(X509_EXTENSION) *exts, const char *oid, const char *name, const char *descr, const unsigned char *data, int len) {
+  int nid = OBJ_create(oid, name, descr);
+  if(nid <= 0) {
+    fprintf(stderr, "Failed creating %s extension object.\n", name);
+    return 0;
+  }
+  ASN1_OCTET_STRING *octets = ASN1_OCTET_STRING_new();
+  if(!octets) {
+    fprintf(stderr, "Failed allocating octets for %s extension.\n", name);
+    return 0;
+  }
+  if(!ASN1_OCTET_STRING_set(octets, data, len)) {
+    fprintf(stderr, "Failed setting octets for %s extension.\n", name);
+    return 0;
+  }
+  X509_EXTENSION *ext = X509_EXTENSION_create_by_NID(NULL, nid, 0, octets);
+  if(!ext) {
+    fprintf(stderr, "Failed creating %s extension.\n", name);
+    return 0;
+  }
+  if(!sk_X509_EXTENSION_push(exts, ext)) {
+    fprintf(stderr, "Failed pushing %s extension.\n", name);
+    return 0;
+  }
+  return 1;
+}
+
 static bool request_certificate(ykpiv_state *state, enum enum_key_format key_format,
     const char *input_file_name, enum enum_slot slot, char *subject, enum enum_hash hash,
-    const char *output_file_name) {
+    const char *output_file_name, int attestation) {
   X509_REQ *req = NULL;
   X509_NAME *name = NULL;
   FILE *input_file = NULL;
@@ -649,7 +721,7 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
   size_t oid_len;
   const unsigned char *oid;
 #if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
-  unsigned char digest[EVP_MAX_MD_SIZE + MAX_OID_LEN];
+  unsigned char digest[EVP_MAX_MD_SIZE + MAX_OID_LEN] = {0};
   unsigned int md_len;
   unsigned int digest_len;
   unsigned char *signinput;
@@ -666,20 +738,69 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
     goto request_out;
   }
 
-  if(isatty(fileno(input_file))) {
-    fprintf(stderr, "Please paste the public key...\n");
+  req = X509_REQ_new();
+  if(!req) {
+    fprintf(stderr, "Failed to allocate request structure.\n");
+    goto request_out;
   }
 
-  if(key_format == key_format_arg_PEM) {
-    public_key = PEM_read_PUBKEY(input_file, NULL, NULL, NULL);
-    if(!public_key) {
-      fprintf(stderr, "Failed loading public key for request.\n");
+  if(attestation) {
+    unsigned char buf[YKPIV_OBJ_MAX_SIZE] = {0};
+    size_t buflen = sizeof(buf);
+    ykpiv_rc rc;
+
+    if((rc = ykpiv_attest(state, key, buf, &buflen)) == YKPIV_OK) {
+      STACK_OF(X509_EXTENSION) *exts = sk_X509_EXTENSION_new_null();
+      add_ext(exts, YKPIV_ATTESTATION_OID ".11", "ykpiv attestation", "Yubico PIV X.509 Attestation", buf, buflen);
+
+      unsigned char *pb = 0;
+      size_t pblen = 0;
+      if((rc = ykpiv_util_read_cert(state, YKPIV_KEY_ATTESTATION, &pb, &pblen)) == YKPIV_OK && pblen > 0) {
+        add_ext(exts, YKPIV_ATTESTATION_OID ".2", "ykpiv attest cert", "Yubico PIV Attestation Certificate", pb, pblen);
+        ykpiv_util_free(state, pb);
+      } else {
+        fprintf(stderr, "Failed reading attestation certificate: %s.\n", ykpiv_strerror(rc));        
+      }
+
+      if(!X509_REQ_add_extensions(req, exts)) {
+        fprintf(stderr, "Failed setting the request extensions.\n");
+        sk_X509_EXTENSION_pop_free(exts, X509_EXTENSION_free);
+        goto request_out;
+      }
+      sk_X509_EXTENSION_pop_free(exts, X509_EXTENSION_free);
+
+      // Extract the public key for the request from the attestation
+      const unsigned char *ptr = buf;
+      X509 *x509 = d2i_X509(NULL, &ptr, buflen);
+      if(x509) {
+        public_key = X509_get_pubkey(x509);
+        X509_free(x509);
+      }
+      if(!public_key) {
+        fprintf(stderr, "Failed extracting public key for request from attestation.\n");
+        goto request_out;
+      }
+    } else {
+      fprintf(stderr, "Failed creating attestation: %s.\n", ykpiv_strerror(rc));
       goto request_out;
     }
   } else {
-    fprintf(stderr, "Only PEM supported for public key input.\n");
-    goto request_out;
+    if(isatty(fileno(input_file))) {
+      fprintf(stderr, "Please paste the public key...\n");
+    }
+
+    if(key_format == key_format_arg_PEM) {
+      public_key = PEM_read_PUBKEY(input_file, NULL, NULL, NULL);
+      if(!public_key) {
+        fprintf(stderr, "Failed loading public key for request.\n");
+        goto request_out;
+      }
+    } else {
+      fprintf(stderr, "Only PEM supported for public key input.\n");
+      goto request_out;
+    }
   }
+
   algorithm = get_algorithm(public_key);
   if(algorithm == 0) {
     goto request_out;
@@ -690,17 +811,14 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
     goto request_out;
   }
 
-  req = X509_REQ_new();
-  if(!req) {
-    fprintf(stderr, "Failed to allocate request structure.\n");
-    goto request_out;
-  }
   if(!X509_REQ_set_pubkey(req, public_key)) {
     fprintf(stderr, "Failed setting the request public key.\n");
     goto request_out;
   }
 
-  X509_REQ_set_version(req, 0);
+  if(X509_REQ_set_version(req, 0) != 1) {
+    fprintf(stderr, "Failed setting the certificate request version.\n");
+  }
 
   name = parse_name(subject);
   if(!name) {
@@ -745,7 +863,7 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
 
   req->sig_alg->algorithm = OBJ_nid2obj(nid);
   {
-    unsigned char signature[1024];
+    unsigned char signature[1024] = {0};
     size_t sig_len = sizeof(signature);
     if(!sign_data(state, signinput, len, signature, &sig_len, algorithm, key)) {
       fprintf(stderr, "Failed signing request.\n");
@@ -767,8 +885,11 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
 #endif
 
   if(key_format == key_format_arg_PEM) {
-    PEM_write_X509_REQ(output_file, req);
-    ret = true;
+    if(PEM_write_X509_REQ(output_file, req) == 1) {
+      ret = true;
+    } else {
+      fprintf(stderr, "Failed writing x509 information\n");
+    }
   } else {
     fprintf(stderr, "Only PEM support available for certificate requests.\n");
   }
@@ -825,7 +946,7 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
   ASN1_INTEGER *sno = ASN1_INTEGER_new();
   BIGNUM *ser = NULL;
 #if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
-  unsigned char digest[EVP_MAX_MD_SIZE + MAX_OID_LEN];
+  unsigned char digest[EVP_MAX_MD_SIZE + MAX_OID_LEN] = {0};
   unsigned int digest_len;
   unsigned int md_len;
   unsigned char *signinput;
@@ -878,7 +999,10 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
     goto selfsign_out;
   }
   if(serial) {
-    ASN1_INTEGER_set(sno, *serial);
+    if(ASN1_INTEGER_set(sno, *serial) != 1) {
+      fprintf(stderr, "Failed to read serial number.\n");
+      goto selfsign_out;
+    }
   } else {
     ser = BN_new();
     if(!ser) {
@@ -898,19 +1022,11 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
     fprintf(stderr, "Failed to set certificate serial.\n");
     goto selfsign_out;
   }
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
   if(!X509_gmtime_adj(X509_get_notBefore(x509), 0)) {
-#else
-  if(!X509_gmtime_adj(X509_getm_notBefore(x509), 0)) {
-#endif
     fprintf(stderr, "Failed to set certificate notBefore.\n");
     goto selfsign_out;
   }
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
   if(!X509_gmtime_adj(X509_get_notAfter(x509), 60L * 60L * 24L * validDays)) {
-#else
-  if(!X509_gmtime_adj(X509_getm_notAfter(x509), 60L * 60L * 24L * validDays)) {
-#endif
     fprintf(stderr, "Failed to set certificate notAfter.\n");
     goto selfsign_out;
   }
@@ -1001,7 +1117,7 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
     goto selfsign_out;
   }
   {
-    unsigned char signature[1024];
+    unsigned char signature[1024] = {0};
     size_t sig_len = sizeof(signature);
     if(!sign_data(state, signinput, len, signature, &sig_len, algorithm, key)) {
       fprintf(stderr, "Failed signing certificate.\n");
@@ -1025,8 +1141,12 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
 #endif
 
   if(key_format == key_format_arg_PEM) {
-    PEM_write_X509(output_file, x509);
-    ret = true;
+    if(PEM_write_X509(output_file, x509) == 1) {
+      ret = true;
+    } else {
+      fprintf(stderr, "Failed writing x509 information\n");
+    }
+
   } else {
     fprintf(stderr, "Only PEM support available for certificate requests.\n");
   }
@@ -1075,7 +1195,7 @@ static bool verify_pin(ykpiv_state *state, const char *pin) {
   res = ykpiv_verify(state, pin, &tries);
   if(res == YKPIV_OK) {
     return true;
-  } else if(res == YKPIV_WRONG_PIN) {
+  } else if(res == YKPIV_WRONG_PIN || res == YKPIV_PIN_LOCKED) {
     if(tries > 0) {
       fprintf(stderr, "Pin verification failed, %d tries left before pin is blocked.\n", tries);
     } else {
@@ -1120,8 +1240,6 @@ static bool change_pin(ykpiv_state *state, enum enum_action action, const char *
   }
   res = op(state, pin, pin_len, new_pin, new_len, &tries);
 
-#pragma GCC diagnostic push
-#pragma GCC diagnostic ignored "-Wswitch-enum"
   switch (res) {
     case YKPIV_OK:
       return true;
@@ -1143,7 +1261,6 @@ static bool change_pin(ykpiv_state *state, enum enum_action action, const char *
       fprintf(stderr, "Failed changing/unblocking code, error: %s\n", ykpiv_strerror(res));
       return false;
   }
-#pragma GCC diagnostic pop
 }
 
 static bool delete_certificate(ykpiv_state *state, enum enum_slot slot) {
@@ -1186,8 +1303,11 @@ static bool read_certificate(ykpiv_state *state, enum enum_slot slot,
     }
 
     if (key_format == key_format_arg_PEM) {
-      PEM_write_X509(output_file, x509);
-      ret = true;
+      if(PEM_write_X509(output_file, x509) == 1) {
+        ret = true;
+      } else {
+        fprintf(stderr, "Failed writing x509 information\n");
+      }
     }
     else {
       if (!SSH_write_X509(output_file, x509)) {
@@ -1222,7 +1342,7 @@ static bool sign_file(ykpiv_state *state, const char *input, const char *output,
   FILE *output_file = NULL;
   int key;
   unsigned int hash_len;
-  unsigned char hashed[EVP_MAX_MD_SIZE];
+  unsigned char hashed[EVP_MAX_MD_SIZE * 2] = {0};
   bool ret = false;
   int algo;
   const EVP_MD *md;
@@ -1260,13 +1380,22 @@ static bool sign_file(ykpiv_state *state, const char *input, const char *output,
     }
 
     mdctx = EVP_MD_CTX_create();
-    EVP_DigestInit_ex(mdctx, md, NULL);
+    if(EVP_DigestInit_ex(mdctx, md, NULL) != 1) {
+      fprintf(stderr, "failed to initialize digest operation\n");
+      goto out;
+    }
     while(!feof(input_file)) {
-      char buf[1024];
+      char buf[1024] = {0};
       size_t len = fread(buf, 1, 1024, input_file);
-      EVP_DigestUpdate(mdctx, buf, len);
+      if(EVP_DigestUpdate(mdctx, buf, len) != 1) {
+        fprintf(stderr, "failed to update digest data\n");
+        goto out;
+      }
+    }
+    if(EVP_DigestFinal_ex(mdctx, hashed, &hash_len) != 1) {
+      fprintf(stderr, "failed to finalize digest operation\n");
+      goto out;
     }
-    EVP_DigestFinal_ex(mdctx, hashed, &hash_len);
 
     if(verbosity) {
       fprintf(stderr, "file hashed as: ");
@@ -1280,7 +1409,7 @@ static bool sign_file(ykpiv_state *state, const char *input, const char *output,
   }
 
   {
-    unsigned char buf[1024];
+    unsigned char buf[1024] = {0};
     size_t len = sizeof(buf);
     if(!sign_data(state, hashed, hash_len, buf, &len, algo, key)) {
       fprintf(stderr, "failed signing file\n");
@@ -1311,10 +1440,10 @@ static void print_cert_info(ykpiv_state *state, enum enum_slot slot, const EVP_M
     FILE *output) {
   int object = (int)ykpiv_util_slot_object(get_slot_hex(slot));
   int slot_name;
-  unsigned char data[YKPIV_OBJ_MAX_SIZE];
+  unsigned char data[YKPIV_OBJ_MAX_SIZE] = {0};
   const unsigned char *ptr = data;
   unsigned long len = sizeof(data);
-  unsigned long cert_len;
+  unsigned long offs, cert_len;
   X509 *x509 = NULL;
   X509_NAME *subj;
   BIO *bio = NULL;
@@ -1331,10 +1460,15 @@ static void print_cert_info(ykpiv_state *state, enum enum_slot slot, const EVP_M
     unsigned int md_len = sizeof(data);
     const ASN1_TIME *not_before, *not_after;
 
-    ptr += get_length(ptr, &cert_len);
+    offs = get_length(ptr, data + len, &cert_len);
+    if(!offs) {
+      fprintf(output, "Invalid cert length.\n");
+      goto cert_out;
+    }
+    ptr += offs;
     x509 = d2i_X509(NULL, &ptr, cert_len);
     if(!x509) {
-      fprintf(output, "Unknown data present.\n");
+      fprintf(output, "Invalid cert data.\n");
       goto cert_out;
     }
     {
@@ -1360,6 +1494,7 @@ static void print_cert_info(ykpiv_state *state, enum enum_slot slot, const EVP_M
         default:
           fprintf(output, "Unknown\n");
       }
+      EVP_PKEY_free(key);
     }
     subj = X509_get_subject_name(x509);
     if(!subj) {
@@ -1367,7 +1502,10 @@ static void print_cert_info(ykpiv_state *state, enum enum_slot slot, const EVP_M
       goto cert_out;
     }
     fprintf(output, "\tSubject DN:\t");
-    X509_NAME_print_ex_fp(output, subj, 0, XN_FLAG_COMPAT);
+    if(X509_NAME_print_ex_fp(output, subj, 0, XN_FLAG_COMPAT) != 1) {
+      fprintf(output, "Failed to write Subject DN.\n");
+      goto cert_out;
+    }
     fprintf(output, "\n");
     subj = X509_get_issuer_name(x509);
     if(!subj) {
@@ -1375,31 +1513,35 @@ static void print_cert_info(ykpiv_state *state, enum enum_slot slot, const EVP_M
       goto cert_out;
     }
     fprintf(output, "\tIssuer DN:\t");
-    X509_NAME_print_ex_fp(output, subj, 0, XN_FLAG_COMPAT);
+    if(X509_NAME_print_ex_fp(output, subj, 0, XN_FLAG_COMPAT) != 1) {
+      fprintf(output, "Failed to write Issuer DN.\n");
+      goto cert_out;
+    }
     fprintf(output, "\n");
-    X509_digest(x509, md, data, &md_len);
+    if(X509_digest(x509, md, data, &md_len) != 1) {
+      fprintf(output, "Failed to digest data.\n");
+      goto cert_out;
+    }
     fprintf(output, "\tFingerprint:\t");
     dump_data(data, md_len, output, false, format_arg_hex);
 
     bio = BIO_new_fp(output, BIO_NOCLOSE | BIO_FP_TEXT);
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
     not_before = X509_get_notBefore(x509);
-#else
-    not_before = X509_get0_notBefore(x509);
-#endif
     if(not_before) {
       fprintf(output, "\tNot Before:\t");
-      ASN1_TIME_print(bio, not_before);
+      if(ASN1_TIME_print(bio, not_before) != 1) {
+        fprintf(output, "Failed to write Not Before time.\n");
+        goto cert_out;
+      }
       fprintf(output, "\n");
     }
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
     not_after = X509_get_notAfter(x509);
-#else
-    not_after = X509_get0_notAfter(x509);
-#endif
     if(not_after) {
       fprintf(output, "\tNot After:\t");
-      ASN1_TIME_print(bio, not_after);
+      if(ASN1_TIME_print(bio, not_after) != 1) {
+        fprintf(output, "Failed to write Not After time.\n");
+        goto cert_out;
+      }
       fprintf(output, "\n");
     }
   } else {
@@ -1419,7 +1561,7 @@ static bool status(ykpiv_state *state, enum enum_hash hash,
                    enum enum_slot slot,
                    const char *output_file_name) {
   const EVP_MD *md;
-  unsigned char buf[YKPIV_OBJ_MAX_SIZE];
+  unsigned char buf[YKPIV_OBJ_MAX_SIZE] = {0};
   long unsigned len = sizeof(buf);
   int i;
   uint32_t serial = 0;
@@ -1487,10 +1629,10 @@ static bool test_signature(ykpiv_state *state, enum enum_slot slot,
     enum enum_key_format cert_format, int verbose) {
   const EVP_MD *md;
   bool ret = false;
-  unsigned char data[1024];
+  unsigned char data[1024] = {0};
   unsigned int data_len;
   X509 *x509 = NULL;
-  EVP_PKEY *pubkey;
+  EVP_PKEY *pubkey = NULL;
   FILE *input_file = open_file(input_file_name, key_file_mode(cert_format, false));
 
   if(!input_file) {
@@ -1517,30 +1659,40 @@ static bool test_signature(ykpiv_state *state, enum enum_slot slot,
 
   md = get_hash(hash, NULL, NULL);
   if(md == NULL) {
-    return false;
+    goto test_out;
   }
 
   {
-    unsigned char rand[128];
+    unsigned char rand[128] = {0};
     EVP_MD_CTX *mdctx;
     if(RAND_bytes(rand, sizeof(rand)) <= 0) {
       fprintf(stderr, "error: no randomness.\n");
-      return false;
+      goto test_out;
     }
 
     mdctx = EVP_MD_CTX_create();
-    EVP_DigestInit_ex(mdctx, md, NULL);
-    EVP_DigestUpdate(mdctx, rand, 128);
-    EVP_DigestFinal_ex(mdctx, data, &data_len);
+    if(EVP_DigestInit_ex(mdctx, md, NULL) != 1) {
+      fprintf(stderr, "Failed to initialize digest operation\n");
+      goto test_out;
+    }
+    if(EVP_DigestUpdate(mdctx, rand, 128) != 1) {
+      fprintf(stderr, "Failed to update digest data\n");
+      goto test_out;
+    }
+    if(EVP_DigestFinal_ex(mdctx, data, &data_len) != 1) {
+      fprintf(stderr, "Failed to finalize digest operation\n");
+      goto test_out;
+    }
     if(verbose) {
       fprintf(stderr, "Test data hashes as: ");
       dump_data(data, data_len, stderr, true, format_arg_hex);
     }
+    EVP_MD_CTX_destroy(mdctx);
   }
 
   {
-    unsigned char signature[1024];
-    unsigned char encoded[1024];
+    unsigned char signature[1024] = {0};
+    unsigned char encoded[1024] = {0};
     unsigned char *ptr = data;
     unsigned int enc_len;
     size_t sig_len = sizeof(signature);
@@ -1609,6 +1761,9 @@ static bool test_signature(ykpiv_state *state, enum enum_slot slot,
     }
   }
 test_out:
+  if(pubkey) {
+    EVP_PKEY_free(pubkey);
+  }
   if(x509) {
     X509_free(x509);
   }
@@ -1622,7 +1777,7 @@ static bool test_decipher(ykpiv_state *state, enum enum_slot slot,
     const char *input_file_name, enum enum_key_format cert_format, int verbose) {
   bool ret = false;
   X509 *x509 = NULL;
-  EVP_PKEY *pubkey;
+  EVP_PKEY *pubkey = NULL;
   EC_KEY *tmpkey = NULL;
   FILE *input_file = open_file(input_file_name, key_file_mode(cert_format, false));
 
@@ -1663,9 +1818,9 @@ static bool test_decipher(ykpiv_state *state, enum enum_slot slot,
     }
     key = get_slot_hex(slot);
     if(YKPIV_IS_RSA(algorithm)) {
-      unsigned char secret[32];
-      unsigned char secret2[32];
-      unsigned char data[256];
+      unsigned char secret[32] = {0};
+      unsigned char secret2[32] = {0};
+      unsigned char data[256] = {0};
       int len;
       size_t len2 = sizeof(data);
       RSA *rsa = EVP_PKEY_get1_RSA(pubkey);
@@ -1704,9 +1859,9 @@ static bool test_decipher(ykpiv_state *state, enum enum_slot slot,
         fprintf(stderr, "Failed unwrapping PKCS1 envelope.\n");
       }
     } else if(YKPIV_IS_EC(algorithm)) {
-      unsigned char secret[48];
-      unsigned char secret2[48];
-      unsigned char public_key[97];
+      unsigned char secret[48] = {0};
+      unsigned char secret2[48] = {0};
+      unsigned char public_key[97] = {0};
       unsigned char *ptr = public_key;
       size_t len = sizeof(secret);
       EC_KEY *ec = EVP_PKEY_get1_EC_KEY(pubkey);
@@ -1722,10 +1877,19 @@ static bool test_decipher(ykpiv_state *state, enum enum_slot slot,
       }
 
       tmpkey = EC_KEY_new_by_curve_name(nid);
-      EC_KEY_generate_key(tmpkey);
-      ECDH_compute_key(secret, len, EC_KEY_get0_public_key(ec), tmpkey, NULL);
+      if(EC_KEY_generate_key(tmpkey) != 1) {
+        fprintf(stderr, "Failed to generate EC key\n");
+        goto decipher_out;
+      }
+      if(ECDH_compute_key(secret, len, EC_KEY_get0_public_key(ec), tmpkey, NULL) == -1) {
+        fprintf(stderr, "Failed to compute ECDH key\n");
+        goto decipher_out;
+      }
 
-      i2o_ECPublicKey(tmpkey, &ptr);
+      if(i2o_ECPublicKey(tmpkey, &ptr) < 0) {
+        fprintf(stderr, "Failed to parse EC public key\n");
+        goto decipher_out;
+      }
       if(ykpiv_decipher_data(state, public_key, (key_len * 2) + 1, secret2, &len, algorithm, key) != YKPIV_OK) {
         fprintf(stderr, "Failed ECDH exchange!\n");
         goto decipher_out;
@@ -1749,6 +1913,9 @@ decipher_out:
   if(tmpkey) {
     EC_KEY_free(tmpkey);
   }
+  if(pubkey) {
+    EVP_PKEY_free(pubkey);
+  }
   if(x509) {
     X509_free(x509);
   }
@@ -1759,7 +1926,7 @@ decipher_out:
 }
 
 static bool list_readers(ykpiv_state *state) {
-  char readers[2048];
+  char readers[2048] = {0};
   char *reader_ptr;
   size_t len = sizeof(readers);
   ykpiv_rc rc = ykpiv_list_readers(state, readers, &len);
@@ -1775,7 +1942,7 @@ static bool list_readers(ykpiv_state *state) {
 
 static bool attest(ykpiv_state *state, enum enum_slot slot,
     enum enum_key_format key_format, const char *output_file_name) {
-  unsigned char data[2048];
+  unsigned char data[2048] = {0};
   size_t len = sizeof(data);
   bool ret = false;
   X509 *x509 = NULL;
@@ -1804,7 +1971,9 @@ static bool attest(ykpiv_state *state, enum enum_slot slot,
       fprintf(stderr, "Failed parsing x509 information.\n");
       goto attest_out;
     }
-    PEM_write_X509(output_file, x509);
+    if(PEM_write_X509(output_file, x509) != 1){
+      fprintf(stderr, "Failed writing x509 information\n");
+    }
   } else {
     fwrite(data, len, 1, output_file);
   }
@@ -1824,7 +1993,7 @@ static bool write_object(ykpiv_state *state, int id,
     const char *input_file_name, int verbosity, enum enum_format format) {
   bool ret = false;
   FILE *input_file = NULL;
-  unsigned char data[YKPIV_OBJ_MAX_SIZE];
+  unsigned char data[YKPIV_OBJ_MAX_SIZE] = {0};
   size_t len = sizeof(data);
   ykpiv_rc res;
 
@@ -1863,7 +2032,7 @@ write_out:
 static bool read_object(ykpiv_state *state, int id, const char *output_file_name,
     enum enum_format format) {
   FILE *output_file = NULL;
-  unsigned char data[YKPIV_OBJ_MAX_SIZE];
+  unsigned char data[YKPIV_OBJ_MAX_SIZE] = {0};
   unsigned long len = sizeof(data);
   bool ret = false;
 
@@ -1895,7 +2064,7 @@ int main(int argc, char *argv[]) {
   unsigned int i;
   int ret = EXIT_SUCCESS;
   bool authed = false;
-  char pwbuf[128];
+  char pwbuf[128] = {0};
   char *password;
 
   if(cmdline_parser(argc, argv, &args_info) != 0) {
@@ -1913,6 +2082,7 @@ int main(int argc, char *argv[]) {
         if(!args_info.subject_arg) {
           fprintf(stderr, "The '%s' action needs a subject (-S) to operate on.\n",
               cmdline_parser_action_values[action]);
+          cmdline_parser_free(&args_info);
           return EXIT_FAILURE;
         }
         /* fall through */
@@ -1927,6 +2097,7 @@ int main(int argc, char *argv[]) {
         if(args_info.slot_arg == slot__NULL) {
           fprintf(stderr, "The '%s' action needs a slot (-s) to operate on.\n",
               cmdline_parser_action_values[action]);
+          cmdline_parser_free(&args_info);
           return EXIT_FAILURE;
         }
         break;
@@ -1934,6 +2105,7 @@ int main(int argc, char *argv[]) {
         if(!args_info.pin_retries_given || !args_info.puk_retries_given) {
           fprintf(stderr, "The '%s' action needs both --pin-retries and --puk-retries arguments.\n",
               cmdline_parser_action_values[action]);
+          cmdline_parser_free(&args_info);
           return EXIT_FAILURE;
         }
         break;
@@ -1942,6 +2114,7 @@ int main(int argc, char *argv[]) {
         if(!args_info.id_given) {
           fprintf(stderr, "The '%s' action needs the --id argument.\n",
               cmdline_parser_action_values[action]);
+          cmdline_parser_free(&args_info);
           return EXIT_FAILURE;
         }
         break;
@@ -1964,11 +2137,14 @@ int main(int argc, char *argv[]) {
 
   if(ykpiv_init(&state, verbosity) != YKPIV_OK) {
     fprintf(stderr, "Failed initializing library.\n");
+    cmdline_parser_free(&args_info);
     return EXIT_FAILURE;
   }
 
   if(ykpiv_connect(state, args_info.reader_arg) != YKPIV_OK) {
-    fprintf(stderr, "Failed to connect to yubikey.\nTry removing and reconnecting the device.");
+    fprintf(stderr, "Failed to connect to yubikey.\nTry removing and reconnecting the device.\n");
+    ykpiv_done(state);
+    cmdline_parser_free(&args_info);
     return EXIT_FAILURE;
   }
 
@@ -1983,7 +2159,9 @@ int main(int argc, char *argv[]) {
           }
           if(!read_pw("Password", pwbuf, sizeof(pwbuf), false, args_info.stdin_input_flag)) {
             fprintf(stderr, "Failed to get password.\n");
-            return false;
+            ykpiv_done(state);
+            cmdline_parser_free(&args_info);
+            return EXIT_FAILURE;
           }
           password = pwbuf;
         }
@@ -1996,9 +2174,9 @@ int main(int argc, char *argv[]) {
       case action_arg_deleteMINUS_certificate:
       case action_arg_writeMINUS_object:
         if(!authed) {
-          unsigned char key[KEY_LEN];
+          unsigned char key[KEY_LEN] = {0};
           size_t key_len = sizeof(key);
-          char keybuf[KEY_LEN*2+2]; /* one extra byte for potential \n */
+          char keybuf[KEY_LEN*2+2] = {0}; /* one extra byte for potential \n */
           char *key_ptr = args_info.key_arg;
           if(verbosity) {
             fprintf(stderr, "Authenticating since action '%s' needs that.\n", cmdline_parser_action_values[action]);
@@ -2006,17 +2184,23 @@ int main(int argc, char *argv[]) {
           if(args_info.key_given && args_info.key_orig == NULL) {
             if(!read_pw("management key", keybuf, sizeof(keybuf), false, args_info.stdin_input_flag)) {
               fprintf(stderr, "Failed to read management key from stdin,\n");
+              ykpiv_done(state);
+              cmdline_parser_free(&args_info);
               return EXIT_FAILURE;
             }
             key_ptr = keybuf;
           }
           if(ykpiv_hex_decode(key_ptr, strlen(key_ptr), key, &key_len) != YKPIV_OK) {
             fprintf(stderr, "Failed decoding key!\n");
+            ykpiv_done(state);
+            cmdline_parser_free(&args_info);
             return EXIT_FAILURE;
           }
 
           if(ykpiv_authenticate(state, key) != YKPIV_OK) {
             fprintf(stderr, "Failed authentication with the application.\n");
+            ykpiv_done(state);
+            cmdline_parser_free(&args_info);
             return EXIT_FAILURE;
           }
           if(verbosity) {
@@ -2054,7 +2238,7 @@ int main(int argc, char *argv[]) {
 
 
   /* openssl setup.. */
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#if (OPENSSL_VERSION_NUMBER < 0x10100000L)
   OpenSSL_add_all_algorithms();
 #endif
 
@@ -2089,7 +2273,7 @@ int main(int argc, char *argv[]) {
           new_mgm_key = new_keybuf;
         }
         if(strlen(new_mgm_key) == (KEY_LEN * 2)){
-          unsigned char new_key[KEY_LEN];
+          unsigned char new_key[KEY_LEN] = {0};
           size_t new_key_len = sizeof(new_key);
           if(ykpiv_hex_decode(new_mgm_key, strlen(new_mgm_key), new_key, &new_key_len) != YKPIV_OK) {
             fprintf(stderr, "Failed decoding new key!\n");
@@ -2105,7 +2289,7 @@ int main(int argc, char *argv[]) {
             fprintf(stderr, "Successfully set new management key.\n");
           }
         } else {
-          fprintf(stderr, "The new management key has to be exactly %d hexidecimal characters.\n", KEY_LEN * 2);
+          fprintf(stderr, "The new management key has to be exactly %d hexadecimal characters.\n", KEY_LEN * 2);
           ret = EXIT_FAILURE;
         }
         break;
@@ -2153,7 +2337,7 @@ int main(int argc, char *argv[]) {
       case action_arg_requestMINUS_certificate:
         if(request_certificate(state, args_info.key_format_arg, args_info.input_arg,
               args_info.slot_arg, args_info.subject_arg, args_info.hash_arg,
-              args_info.output_arg) == false) {
+              args_info.output_arg, args_info.attestation_flag) == false) {
           ret = EXIT_FAILURE;
         } else {
           fprintf(stderr, "Successfully generated a certificate request.\n");
@@ -2165,7 +2349,10 @@ int main(int argc, char *argv[]) {
 
         if(!pin) {
           if (!read_pw("PIN", pinbuf, sizeof(pinbuf), false, args_info.stdin_input_flag)) {
-            return false;
+            fprintf(stderr, "Failed to get PIN.\n");
+            ykpiv_done(state);
+            cmdline_parser_free(&args_info);
+            return EXIT_FAILURE;
           }
           pin = pinbuf;
         }
@@ -2188,13 +2375,19 @@ int main(int argc, char *argv[]) {
 
         if(!pin) {
           if (!read_pw(name, pinbuf, sizeof(pinbuf), false, args_info.stdin_input_flag)) {
-            return false;
+            fprintf(stderr, "Failed to get %s.\n", name);
+            ykpiv_done(state);
+            cmdline_parser_free(&args_info);
+            return EXIT_FAILURE;
           }
           pin = pinbuf;
         }
         if(!new_pin) {
           if (!read_pw(new_name, new_pinbuf, sizeof(new_pinbuf), true, args_info.stdin_input_flag)) {
-            return false;
+            fprintf(stderr, "Failed to get %s.\n", new_name);
+            ykpiv_done(state);
+            cmdline_parser_free(&args_info);
+            return EXIT_FAILURE;
           }
           new_pin = new_pinbuf;
         }
@@ -2297,8 +2490,9 @@ int main(int argc, char *argv[]) {
   }
 
   ykpiv_done(state);
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#if (OPENSSL_VERSION_NUMBER < 0x10100000L)
   EVP_cleanup();
 #endif
+  cmdline_parser_free(&args_info);
   return ret;
 }
