@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2015-2016 Yubico AB
+ * Copyright (c) 2015-2020 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -31,8 +31,8 @@
 #include <string.h>
 #include "mechanisms.h"
 #include "objects.h"
-#include "../tool/openssl-compat.h" // TODO: share this better?
-#include "../tool/util.h" // TODO: share this better?
+#include "../common/openssl-compat.h"
+#include "../common/util.h"
 #include "openssl_utils.h"
 #include "utils.h"
 #include "debug.h"
@@ -190,7 +190,7 @@ CK_RV sign_mechanism_init(ykcs11_session_t *session, ykcs11_pkey_t *key, CK_MECH
         DBG("Mechanism %lu requires PSS parameters", session->op_info.mechanism);
         return CKR_MECHANISM_PARAM_INVALID;
       }
-      CK_RSA_PKCS_PSS_PARAMS_PTR pss = mech->pParameter;
+      CK_RSA_PKCS_PSS_PARAMS *pss = mech->pParameter;
       session->op_info.op.sign.padding = RSA_PKCS1_PSS_PADDING;
       session->op_info.op.sign.pss_md = EVP_MD_by_mechanism(pss->hashAlg);
       session->op_info.op.sign.mgf1_md = EVP_MD_by_mechanism(pss->mgf);
@@ -258,7 +258,7 @@ CK_RV sign_mechanism_final(ykcs11_session_t *session, CK_BYTE_PTR sig, CK_ULONG_
   }
 
   CK_ULONG padlen = session->op_info.out_len;
-  CK_BYTE buf[1024];
+  CK_BYTE buf[1024] = {0};
 
   // Apply padding
   switch(session->op_info.op.sign.padding) {
@@ -290,7 +290,7 @@ CK_RV sign_mechanism_final(ykcs11_session_t *session, CK_BYTE_PTR sig, CK_ULONG_
   }
 
   // Sign with PIV
-  unsigned char sigbuf[256];
+  unsigned char sigbuf[256] = {0};
   size_t siglen = sizeof(sigbuf);
   ykpiv_rc rcc = ykpiv_sign_data(session->slot->piv_state, session->op_info.buf, session->op_info.buf_len, sigbuf, &siglen, session->op_info.op.sign.algorithm, session->op_info.op.sign.piv_key);
   if(rcc == YKPIV_OK) {
@@ -395,7 +395,7 @@ CK_RV verify_mechanism_init(ykcs11_session_t *session, ykcs11_pkey_t *key, CK_ME
   }
 
   ykcs11_rsa_t *rsa = EVP_PKEY_get0_RSA(key);
-  CK_RSA_PKCS_PSS_PARAMS_PTR pss = NULL;
+  CK_RSA_PKCS_PSS_PARAMS *pss = NULL;
 
   switch (session->op_info.mechanism) {
     case CKM_RSA_X_509:
@@ -485,9 +485,18 @@ CK_RV verify_mechanism_init(ykcs11_session_t *session, ykcs11_pkey_t *key, CK_ME
         DBG("Mechanism %lu requires PSS parameters to specify hashAlg %s", session->op_info.mechanism, EVP_MD_name(md));
         return CKR_ARGUMENTS_BAD;
       }
-      EVP_PKEY_CTX_set_signature_md(session->op_info.op.verify.pkey_ctx, EVP_MD_by_mechanism(pss->hashAlg));
-      EVP_PKEY_CTX_set_rsa_mgf1_md(session->op_info.op.verify.pkey_ctx, EVP_MD_by_mechanism(pss->mgf));
-      EVP_PKEY_CTX_set_rsa_pss_saltlen(session->op_info.op.verify.pkey_ctx, pss->sLen);
+      if(EVP_PKEY_CTX_set_signature_md(session->op_info.op.verify.pkey_ctx, EVP_MD_by_mechanism(pss->hashAlg)) <= 0) {
+        DBG("Failed to set signature");
+        return CKR_FUNCTION_FAILED;
+      }
+      if(EVP_PKEY_CTX_set_rsa_mgf1_md(session->op_info.op.verify.pkey_ctx, EVP_MD_by_mechanism(pss->mgf)) <= 0) {
+        DBG("Failed to set PSS MGF type parameter");
+        return CKR_FUNCTION_FAILED;
+      }
+      if(EVP_PKEY_CTX_set_rsa_pss_saltlen(session->op_info.op.verify.pkey_ctx, pss->sLen) <= 0) {
+        DBG("Failed to set PSS salt length");
+        return CKR_FUNCTION_FAILED;
+      }
     }
   }
 
@@ -501,7 +510,7 @@ CK_RV verify_mechanism_final(ykcs11_session_t *session, CK_BYTE_PTR sig, CK_ULON
 
   int rc;
 
-  CK_BYTE der[1024];
+  CK_BYTE der[1024] = {0};
   if(!session->op_info.op.verify.padding) {
     if(sig_len > sizeof(der)) {
       DBG("do_apply_DER_encoding_to_ECSIG failed because signature was too large (%lu)", sig_len);
@@ -567,21 +576,26 @@ CK_RV check_pubkey_template(gen_info_t *gen, CK_MECHANISM_PTR mechanism, CK_ATTR
   for (CK_ULONG i = 0; i < n; i++) {
     switch (templ[i].type) {
     case CKA_CLASS:
-      if (*((CK_ULONG_PTR) templ[i].pValue) != CKO_PUBLIC_KEY)
+      if (*((CK_ULONG_PTR) templ[i].pValue) != CKO_PUBLIC_KEY) {
+        DBG("Bad CKA_CLASS");
         return CKR_TEMPLATE_INCONSISTENT;
-
+      }
       break;
 
     case CKA_KEY_TYPE:
       if ((rsa == CK_TRUE  && (*((CK_KEY_TYPE *)templ[i].pValue)) != CKK_RSA) ||
-          (rsa == CK_FALSE && (*((CK_KEY_TYPE *)templ[i].pValue)) != CKK_ECDSA))
+          (rsa == CK_FALSE && (*((CK_KEY_TYPE *)templ[i].pValue)) != CKK_ECDSA)) {
+        DBG("Bad CKA_KEY_TYPE");
         return CKR_TEMPLATE_INCONSISTENT;
+      }
 
       break;
 
     case CKA_PUBLIC_EXPONENT:
-      if (rsa == CK_FALSE)
-        return CKR_ATTRIBUTE_VALUE_INVALID;
+      if (rsa == CK_FALSE) {
+        DBG("Non-RSA key can't have CKA_PUBLIC_EXPONENT");
+        return CKR_TEMPLATE_INCONSISTENT;
+      }
 
       // Only support F4
       if (templ[i].ulValueLen != 3 || memcmp((CK_BYTE_PTR)templ[i].pValue, F4, 3) != 0) {
@@ -592,9 +606,10 @@ CK_RV check_pubkey_template(gen_info_t *gen, CK_MECHANISM_PTR mechanism, CK_ATTR
       break;
 
     case CKA_MODULUS_BITS:
-      if (rsa == CK_FALSE)
-        return CKR_ATTRIBUTE_VALUE_INVALID;
-
+      if (rsa == CK_FALSE) {
+        DBG("Non-RSA key can't have CKA_MODULUS_BITS");
+        return CKR_TEMPLATE_INCONSISTENT;
+      }
       switch(*(CK_ULONG_PTR)templ[i].pValue) {
         case 1024:
           gen->algorithm = YKPIV_ALGO_RSA1024;
@@ -609,22 +624,30 @@ CK_RV check_pubkey_template(gen_info_t *gen, CK_MECHANISM_PTR mechanism, CK_ATTR
       break;
 
     case CKA_EC_PARAMS:
+      if (rsa == CK_TRUE) {
+        DBG("RSA key can't have CKA_EC_PARAMS");
+        return CKR_TEMPLATE_INCONSISTENT;
+      }
       // Support PRIME256V1 and SECP384R1
       if (templ[i].ulValueLen == 10 && memcmp((CK_BYTE_PTR)templ[i].pValue, PRIME256V1, 10) == 0)
         gen->algorithm = YKPIV_ALGO_ECCP256;
       else if(templ[i].ulValueLen == 7 && memcmp((CK_BYTE_PTR)templ[i].pValue, SECP384R1, 7) == 0)
         gen->algorithm = YKPIV_ALGO_ECCP384;
-      else
-        return CKR_FUNCTION_FAILED;
+      else {
+        DBG("Bad CKA_EC_PARAMS");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
       break;
 
     case CKA_ID:
-      if (find_pubk_object(*((CK_BYTE_PTR)templ[i].pValue)) == PIV_INVALID_OBJ)
+      if (find_pubk_object(*((CK_BYTE_PTR)templ[i].pValue)) == PIV_INVALID_OBJ) {
+        DBG("Bad CKA_ID");
         return CKR_ATTRIBUTE_VALUE_INVALID;
-
+      }
       gen->key_id = *((CK_BYTE_PTR)templ[i].pValue);
       break;
 
+    case CKA_SENSITIVE:
     case CKA_TOKEN:
     case CKA_ENCRYPT:
     case CKA_VERIFY:
@@ -632,6 +655,7 @@ CK_RV check_pubkey_template(gen_info_t *gen, CK_MECHANISM_PTR mechanism, CK_ATTR
     case CKA_DERIVE:
     case CKA_PRIVATE:
     case CKA_LABEL:
+    case CKA_SUBJECT:
       // Ignore these attributes for now
       break;
 
@@ -652,28 +676,32 @@ CK_RV check_pvtkey_template(gen_info_t *gen, CK_MECHANISM_PTR mechanism, CK_ATTR
   for (CK_ULONG i = 0; i < n; i++) {
     switch (templ[i].type) {
     case CKA_CLASS:
-      if (*((CK_ULONG_PTR)templ[i].pValue) != CKO_PRIVATE_KEY)
+      if (*((CK_ULONG_PTR)templ[i].pValue) != CKO_PRIVATE_KEY) {
+        DBG("Bad CKA_CLASS");
         return CKR_TEMPLATE_INCONSISTENT;
-
+      }
       break;
 
     case CKA_KEY_TYPE:
       if ((rsa == CK_TRUE  && (*((CK_KEY_TYPE *)templ[i].pValue)) != CKK_RSA) ||
-          (rsa == CK_FALSE && (*((CK_KEY_TYPE *)templ[i].pValue)) != CKK_ECDSA))
+          (rsa == CK_FALSE && (*((CK_KEY_TYPE *)templ[i].pValue)) != CKK_ECDSA)) {
+        DBG("Bad CKA_KEY_TYPE");
         return CKR_TEMPLATE_INCONSISTENT;
-
+      }
       break;
 
     case CKA_ID:
-      if (find_pvtk_object(*((CK_BYTE_PTR)templ[i].pValue)) == PIV_INVALID_OBJ)
+      if (find_pvtk_object(*((CK_BYTE_PTR)templ[i].pValue)) == PIV_INVALID_OBJ) {
+        DBG("Bad CKA_ID");
         return CKR_ATTRIBUTE_VALUE_INVALID;
-
+      }
       // Check if ID was already specified in the public key template
       // In that case it has to match
       if (gen->key_id != 0 &&
-          gen->key_id != *((CK_BYTE_PTR)templ[i].pValue))
+          gen->key_id != *((CK_BYTE_PTR)templ[i].pValue)) {
+        DBG("Inconsistent CKA_ID");
         return CKR_TEMPLATE_INCONSISTENT;
-
+      }
       gen->key_id = *((CK_BYTE_PTR)templ[i].pValue);
       break;
 
@@ -685,6 +713,7 @@ CK_RV check_pvtkey_template(gen_info_t *gen, CK_MECHANISM_PTR mechanism, CK_ATTR
     case CKA_TOKEN:
     case CKA_DERIVE:
     case CKA_LABEL:
+    case CKA_SUBJECT:
       // Ignore these attributes for now
       break;
 
@@ -698,6 +727,44 @@ CK_RV check_pvtkey_template(gen_info_t *gen, CK_MECHANISM_PTR mechanism, CK_ATTR
 
 }
 
+CK_RV validate_derive_key_attribute(CK_ATTRIBUTE_TYPE type, void *value) {
+  switch (type) {
+    case CKA_TOKEN:
+      if (*((CK_BBOOL *) value) != CK_FALSE) {
+        DBG("Derived key can only be a session object");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_CLASS:
+      if (*((CK_ULONG_PTR) value) != CKO_SECRET_KEY) {
+        DBG("Derived key class is unsupported");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_KEY_TYPE:
+      if (*((CK_ULONG_PTR) value) != CKK_GENERIC_SECRET) {
+        DBG("Derived key type is unsupported");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_EXTRACTABLE:
+      if (*((CK_BBOOL *) value) != CK_TRUE) {
+        DBG("The derived key must be extractable");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    default:
+      DBG("ECDH key derive template contains the ignored attribute: %lx", type);
+      break;
+  }
+
+  return CKR_OK;
+}
+
 CK_RV digest_mechanism_init(ykcs11_session_t *session, CK_MECHANISM_PTR mech) {
 
   session->op_info.mechanism = mech->mechanism;
@@ -806,7 +873,7 @@ CK_RV decrypt_mechanism_init(ykcs11_session_t *session, ykcs11_pkey_t *key, CK_M
     if(mech->pParameter == NULL || mech->ulParameterLen != sizeof(CK_RSA_PKCS_OAEP_PARAMS)) {
         return CKR_MECHANISM_PARAM_INVALID;
     }
-    CK_RSA_PKCS_OAEP_PARAMS_PTR oaep = mech->pParameter;
+    CK_RSA_PKCS_OAEP_PARAMS *oaep = mech->pParameter;
     DBG("OAEP params : hashAlg 0x%lx mgf 0x%lx source 0x%lx pSourceData %p ulSourceDataLen %lu", oaep->hashAlg, oaep->mgf, oaep->source, oaep->pSourceData, oaep->ulSourceDataLen);
     session->op_info.op.encrypt.oaep_md = EVP_MD_by_mechanism(oaep->hashAlg);
     session->op_info.op.encrypt.mgf1_md = EVP_MD_by_mechanism(oaep->mgf);
@@ -833,7 +900,7 @@ CK_RV decrypt_mechanism_init(ykcs11_session_t *session, ykcs11_pkey_t *key, CK_M
 
 CK_RV decrypt_mechanism_final(ykcs11_session_t *session, CK_BYTE_PTR data, CK_ULONG_PTR data_len, CK_ULONG key_len) {
   ykpiv_rc piv_rv;
-  CK_BYTE  dec[1024];
+  CK_BYTE  dec[1024] = {0};
   size_t   dec_len = sizeof(dec);
   int      cb_len;
 
