@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2015-2016 Yubico AB
+ * Copyright (c) 2015-2016,2019-2020 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -30,7 +30,7 @@
 
 #include "obj_types.h"
 #include "objects.h"
-#include <ykpiv.h>
+#include "ykpiv.h"
 #include <string.h>
 #include <stdlib.h>
 #include "openssl_utils.h"
@@ -46,6 +46,7 @@ static CK_RV get_coa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templa
 static CK_RV get_proa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR template);
 static CK_RV get_puoa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR template);
 static CK_RV get_atst(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR template);
+static CK_RV get_skoa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR template);
 
 //TODO: this is mostly a snippet from OpenSC how to give credit?     Less and less so now
 /* Must be in order, and one per enumerated PIV_OBJ */
@@ -190,7 +191,8 @@ static piv_obj_t piv_objects[] = {
   {PIV_PUBK_OBJ_RETIRED18, "Public key for Retired Key 18", get_puoa, 22},
   {PIV_PUBK_OBJ_RETIRED19, "Public key for Retired Key 19", get_puoa, 23},
   {PIV_PUBK_OBJ_RETIRED20, "Public key for Retired Key 20", get_puoa, 24},
-  {PIV_PUBK_OBJ_ATTESTATION, "Public key for PIV Attestation", get_puoa, 25}
+  {PIV_PUBK_OBJ_ATTESTATION, "Public key for PIV Attestation", get_puoa, 25},
+  {PIV_SECRET_OBJ, "Generic secret key", get_skoa, 0}
 };
 
 static piv_data_obj_t data_objects[] = {
@@ -385,7 +387,7 @@ static CK_RV get_doa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templa
 /* Get certificate object attribute */
 static CK_RV _get_coa(ykcs11_x509_t **certs, piv_obj_id_t obj, CK_ATTRIBUTE_PTR template, CK_BBOOL token) {
   CK_BYTE_PTR data;
-  CK_BYTE     b_tmp[YKPIV_OBJ_MAX_SIZE]; // Max cert value for ykpiv
+  CK_BYTE     b_tmp[YKPIV_OBJ_MAX_SIZE] = {0}; // Max cert value for ykpiv
   CK_ULONG    ul_tmp;
   CK_ULONG    len = 0;
   CK_RV       rv;
@@ -512,7 +514,7 @@ static CK_RV get_atst(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
 /* Get private key object attribute */
 static CK_RV get_proa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR template) {
   CK_BYTE_PTR data;
-  CK_BYTE     b_tmp[1024];
+  CK_BYTE     b_tmp[1024] = {0};
   CK_ULONG    ul_tmp;
   CK_ULONG    len = 0;
   CK_RV       rv;
@@ -745,7 +747,7 @@ static CK_RV get_proa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
 /* Get public key object attribute */
 static CK_RV get_puoa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR template) {
   CK_BYTE_PTR data;
-  CK_BYTE     b_tmp[1024];
+  CK_BYTE     b_tmp[1024] = {0};
   CK_ULONG    ul_tmp;
   CK_ULONG    len = 0;
   DBG("For public key object %u, get ", obj);
@@ -946,6 +948,147 @@ static CK_RV get_puoa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
   return CKR_OK;
 }
 
+/* Get secret key object attribute */
+static CK_RV get_skoa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR template) {
+  CK_BYTE_PTR data;
+  CK_BYTE     tmp;
+  CK_ULONG    ul_tmp;
+  CK_ULONG    len = 0;
+  DBG("For secret key object %u, get ", obj);
+
+  switch (template->type) {
+  case CKA_CLASS:
+    DBG("CLASS");
+    len = sizeof(CK_ULONG);
+    ul_tmp = CKO_SECRET_KEY;
+    data = (CK_BYTE_PTR)&ul_tmp;
+    break;
+
+  case CKA_KEY_TYPE:
+    DBG("KEY_TYPE");
+    len = sizeof(CK_ULONG);
+    ul_tmp = CKK_GENERIC_SECRET;
+    data = (CK_BYTE_PTR)&ul_tmp;
+    break;
+
+  case CKA_TOKEN:
+    DBG("TOKEN");
+    len = sizeof(CK_BBOOL);
+    tmp = CK_FALSE;
+    data = &tmp;
+    break;
+
+  case CKA_PRIVATE:
+    DBG("PRIVATE");
+    len = sizeof(CK_BBOOL);
+    tmp = CK_TRUE;
+    data = &tmp;
+    break;
+
+  case CKA_LOCAL:
+    DBG("LOCAL");
+    len = sizeof(CK_BBOOL);
+    tmp = CK_FALSE;
+    data = &tmp;
+    break;
+
+  case CKA_SENSITIVE:
+    DBG("SENSITIVE");
+    len = sizeof(CK_BBOOL);
+    tmp = CK_FALSE;
+    data = &tmp;
+    break;
+
+  case CKA_ALWAYS_SENSITIVE:
+    DBG("ALWAYS_SENSITIVE");
+    len = sizeof(CK_BBOOL);
+    tmp = CK_FALSE;
+    data = &tmp;
+    break;
+
+  case CKA_EXTRACTABLE:
+    DBG("EXTRACTABLE");
+    len = sizeof(CK_BBOOL);
+    tmp = CK_TRUE;
+    data = &tmp;
+    break;
+
+  case CKA_NEVER_EXTRACTABLE:
+    DBG("NEVER_EXTRACTABLE");
+    len = sizeof(CK_BBOOL);
+    tmp = CK_FALSE;
+    data = &tmp;
+    break;
+
+  case CKA_ENCRYPT:
+    DBG("ENCRYPT");
+    len = sizeof(CK_BBOOL);
+    tmp = CK_FALSE;
+    data = &tmp;
+    break;
+
+  case CKA_DECRYPT:
+    DBG("DECRYPT");
+    len = sizeof(CK_BBOOL);
+    tmp = CK_FALSE;
+    data = &tmp;
+    break;
+
+  case CKA_DERIVE:
+    DBG("DERIVE");
+    len = sizeof(CK_BBOOL);
+    tmp = CK_FALSE;
+    data = &tmp;
+    break;
+
+  case CKA_MODIFIABLE:
+    DBG("MODIFIABLE");
+    len = sizeof(CK_BBOOL);
+    tmp = CK_FALSE;
+    data = &tmp;
+    break;
+
+  case CKA_LABEL:
+    DBG("LABEL");
+    len = strlen(piv_objects[obj].label);
+    data = (CK_BYTE_PTR) piv_objects[obj].label;
+    break;
+
+  case CKA_ID:
+    DBG("ID");
+    len = sizeof(CK_BYTE);
+    tmp = piv_objects[obj].sub_id;
+    data = &tmp;
+    break;
+
+  case CKA_VALUE:
+    DBG("VALUE");
+    len = s->data[piv_objects[obj].sub_id].len;
+    data = s->data[piv_objects[obj].sub_id].data;
+    break;
+
+  default:
+    DBG("UNKNOWN ATTRIBUTE %lx (%lu)", template[0].type, template[0].type);
+    template->ulValueLen = CK_UNAVAILABLE_INFORMATION;
+    return CKR_ATTRIBUTE_TYPE_INVALID;
+  }
+
+  /* Just get the length */
+  if (template->pValue == NULL) {
+    template->ulValueLen = len;
+    return CKR_OK;
+  }
+
+  /* Actually get the attribute */
+  if (template->ulValueLen < len)
+    return CKR_BUFFER_TOO_SMALL;
+
+  template->ulValueLen = len;
+  memcpy(template->pValue, data, len);
+
+  return CKR_OK;
+}
+
 CK_ULONG piv_2_ykpiv(piv_obj_id_t id) {
   switch(id) {
   case PIV_DATA_OBJ_CCC:
@@ -975,9 +1118,14 @@ CK_ULONG piv_2_ykpiv(piv_obj_id_t id) {
   case PIV_DATA_OBJ_IRIS_IMAGE:
     return YKPIV_OBJ_IRIS;
 
-  //case PIV_DATA_OBJ_BITGT:
-  //case PIV_DATA_OBJ_SM_SIGNER:
-  //case PIV_DATA_OBJ_PC_REF_DATA:
+  case PIV_DATA_OBJ_BITGT:
+    return YKPIV_OBJ_BITGT;
+
+  case PIV_DATA_OBJ_SM_SIGNER:
+    return YKPIV_OBJ_SM_SIGNER;
+  
+  case PIV_DATA_OBJ_PC_REF_DATA:
+    return YKPIV_OBJ_PC_REF_DATA;
 
   case PIV_DATA_OBJ_X509_PIV_AUTH:
   case PIV_CERT_OBJ_X509_PIV_AUTH:
@@ -1173,6 +1321,13 @@ CK_BBOOL is_present(ykcs11_slot_t *s, piv_obj_id_t id) {
 
 CK_BBOOL add_object(ykcs11_slot_t *s, piv_obj_id_t id) {
   if(s->n_objects < sizeof(s->objects) / sizeof(s->objects[0])) {
+    // We can't use is_present here because the objects might not be sorted
+    for(CK_ULONG i = 0; i < s->n_objects; i++) {
+      if(id == s->objects[i]) {
+        DBG("Couldn't add object %u because it is already present", id);
+        return false;
+      }
+    }
     s->objects[s->n_objects++] = id;
     DBG("Added object %u, slot contains %lu objects", id, s->n_objects);
     return true;
@@ -1182,7 +1337,7 @@ CK_BBOOL add_object(ykcs11_slot_t *s, piv_obj_id_t id) {
 }
 
 CK_BBOOL is_local_key(ykcs11_slot_t *s, piv_obj_id_t id) {
-  return is_present(s, find_atst_object(piv_objects[id].sub_id));
+  return s->local[piv_objects[id].sub_id];
 }
 
 CK_RV get_attribute(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR template) {
@@ -1191,7 +1346,7 @@ CK_RV get_attribute(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templat
 
 CK_BBOOL attribute_match(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR attribute) {
 
-  CK_BYTE data[4096];
+  CK_BYTE data[4096] = {0};
   CK_ATTRIBUTE to_match = { attribute->type, data, sizeof(data) };
 
   if (get_attribute(s, obj, &to_match) != CKR_OK)
