@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2015-2016 Yubico AB
+ * Copyright (c) 2015-2016,2019 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -33,8 +33,6 @@
 
 #include "pkcs11.h"
 
-typedef CK_FLAGS * CK_FLAGS_PTR;
-
 // YUBICO specific attributes
 #define CKA_TOUCH_PIN_DEFAULT 0x00000000U
 #define CKA_TOUCH_ALWAYS      0x00000001U
@@ -43,31 +41,4 @@ typedef CK_FLAGS * CK_FLAGS_PTR;
 #define CKA_PIN_NEVER         0x00000008U
 #define CKA_TOUCH_NEVER       0x00000016U
 
-// Standard stuff that we use but is not in pkcs11.h
-
-#define CKG_MGF1_SHA1			  (0x1UL)
-#define CKG_MGF1_SHA256			(0x2UL)
-#define CKG_MGF1_SHA384			(0x3UL)
-#define CKG_MGF1_SHA512			(0x4UL)
-#define CKG_MGF1_SHA224			(0x5UL)
-
-#define CKZ_DATA_SPECIFIED  (0x1UL) // = CK_BYTE. The only supported option for CK_RSA_PKCS_OAEP_SOURCE_TYPE
-
-typedef unsigned long CK_RSA_PKCS_MGF_TYPE;
-typedef unsigned long CK_RSA_PKCS_OAEP_SOURCE_TYPE;
-
-typedef struct {
-  CK_MECHANISM_TYPE       hashAlg;
-  CK_RSA_PKCS_MGF_TYPE    mgf;
-  CK_ULONG                sLen;
-} CK_RSA_PKCS_PSS_PARAMS, *CK_RSA_PKCS_PSS_PARAMS_PTR;
-
-typedef struct {
-  CK_MECHANISM_TYPE             hashAlg;
-  CK_RSA_PKCS_MGF_TYPE          mgf;
-  CK_RSA_PKCS_OAEP_SOURCE_TYPE  source;
-  CK_VOID_PTR                   pSourceData;
-  CK_ULONG                      ulSourceDataLen;
-} CK_RSA_PKCS_OAEP_PARAMS, *CK_RSA_PKCS_OAEP_PARAMS_PTR;
-
 #endif
