@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2015-2016 Yubico AB
+ * Copyright (c) 2015-2017,2019-2020 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -28,9 +28,9 @@
  *
  */
 
-#include "../../tool/openssl-compat.h"
-#include <ykcs11.h>
-#include <ykcs11-version.h>
+#include "../../common/openssl-compat.h"
+#include "ykcs11.h"
+#include "ykcs11-config.h"
 
 #include <string.h>
 
@@ -42,10 +42,10 @@
 
 #include "ykcs11_tests_util.h"
 
-#pragma clang diagnostic push
-#pragma clang diagnostic ignored "-Wpointer-sign"
+//#pragma clang diagnostic push
+//#pragma clang diagnostic ignored "-Wpointer-sign"
 
-#ifdef __MINGW32__
+#ifdef _WIN32
 #define dprintf(fd, ...) fprintf(stdout, __VA_ARGS__)
 #endif
 
@@ -70,9 +70,9 @@ static void _asrt(const char *file, int line, CK_ULONG check, CK_ULONG expected,
 
 }
 
-static void get_functions(CK_FUNCTION_LIST_PTR_PTR funcs) {
+static void get_functions() {
 
-  if (C_GetFunctionList(funcs) != CKR_OK) {
+  if (C_GetFunctionList(&funcs) != CKR_OK) {
     fprintf(stderr, "Get function list failed\n");
     exit(EXIT_FAILURE);
   }
@@ -95,6 +95,7 @@ static void test_lib_info() {
   const CK_ULONG CRYPTOKI_VERSION_MIN = 40;
 
   CK_INFO info;
+  asrt(funcs->C_Initialize(NULL), CKR_OK, "INITIALIZE");
   asrt(funcs->C_GetInfo(&info), CKR_OK, "GET_INFO");
   asrt(strncmp(info.manufacturerID, MANUFACTURER_ID, strlen(MANUFACTURER_ID)), 0, "MANUFACTURER");
   asrt(info.cryptokiVersion.major, CRYPTOKI_VERSION_MAJ, "CK_MAJ");
@@ -102,6 +103,7 @@ static void test_lib_info() {
   asrt(info.libraryVersion.major, YKCS11_VERSION_MAJOR, "LIB_MAJ");
   asrt(info.libraryVersion.minor, ((YKCS11_VERSION_MINOR * 10) + YKCS11_VERSION_PATCH), "LIB_MIN");
   asrt(strncmp(info.libraryDescription, YKCS11_DESCRIPTION, strlen(YKCS11_DESCRIPTION)), 0, "LIB_DESC");
+  asrt(funcs->C_Finalize(NULL), CKR_OK, "FINALIZE");
   dprintf(0, "TEST END: test_lib_info()\n");
 }
 
@@ -194,6 +196,7 @@ static void test_mechanism_list_and_info() {
     CKM_ECDSA_SHA224,
     CKM_ECDSA_SHA256,
     CKM_ECDSA_SHA384,
+    CKM_ECDH1_DERIVE,
     CKM_SHA_1,
     CKM_SHA256,
     CKM_SHA384,
@@ -220,6 +223,7 @@ static void test_mechanism_list_and_info() {
     {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY},
     {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY},
     {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY},
+    {256, 384, CKF_HW | CKF_DERIVE},
     {0, 0, CKF_DIGEST},
     {0, 0, CKF_DIGEST},
     {0, 0, CKF_DIGEST},
@@ -247,7 +251,7 @@ static void test_mechanism_list_and_info() {
 static void test_session() {
   dprintf(0, "TEST START: test_session()\n");
 
-  CK_SESSION_HANDLE session1, session2;
+  CK_SESSION_HANDLE session1;
   CK_SESSION_INFO   info;
 
   init_connection();
@@ -339,7 +343,7 @@ static void test_max_multiple_sessions() {
   asrt(funcs->C_OpenSession(0, CKF_SERIAL_SESSION | CKF_RW_SESSION, NULL, NULL, &session), CKR_SESSION_COUNT, "MaxMultipleSession_OpenSession_TooMany");
 
   asrt(funcs->C_CloseAllSessions(0), CKR_OK, "MaxMultipleSessions_CloseAllSessions");
-  for(int i=1; i<=17; i++) {
+  for(i=1; i<=17; i++) {
     asrt(funcs->C_GetSessionInfo(i, &info), CKR_SESSION_HANDLE_INVALID, "MaxMultipleSessions_closedSessionsInfo");
   }
 
@@ -383,11 +387,11 @@ static void test_login() {
 static void test_login_order() {
   dprintf(0, "TEST START: test_login_order()\n");
   CK_BYTE     params[] = {0x06, 0x08, 0x2a, 0x86, 0x48, 0xce, 0x3d, 0x03, 0x01, 0x07};
-  CK_BYTE     data[32];
-  CK_BYTE     sig[128];
+  CK_BYTE     data[32] = {0};
+  CK_BYTE     sig[128] = {0};
   CK_BYTE     i;
   CK_ULONG    recv_len = sizeof(sig);
-  CK_OBJECT_HANDLE privkey[2], pubkey[2];
+  CK_OBJECT_HANDLE privkey[2]={0}, pubkey[2]={0};
   CK_SESSION_HANDLE session1, session2;
 
   CK_MECHANISM sign_mech = {CKM_ECDSA, NULL, 0};
@@ -412,8 +416,8 @@ static void test_login_order() {
   dprintf(0, "TEST END: test_login_order()\n");
 }
 
-static void test_generate_eccp(CK_BYTE* key_params, CK_ULONG key_params_len, CK_BYTE n_keys) {
-  CK_OBJECT_HANDLE obj_pvtkey[N_ALL_KEYS], obj_pubkey[N_ALL_KEYS];
+static void test_generate_eccp(CK_BYTE* key_params, CK_ULONG key_params_len, int curve, CK_BYTE n_keys) {
+  CK_OBJECT_HANDLE obj_pvtkey[N_ALL_KEYS]={0}, obj_pubkey[N_ALL_KEYS]={0};
   CK_SESSION_HANDLE session;
 
   init_connection();
@@ -421,6 +425,7 @@ static void test_generate_eccp(CK_BYTE* key_params, CK_ULONG key_params_len, CK_
 
   generate_ec_keys(funcs, session, n_keys, key_params, key_params_len, obj_pubkey, obj_pvtkey);
   test_ec_sign_simple(funcs, session, obj_pvtkey, n_keys, NULL, 0);
+  test_ec_ecdh_simple(funcs, session, obj_pvtkey, n_keys, curve);
 
   destroy_test_objects(funcs, session, obj_pvtkey, n_keys);
   asrt(funcs->C_CloseSession(session), CKR_OK, "CloseSession");
@@ -431,9 +436,9 @@ static void test_generate_eccp256() {
   dprintf(0, "TEST START: test_generate_eccp256()\n");
   CK_BYTE     params[] = {0x06, 0x08, 0x2a, 0x86, 0x48, 0xce, 0x3d, 0x03, 0x01, 0x07};
   if(is_neo) {
-    test_generate_eccp(params, sizeof(params), N_SELECTED_KEYS);
+    test_generate_eccp(params, sizeof(params), NID_X9_62_prime256v1, N_SELECTED_KEYS);
   } else {
-    test_generate_eccp(params, sizeof(params), N_ALL_KEYS);
+    test_generate_eccp(params, sizeof(params), NID_X9_62_prime256v1, N_ALL_KEYS);
   }
   dprintf(0, "TEST END: test_generate_eccp256()\n");
 }
@@ -442,14 +447,14 @@ static void test_generate_eccp384() {
   dprintf(0, "TEST START: test_generate_eccp384()\n");
   CK_BYTE     params[] = {0x06, 0x05, 0x2b, 0x81, 0x04, 0x00, 0x22};
   if(!is_neo) {
-    test_generate_eccp(params, sizeof(params), N_ALL_KEYS);
+    test_generate_eccp(params, sizeof(params), NID_secp384r1, N_ALL_KEYS);
   }
   dprintf(0, "TEST END: test_generate_eccp384()\n");
 }
 
 static void test_sign_eccp(CK_BYTE* key_params, CK_ULONG key_params_len, CK_ULONG key_len, int curve) {
   EC_KEY      *eck;
-  CK_OBJECT_HANDLE obj_pvtkey[N_SELECTED_KEYS], obj_cert[N_SELECTED_KEYS];
+  CK_OBJECT_HANDLE obj_pvtkey[N_SELECTED_KEYS]={0}, obj_cert[N_SELECTED_KEYS]={0};
   CK_SESSION_HANDLE session;
 
   init_connection();
@@ -466,6 +471,7 @@ static void test_sign_eccp(CK_BYTE* key_params, CK_ULONG key_params_len, CK_ULON
     test_ec_sign_thorough(funcs, session, obj_pvtkey, CKM_ECDSA_SHA384, eck, key_len);
   }
 
+  EC_KEY_free(eck);
   destroy_test_objects(funcs, session, obj_pvtkey, N_SELECTED_KEYS);
   asrt(funcs->C_CloseSession(session), CKR_OK, "CloseSession");
   asrt(funcs->C_Finalize(NULL), CKR_OK, "FINALIZE");
@@ -488,7 +494,7 @@ static void test_sign_eccp384() {
 }
 
 static void test_generate_rsa(CK_ULONG key_size, CK_BYTE n_keys) {
-  CK_OBJECT_HANDLE obj_pvtkey[N_ALL_KEYS], obj_pubkey[N_ALL_KEYS];
+  CK_OBJECT_HANDLE obj_pvtkey[N_ALL_KEYS]={0}, obj_pubkey[N_ALL_KEYS]={0};
   CK_SESSION_HANDLE session;
 
   init_connection();
@@ -526,18 +532,18 @@ static void test_key_attributes() {
   asrt(funcs->C_OpenSession(0, CKF_SERIAL_SESSION | CKF_RW_SESSION, NULL, NULL, &session), CKR_OK, "OpenSession1");
 
   generate_ec_keys(funcs, session, 1, params_eccp256, sizeof(params_eccp256), &pubkey, &privkey);
-  test_pubkey_attributes_ec(funcs, session, pubkey, 256, "Public key for PIV Authentication", 67, params_eccp256, sizeof(params_eccp256), is_neo);
-  test_privkey_attributes_ec(funcs, session, privkey, 256, "Private key for PIV Authentication", 67, params_eccp256, sizeof(params_eccp256), CK_FALSE, is_neo);
+  test_pubkey_attributes_ec(funcs, session, pubkey, 256, "Public key for PIV Authentication", 67, params_eccp256, sizeof(params_eccp256));
+  test_privkey_attributes_ec(funcs, session, privkey, 256, "Private key for PIV Authentication", 67, params_eccp256, sizeof(params_eccp256), CK_FALSE);
   
   if(!is_neo) {
     generate_ec_keys(funcs, session, 1, params_eccp384, sizeof(params_eccp384), &pubkey, &privkey);
-    test_pubkey_attributes_ec(funcs, session, pubkey, 384, "Public key for PIV Authentication", 99, params_eccp384, sizeof(params_eccp384), 0);
-    test_privkey_attributes_ec(funcs, session, privkey, 384, "Private key for PIV Authentication", 99, params_eccp384, sizeof(params_eccp384), CK_FALSE, CK_FALSE);
+    test_pubkey_attributes_ec(funcs, session, pubkey, 384, "Public key for PIV Authentication", 99, params_eccp384, sizeof(params_eccp384));
+    test_privkey_attributes_ec(funcs, session, privkey, 384, "Private key for PIV Authentication", 99, params_eccp384, sizeof(params_eccp384), CK_FALSE);
   }
   
   generate_rsa_keys(funcs, session, 1024, 1, &pubkey, &privkey);
-  test_pubkey_attributes_rsa(funcs, session, pubkey, 1024, "Public key for PIV Authentication", 128, e, sizeof(e), is_neo);
-  test_privkey_attributes_rsa(funcs, session, privkey, 1024, "Private key for PIV Authentication", 128, e, sizeof(e), CK_FALSE, is_neo);
+  test_pubkey_attributes_rsa(funcs, session, pubkey, 1024, "Public key for PIV Authentication", 128, e, sizeof(e));
+  test_privkey_attributes_rsa(funcs, session, privkey, 1024, "Private key for PIV Authentication", 128, e, sizeof(e), CK_FALSE);
 
   destroy_test_objects(funcs, session, &privkey, 1);
   asrt(funcs->C_CloseSession(session), CKR_OK, "CloseSession");
@@ -559,7 +565,7 @@ static void test_find_objects() {
 
   asrt(funcs->C_Login(session, CKU_USER, "123456", 6), CKR_OK, "LOGIN USER");
 
-  CK_OBJECT_HANDLE found_obj[10];
+  CK_OBJECT_HANDLE found_obj[10] = {0};
   CK_ULONG n_found_obj = 0;
   CK_ULONG object_class;
   CK_BYTE ckaid = 0;
@@ -629,7 +635,7 @@ static void test_find_objects() {
 }
 
 static CK_OBJECT_HANDLE get_public_key_handle(CK_SESSION_HANDLE session, CK_OBJECT_HANDLE privkey) {
-  CK_OBJECT_HANDLE found_obj[10];
+  CK_OBJECT_HANDLE found_obj[10] = {0};
   CK_ULONG n_found_obj = 0;
   CK_ULONG class_pub = CKO_PUBLIC_KEY;
   CK_BYTE ckaid = 0;
@@ -652,7 +658,7 @@ static CK_OBJECT_HANDLE get_public_key_handle(CK_SESSION_HANDLE session, CK_OBJE
 
 static void test_import_eccp(CK_BYTE* key_params, CK_ULONG key_params_len, CK_ULONG key_len, int curve, CK_BYTE n_keys) {
   EC_KEY            *eck;
-  CK_OBJECT_HANDLE  obj_cert[N_ALL_KEYS], obj_pvtkey[N_ALL_KEYS];
+  CK_OBJECT_HANDLE  obj_cert[N_ALL_KEYS]={0}, obj_pvtkey[N_ALL_KEYS]={0};
   CK_SESSION_HANDLE session;
 
   init_connection();
@@ -663,7 +669,9 @@ static void test_import_eccp(CK_BYTE* key_params, CK_ULONG key_params_len, CK_UL
     exit(EXIT_FAILURE);
 
   test_ec_sign_simple(funcs, session, obj_pvtkey, n_keys, eck, key_len);
+  test_ec_ecdh_simple(funcs, session, obj_pvtkey, n_keys, curve);
 
+  EC_KEY_free(eck);
   destroy_test_objects(funcs, session, obj_cert, n_keys);
   asrt(funcs->C_CloseSession(session), CKR_OK, "CloseSession");
   asrt(funcs->C_Finalize(NULL), CKR_OK, "FINALIZE");
@@ -692,18 +700,20 @@ static void test_import_eccp384() {
 static void test_import_rsa(CK_ULONG key_size, CK_BYTE n_keys) {
   EVP_PKEY    *evp = EVP_PKEY_new();
   RSA         *rsak = RSA_new();
-  CK_OBJECT_HANDLE obj_cert[N_ALL_KEYS], obj_pvtkey[N_ALL_KEYS];
+  CK_OBJECT_HANDLE obj_cert[N_ALL_KEYS]={0}, obj_pvtkey[N_ALL_KEYS]={0};
   CK_SESSION_HANDLE session;
 
   init_connection();
   asrt(funcs->C_OpenSession(0, CKF_SERIAL_SESSION | CKF_RW_SESSION, NULL, NULL, &session), CKR_OK, "OpenSession1");
 
-  import_rsa_key(funcs, session, key_size, evp, rsak, n_keys, obj_cert, obj_pvtkey);
+  import_rsa_key(funcs, session, key_size, &evp, &rsak, n_keys, obj_cert, obj_pvtkey);
   if (evp == NULL || rsak == NULL)
     exit(EXIT_FAILURE);
 
   test_rsa_sign_simple(funcs, session, obj_pvtkey, n_keys, evp);
 
+  RSA_free(rsak);
+  EVP_PKEY_free(evp);
   destroy_test_objects(funcs, session, obj_cert, n_keys);
   asrt(funcs->C_CloseSession(session), CKR_OK, "CloseSession");
   asrt(funcs->C_Finalize(NULL), CKR_OK, "FINALIZE");
@@ -725,13 +735,13 @@ dprintf(0, "TEST END: test_import_rsakeys()\n");
 static void test_sign_rsa(CK_ULONG key_size) {
   EVP_PKEY    *evp = EVP_PKEY_new();
   RSA         *rsak = RSA_new();
-  CK_OBJECT_HANDLE obj_cert[N_SELECTED_KEYS], obj_pvtkey[N_SELECTED_KEYS];
+  CK_OBJECT_HANDLE obj_cert[N_SELECTED_KEYS]={0}, obj_pvtkey[N_SELECTED_KEYS]={0};
   CK_SESSION_HANDLE session;
 
   init_connection();
   asrt(funcs->C_OpenSession(0, CKF_SERIAL_SESSION | CKF_RW_SESSION, NULL, NULL, &session), CKR_OK, "OpenSession1");
 
-  import_rsa_key(funcs, session, key_size, evp, rsak, N_SELECTED_KEYS, obj_cert, obj_pvtkey);
+  import_rsa_key(funcs, session, key_size, &evp, &rsak, N_SELECTED_KEYS, obj_cert, obj_pvtkey);
   if (evp == NULL || rsak == NULL)
     exit(EXIT_FAILURE);
 
@@ -745,6 +755,8 @@ static void test_sign_rsa(CK_ULONG key_size) {
   test_rsa_sign_pss(funcs, session, obj_pvtkey, N_SELECTED_KEYS, rsak, CKM_SHA256_RSA_PKCS_PSS);
   test_rsa_sign_pss(funcs, session, obj_pvtkey, N_SELECTED_KEYS, rsak, CKM_SHA384_RSA_PKCS_PSS);
 
+  RSA_free(rsak);
+  EVP_PKEY_free(evp);
   destroy_test_objects(funcs, session, obj_cert, N_SELECTED_KEYS);
   asrt(funcs->C_CloseSession(session), CKR_OK, "CloseSession");
   asrt(funcs->C_Finalize(NULL), CKR_OK, "FINALIZE");
@@ -761,13 +773,13 @@ static void test_decrypt_RSA() {
   dprintf(0, "TEST START: test_decrypt_RSA()\n");
   EVP_PKEY    *evp = EVP_PKEY_new();
   RSA         *rsak = RSA_new();
-  CK_OBJECT_HANDLE obj_cert[N_SELECTED_KEYS], obj_pvtkey[N_SELECTED_KEYS];
+  CK_OBJECT_HANDLE obj_cert[N_SELECTED_KEYS]={0}, obj_pvtkey[N_SELECTED_KEYS]={0};
   CK_SESSION_HANDLE session;
 
   init_connection();
   asrt(funcs->C_OpenSession(0, CKF_SERIAL_SESSION | CKF_RW_SESSION, NULL, NULL, &session), CKR_OK, "OpenSession1");
 
-  import_rsa_key(funcs, session, 1024, evp, rsak, N_SELECTED_KEYS, obj_cert, obj_pvtkey);
+  import_rsa_key(funcs, session, 1024, &evp, &rsak, N_SELECTED_KEYS, obj_cert, obj_pvtkey);
   if (evp == NULL ||  rsak == NULL)
     exit(EXIT_FAILURE);
 
@@ -777,7 +789,9 @@ static void test_decrypt_RSA() {
 
   test_rsa_decrypt_oaep(funcs, session, obj_pvtkey, N_SELECTED_KEYS, CKM_SHA_1, rsak);
   test_rsa_decrypt_oaep(funcs, session, obj_pvtkey, N_SELECTED_KEYS, CKM_SHA256, rsak);
-  
+
+  RSA_free(rsak);
+  EVP_PKEY_free(evp);
   destroy_test_objects(funcs, session, obj_cert, N_SELECTED_KEYS);
   asrt(funcs->C_CloseSession(session), CKR_OK, "CloseSession");
   asrt(funcs->C_Finalize(NULL), CKR_OK, "FINALIZE");
@@ -788,13 +802,13 @@ static void test_encrypt_RSA() {
   dprintf(0, "TEST START: test_encrypt_RSA()\n");
   EVP_PKEY    *evp = EVP_PKEY_new();
   RSA         *rsak = RSA_new();
-  CK_OBJECT_HANDLE obj_cert[N_SELECTED_KEYS], obj_pvtkey[N_SELECTED_KEYS];
+  CK_OBJECT_HANDLE obj_cert[N_SELECTED_KEYS]={0}, obj_pvtkey[N_SELECTED_KEYS]={0};
   CK_SESSION_HANDLE session;
 
   init_connection();
   asrt(funcs->C_OpenSession(0, CKF_SERIAL_SESSION | CKF_RW_SESSION, NULL, NULL, &session), CKR_OK, "OpenSession1");
 
-  import_rsa_key(funcs, session, 1024, evp, rsak, N_SELECTED_KEYS, obj_cert, obj_pvtkey);
+  import_rsa_key(funcs, session, 1024, &evp, &rsak, N_SELECTED_KEYS, obj_cert, obj_pvtkey);
   if (evp == NULL ||  rsak == NULL)
     exit(EXIT_FAILURE);
 
@@ -802,6 +816,8 @@ static void test_encrypt_RSA() {
   test_rsa_encrypt(funcs, session, obj_pvtkey, N_SELECTED_KEYS, rsak, CKM_RSA_X_509, RSA_NO_PADDING);
   test_rsa_encrypt(funcs, session, obj_pvtkey, N_SELECTED_KEYS, rsak, CKM_RSA_PKCS_OAEP, RSA_PKCS1_OAEP_PADDING);
 
+  RSA_free(rsak);
+  EVP_PKEY_free(evp);
   destroy_test_objects(funcs, session, obj_cert, N_SELECTED_KEYS);
   asrt(funcs->C_CloseSession(session), CKR_OK, "CloseSession");
   asrt(funcs->C_Finalize(NULL), CKR_OK, "FINALIZE");
@@ -827,25 +843,26 @@ static void test_digest() {
 #endif
 
 int destruction_confirmed(void) {
-  char *confirmed = getenv("YKPIV_ENV_HWTESTS_CONFIRMED");
-  if (confirmed && confirmed[0] == '1')
-    return 1;
-  // Use dprintf() to write directly to stdout, since automake eats the standard stdout/stderr pointers.
-  dprintf(0, "\n***\n*** Hardware tests skipped.  Run \"make hwcheck\".\n***\n\n");
-  return 0;
+#ifdef _WIN32
+  return 1;
+#else
+  return system("../../../tools/confirm.sh") == 0;
+#endif
 }
 
 int main(void) {
 
-  get_functions(&funcs);
+  get_functions();
 
   test_lib_info();
 
 #if HW_TESTS
   // Require user confirmation to continue, since this test suite will clear
   // any data stored on connected keys.
-  if (!destruction_confirmed())
+  if (!destruction_confirmed()) {
+    dprintf(0, "\n***\n*** Hardware tests skipped.\n***\n\n");
     exit(77); // exit code 77 == skipped tests
+  }
 
   test_initalize();
   // Require YK4, YK5 or NEO to continue.  Skip if different model found.
@@ -879,4 +896,4 @@ int main(void) {
   return EXIT_SUCCESS;
 }
 
-#pragma clang diagnostic pop
\ No newline at end of file
+//#pragma clang diagnostic pop
\ No newline at end of file
