@@ -1,4 +1,34 @@
-#include "../../tool/openssl-compat.h"
+/*
+ * Copyright (c) 2019-2020 Yubico AB
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are
+ * met:
+ *
+ *   * Redistributions of source code must retain the above copyright
+ *     notice, this list of conditions and the following disclaimer.
+ *
+ *   * Redistributions in binary form must reproduce the above
+ *     copyright notice, this list of conditions and the following
+ *     disclaimer in the documentation and/or other materials provided
+ *     with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+ * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ *
+ */
+
+#include "../../common/openssl-compat.h"
 
 #include <string.h>
 #include <stdlib.h>
@@ -55,7 +85,7 @@ void dump_hex(const unsigned char *buf, unsigned int len, FILE *output, int spac
 
 static CK_OBJECT_HANDLE get_public_key_handle(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, 
                         CK_OBJECT_HANDLE privkey) {
-  CK_OBJECT_HANDLE found_obj[10];
+  CK_OBJECT_HANDLE found_obj[10] = {0};
   CK_ULONG n_found_obj = 0;
   CK_ULONG class_pub = CKO_PUBLIC_KEY;
   CK_BYTE ckaid = 0;
@@ -92,7 +122,7 @@ static CK_RV get_hash(CK_MECHANISM_TYPE mech,
     return CKR_FUNCTION_FAILED;
   }
 
-  CK_BYTE hashed_data[512];
+  CK_BYTE hashed_data[512] = {0};
   switch(mech) {
     case CKM_SHA_1:
     case CKM_RSA_PKCS_PSS:
@@ -132,7 +162,7 @@ static CK_RV get_digest(CK_MECHANISM_TYPE mech,
     return CKR_FUNCTION_FAILED;
   }
 
-  CK_BYTE hashed_data[512];
+  CK_BYTE hashed_data[512] = {0};
   switch(mech) {
     case CKM_ECDSA:
     case CKM_RSA_PKCS:
@@ -244,13 +274,13 @@ static CK_MECHANISM_TYPE get_md_of(CK_MECHANISM_TYPE mech) {
 
 void test_digest_func(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_MECHANISM_TYPE mech_type) {
   CK_BYTE     i;
-  CK_BYTE     data[32];
+  CK_BYTE     data[32] = {0};
   CK_ULONG    data_len = sizeof(data);
-  CK_BYTE     digest[128];
+  CK_BYTE     digest[128] = {0};
   CK_ULONG    digest_len;
-  CK_BYTE     digest_update[128];
+  CK_BYTE     digest_update[128] = {0};
   CK_ULONG    digest_update_len;
-  CK_BYTE     hdata[128];
+  CK_BYTE     hdata[128] = {0};
   CK_ULONG    hdata_len;
 
   CK_MECHANISM mech = {mech_type, NULL, 0};
@@ -282,16 +312,16 @@ EC_KEY* import_ec_key(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_
   EVP_PKEY       *evp;
   EC_KEY         *eck;
   const BIGNUM   *bn;
-  CK_CHAR        pvt[key_len];
   X509           *cert;
-  ASN1_TIME      *tm;
-  CK_BYTE        i, j;
+  CK_BYTE        i;
+  CK_CHAR        *pvt;
+  pvt = malloc(key_len);
 
   CK_ULONG    class_k = CKO_PRIVATE_KEY;
   CK_ULONG    class_c = CKO_CERTIFICATE;
   CK_ULONG    kt = CKK_ECDSA;
   CK_BYTE     id = 0;
-  CK_BYTE     value_c[3100];
+  CK_BYTE     value_c[3100] = {0};
   CK_ULONG    cert_len;
 
   unsigned char  *p;
@@ -301,7 +331,7 @@ EC_KEY* import_ec_key(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_
     {CKA_KEY_TYPE, &kt, sizeof(kt)},
     {CKA_ID, &id, sizeof(id)},
     {CKA_EC_PARAMS, ec_params, ec_params_len},
-    {CKA_VALUE, pvt, sizeof(pvt)}
+    {CKA_VALUE, pvt, key_len}
   };
 
   CK_ATTRIBUTE publicKeyTemplate[] = {
@@ -364,27 +394,30 @@ EC_KEY* import_ec_key(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_
   }
 
   asrt(funcs->C_Logout(session), CKR_OK, "Logout SO");
-
+  free(pvt);
+  X509_free(cert);
+  EVP_PKEY_free(evp);
   return eck;
 }
 
-void import_rsa_key(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, int keylen, EVP_PKEY* evp, RSA* rsak,
+void import_rsa_key(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, int keylen, EVP_PKEY** evp, RSA** rsak,
                     CK_BYTE n_keys, CK_OBJECT_HANDLE_PTR obj_cert, CK_OBJECT_HANDLE_PTR obj_pvtkey) {
   X509        *cert;
-  ASN1_TIME   *tm;
-  CK_BYTE     i, j;
+  CK_BYTE     i;
   CK_BYTE     e[] = {0x01, 0x00, 0x01};
-  CK_BYTE     p[keylen / 16];
-  CK_BYTE     q[keylen / 16];
-  CK_BYTE     dp[keylen / 16];
-  CK_BYTE     dq[keylen / 16];
-  CK_BYTE     qinv[keylen / 16];
+  CK_BYTE     *p, *q, *dp, *dq, *qinv;
+  p = malloc(keylen / 16);
+  q = malloc(keylen / 16);
+  dp = malloc(keylen / 16);
+  dq = malloc(keylen / 16);
+  qinv = malloc(keylen / 16);
+
   BIGNUM      *e_bn;
   CK_ULONG    class_k = CKO_PRIVATE_KEY;
   CK_ULONG    class_c = CKO_CERTIFICATE;
   CK_ULONG    kt = CKK_RSA;
   CK_BYTE     id = 0;
-  CK_BYTE     value_c[3100];
+  CK_BYTE     value_c[3100] = {0};
   CK_ULONG    cert_len;
   const BIGNUM *bp, *bq, *biqmp, *bdmp1, *bdmq1;
 
@@ -397,11 +430,11 @@ void import_rsa_key(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, int k
     {CKA_KEY_TYPE, &kt, sizeof(kt)},
     {CKA_ID, &id, sizeof(id)},
     {CKA_PUBLIC_EXPONENT, e, sizeof(e)},
-    {CKA_PRIME_1, p, sizeof(p)},
-    {CKA_PRIME_2, q, sizeof(q)},
-    {CKA_EXPONENT_1, dp, sizeof(dp)},
-    {CKA_EXPONENT_2, dq, sizeof(dq)},
-    {CKA_COEFFICIENT, qinv, sizeof(qinv)}
+    {CKA_PRIME_1, p, (keylen / 16)},
+    {CKA_PRIME_2, q, (keylen / 16)},
+    {CKA_EXPONENT_1, dp, (keylen / 16)},
+    {CKA_EXPONENT_2, dq, (keylen / 16)},
+    {CKA_COEFFICIENT, qinv, (keylen / 16)}
   };
 
   CK_ATTRIBUTE publicKeyTemplate[] = {
@@ -417,10 +450,10 @@ void import_rsa_key(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, int k
     exit(EXIT_FAILURE);
 
   do {
-    asrt(RSA_generate_key_ex(rsak, keylen, e_bn, NULL), 1, "GENERATE RSAK");
+    asrt(RSA_generate_key_ex(*rsak, keylen, e_bn, NULL), 1, "GENERATE RSAK");
 
-    RSA_get0_factors(rsak, &bp, &bq);
-    RSA_get0_crt_params(rsak, &bdmp1, &bdmq1, &biqmp);
+    RSA_get0_factors(*rsak, &bp, &bq);
+    RSA_get0_crt_params(*rsak, &bdmp1, &bdmq1, &biqmp);
     p_len = BN_bn2bin(bp, p);
     q_len = BN_bn2bin(bq, q);
     dp_len = BN_bn2bin(bdmp1, dp);
@@ -430,7 +463,7 @@ void import_rsa_key(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, int k
   } while(!len_correct);
 
 
-  if (EVP_PKEY_set1_RSA(evp, rsak) == 0)
+  if (EVP_PKEY_set1_RSA(*evp, *rsak) == 0)
     exit(EXIT_FAILURE);
 
   cert = X509_new();
@@ -445,10 +478,10 @@ void import_rsa_key(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, int k
   X509_gmtime_adj(X509_get_notBefore(cert), 0);
   X509_gmtime_adj(X509_get_notAfter(cert), 0);
 
-  if (X509_set_pubkey(cert, evp) == 0)
+  if (X509_set_pubkey(cert, *evp) == 0)
     exit(EXIT_FAILURE);
 
-  if (X509_sign(cert, evp, EVP_sha1()) == 0)
+  if (X509_sign(cert, *evp, EVP_sha1()) == 0)
     exit(EXIT_FAILURE);
 
   px = value_c;
@@ -469,7 +502,13 @@ void import_rsa_key(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, int k
 
   asrt(funcs->C_Logout(session), CKR_OK, "Logout SO");
 
-  evp = X509_get_pubkey(cert);
+  X509_free(cert);
+  BN_free(e_bn);
+  free(p);
+  free(q);
+  free(dp);
+  free(dq);
+  free(qinv);
 }
 
 void generate_ec_keys(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_BYTE n_keys,
@@ -508,7 +547,7 @@ void generate_ec_keys(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_
 
 void generate_rsa_keys(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_ULONG key_size, CK_BYTE n_keys,
                       CK_OBJECT_HANDLE_PTR obj_pubkey, CK_OBJECT_HANDLE_PTR obj_pvtkey) {
-  CK_BYTE     i, j;
+  CK_BYTE     i;
   CK_BYTE     e[] = {0x01, 0x00, 0x01};
   CK_ULONG    class_k = CKO_PRIVATE_KEY;
   CK_ULONG    class_c = CKO_PUBLIC_KEY;
@@ -541,67 +580,62 @@ void generate_rsa_keys(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK
 }
 
 static void construct_der_encoded_sig(CK_BYTE sig[], CK_BYTE_PTR der_encoded, CK_ULONG key_len) {
-    CK_ULONG    der_encoded_len;
-    CK_BYTE_PTR der_ptr;
-    CK_BYTE_PTR r_ptr;
-    CK_BYTE_PTR s_ptr;
-    CK_ULONG    r_len;
-    CK_ULONG    s_len;
-    const EVP_MD *md;
-    EVP_MD_CTX *mdctx;
+  CK_BYTE_PTR der_ptr;
+  CK_BYTE_PTR r_ptr;
+  CK_BYTE_PTR s_ptr;
+  CK_ULONG r_len;
+  CK_ULONG s_len;
 
-      r_len = key_len;
-      s_len = key_len;
+  r_len = key_len;
+  s_len = key_len;
 
-      der_ptr = der_encoded;
-      *der_ptr++ = 0x30;
-      *der_ptr++ = 0xff; // placeholder, fix below
+  der_ptr = der_encoded;
+  *der_ptr++ = 0x30;
+  *der_ptr++ = 0xff; // placeholder, fix below
 
-      r_ptr = sig;
+  r_ptr = sig;
 
-      *der_ptr++ = 0x02;
-      *der_ptr++ = r_len;
-      if (*r_ptr >= 0x80) {
-        *(der_ptr - 1) = *(der_ptr - 1) + 1;
-        *der_ptr++ = 0x00;
-      }
-      else if (*r_ptr == 0x00 && *(r_ptr + 1) < 0x80) {
-        r_len--;
-        *(der_ptr - 1) = *(der_ptr - 1) - 1;
-        r_ptr++;
-      }
-      memcpy(der_ptr, r_ptr, r_len);
-      der_ptr+= r_len;
+  *der_ptr++ = 0x02;
+  *der_ptr++ = r_len;
+  if (*r_ptr >= 0x80) {
+    *(der_ptr - 1) = *(der_ptr - 1) + 1;
+    *der_ptr++ = 0x00;
+  } else if (*r_ptr == 0x00 && *(r_ptr + 1) < 0x80) {
+    r_len--;
+    *(der_ptr - 1) = *(der_ptr - 1) - 1;
+    r_ptr++;
+  }
+  memcpy(der_ptr, r_ptr, r_len);
+  der_ptr += r_len;
 
-      s_ptr = sig + key_len;
+  s_ptr = sig + key_len;
 
-      *der_ptr++ = 0x02;
-      *der_ptr++ = s_len;
-      if (*s_ptr >= 0x80) {
-        *(der_ptr - 1) = *(der_ptr - 1) + 1;
-        *der_ptr++ = 0x00;
-      }
-      else if (*s_ptr == 0x00 && *(s_ptr + 1) < 0x80) {
-        s_len--;
-        *(der_ptr - 1) = *(der_ptr - 1) - 1;
-        s_ptr++;
-      }
-      memcpy(der_ptr, s_ptr, s_len);
-      der_ptr+= s_len;
+  *der_ptr++ = 0x02;
+  *der_ptr++ = s_len;
+  if (*s_ptr >= 0x80) {
+    *(der_ptr - 1) = *(der_ptr - 1) + 1;
+    *der_ptr++ = 0x00;
+  } else if (*s_ptr == 0x00 && *(s_ptr + 1) < 0x80) {
+    s_len--;
+    *(der_ptr - 1) = *(der_ptr - 1) - 1;
+    s_ptr++;
+  }
+  memcpy(der_ptr, s_ptr, s_len);
+  der_ptr += s_len;
 
-      der_encoded[1] = der_ptr - der_encoded - 2;
+  der_encoded[1] = der_ptr - der_encoded - 2;
 }
 
 void test_ec_sign_simple(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_OBJECT_HANDLE_PTR obj_pvtkey, 
                          CK_BYTE n_keys, EC_KEY *eck, CK_ULONG key_len) {
                     
   CK_BYTE     i;
-  CK_BYTE     data[32];
+  CK_BYTE     data[32] = {0};
   CK_ULONG    data_len;
-  CK_BYTE     sig[256];
+  CK_BYTE     sig[256] = {0};
   CK_ULONG    sig_len;
 
-  CK_BYTE     der_encoded[116];
+  CK_BYTE     der_encoded[116] = {0};
 
   CK_MECHANISM mech = {CKM_ECDSA, NULL, 0};
 
@@ -630,18 +664,80 @@ void test_ec_sign_simple(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session,
   asrt(funcs->C_Logout(session), CKR_OK, "Logout USER");  
 }
 
+void test_ec_ecdh_simple(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_OBJECT_HANDLE_PTR obj_pvtkey, 
+                         CK_BYTE n_keys, int curve) {
+                    
+  CK_BYTE     i;
+  CK_BYTE     pubkey[128]={0}, pubkey2[128]={0}, secret[128]={0}, secret2[128]={0};
+
+  CK_ULONG    cls = CKO_SECRET_KEY;
+  CK_ULONG    kt = CKK_GENERIC_SECRET;
+
+  CK_BBOOL    _false = CK_FALSE;
+  CK_BBOOL    _true = CK_TRUE;
+
+  EC_KEY *tmpkey = EC_KEY_new_by_curve_name(curve);
+
+  if (tmpkey == NULL)
+    exit(EXIT_FAILURE);
+
+  asrt(EC_KEY_generate_key(tmpkey), 1, "GENERATE ECK");
+
+  int bits = EC_GROUP_get_degree(EC_KEY_get0_group(tmpkey));
+  unsigned char *ptr = pubkey;
+  asrt(i2o_ECPublicKey(tmpkey, &ptr), bits / 4 + 1, "ENCODE ECK");
+
+  CK_ECDH1_DERIVE_PARAMS params = {CKD_NULL, 0, NULL, ptr-pubkey, pubkey};
+  CK_MECHANISM mech = {CKM_ECDH1_DERIVE, &params, sizeof(params)};
+  CK_OBJECT_HANDLE sk;
+
+  for (i = 0; i < n_keys; i++) {
+
+    CK_ATTRIBUTE deriveKeyTemplate[] = {
+      {CKA_TOKEN, &_false, sizeof(_false)},
+      {CKA_CLASS, &cls, sizeof(cls)},
+      {CKA_KEY_TYPE, &kt, sizeof(kt)},
+      {CKA_EXTRACTABLE, &_true, sizeof(_true)},
+    };
+
+    CK_ATTRIBUTE pointTemplate[] = {
+      {CKA_EC_POINT, pubkey2, sizeof(pubkey2)},
+    };
+
+    CK_ATTRIBUTE valueTemplate[] = {
+      {CKA_VALUE, secret2, sizeof(secret2)},
+    };
+
+    asrt(funcs->C_Login(session, CKU_USER, (CK_CHAR_PTR)"123456", 6), CKR_OK, "Login USER");
+    asrt(funcs->C_GetAttributeValue(session, obj_pvtkey[i], pointTemplate, 1), CKR_OK, "GetAttributeValue");
+    asrt(funcs->C_DeriveKey(session, &mech, obj_pvtkey[i], deriveKeyTemplate, 4, &sk), CKR_OK, "DeriveKey");
+    asrt(funcs->C_GetAttributeValue(session, sk, valueTemplate, 1), CKR_OK, "GetAttributeValue");
+    asrt(funcs->C_DestroyObject(session, sk), CKR_OK, "DestroyObject");
+    asrt(funcs->C_Logout(session), CKR_OK, "Logout USER");
+    // Skip DER encoding
+    ptr = pointTemplate->pValue;
+    ptr += 2;
+    EC_KEY *pk = EC_KEY_new_by_curve_name(curve);
+    pk = o2i_ECPublicKey(&pk, &ptr, pointTemplate->ulValueLen - 2);
+    asrt(ECDH_compute_key(secret, sizeof(secret), EC_KEY_get0_public_key(pk), tmpkey, NULL), bits / 8, "ECDH_compute_key");
+    asrt(memcmp(secret, secret2, bits / 8), 0, "Compare secrets");
+    EC_KEY_free(pk);
+  }
+  EC_KEY_free(tmpkey);
+}
+
 void test_ec_sign_thorough(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_OBJECT_HANDLE_PTR obj_pvtkey, 
                            CK_MECHANISM_TYPE mech_type, EC_KEY *eck, CK_ULONG key_len) {
                     
   CK_BYTE       i, j;
-  CK_BYTE       data[32];
+  CK_BYTE       data[32] = {0};
   CK_ULONG      data_len;
-  CK_BYTE       hdata[64];
+  CK_BYTE       hdata[64] = {0};
   unsigned int  hdata_len;  
   CK_BYTE*      sig;
   CK_ULONG      sig_len;
 
-  CK_BYTE     der_encoded[116];
+  CK_BYTE     der_encoded[116] = {0};
   const EVP_MD *md;
   EVP_MD_CTX *mdctx;
 
@@ -682,6 +778,7 @@ void test_ec_sign_thorough(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session
           EVP_DigestInit_ex(mdctx, md, NULL);
           EVP_DigestUpdate(mdctx, data, data_len);
           EVP_DigestFinal_ex(mdctx, hdata, &hdata_len);
+          EVP_MD_CTX_destroy(mdctx);
         }
 
         construct_der_encoded_sig(sig, der_encoded, key_len);
@@ -697,14 +794,11 @@ void test_ec_sign_thorough(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session
 void test_rsa_sign_simple(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_OBJECT_HANDLE_PTR obj_pvtkey, 
                           CK_BYTE n_keys, EVP_PKEY* evp) {
   CK_BYTE     i;
-  CK_BYTE     data[32];
-  CK_BYTE     sig[256];
+  CK_BYTE     data[32] = {0};
+  CK_BYTE     sig[256] = {0};
   CK_ULONG    sig_len;
   EVP_PKEY_CTX *ctx = NULL;
 
-  CK_BYTE     hdata[512];
-  CK_ULONG    hdata_len;
-
   CK_OBJECT_HANDLE obj_pubkey;
   CK_MECHANISM mech = {CKM_RSA_PKCS, NULL, 0};
 
@@ -729,6 +823,7 @@ void test_rsa_sign_simple(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session,
       asrt(EVP_PKEY_verify_init(ctx) > 0, 1, "EVP_KEY_verify_init");
       asrt(EVP_PKEY_CTX_set_signature_md(ctx, NULL) > 0, 1, "EVP_PKEY_CTX_set_signature_md");
       asrt(EVP_PKEY_verify(ctx, sig, sig_len, data, 32), 1, "EVP_PKEY_verify");
+      EVP_PKEY_CTX_free(ctx);
     } else {
       // Internal verification: Verify
       asrt(funcs->C_VerifyInit(session, &mech, obj_pubkey), CKR_OK, "VERIFY INIT");
@@ -742,14 +837,14 @@ void test_rsa_sign_simple(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session,
 void test_rsa_sign_thorough(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_OBJECT_HANDLE_PTR obj_pvtkey, 
                             CK_BYTE n_keys, EVP_PKEY* evp, CK_MECHANISM_TYPE mech_type) {
   CK_BYTE     i, j;
-  CK_BYTE     data[32];
+  CK_BYTE     data[32] = {0};
   CK_BYTE*    sig;
   CK_BYTE*    sig_update;
   CK_ULONG    sig_len;
   CK_ULONG    sig_update_len;
   EVP_PKEY_CTX *ctx = NULL;
 
-  CK_BYTE     hdata[512];
+  CK_BYTE     hdata[512] = {0};
   CK_ULONG    hdata_len;
 
   CK_OBJECT_HANDLE obj_pubkey;
@@ -780,6 +875,7 @@ void test_rsa_sign_thorough(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE sessio
         asrt(EVP_PKEY_verify_init(ctx) > 0, 1, "EVP_KEY_verify_init");
         asrt(EVP_PKEY_CTX_set_signature_md(ctx, NULL) > 0, 1, "EVP_PKEY_CTX_set_signature_md");
         asrt(EVP_PKEY_verify(ctx, sig, sig_len, hdata, hdata_len), 1, "EVP_PKEY_verify");
+        EVP_PKEY_CTX_free(ctx);
       }
       
       // Internal verification: Verify
@@ -824,7 +920,7 @@ void test_rsa_sign_pss(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK
   CK_ULONG    sig_update_len;
 
   CK_BYTE*     pss_buf;
-  CK_BYTE      digest_data[256];
+  CK_BYTE      digest_data[256] = {0};
   unsigned int digest_data_len = sizeof(digest_data);
   EVP_MD_CTX   *md_ctx;
 
@@ -864,6 +960,7 @@ void test_rsa_sign_pss(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK
           asrt(EVP_DigestInit_ex(md_ctx, get_md_type(mech_type), NULL), 1, "DIGEST INIT");
           asrt(EVP_DigestUpdate(md_ctx, data, pss_params.sLen), 1, "DIGEST UPDATE");
           asrt(EVP_DigestFinal_ex(md_ctx, digest_data, &digest_data_len), 1, "DIGEST FINAL");
+          EVP_MD_CTX_destroy(md_ctx);
 
           asrt(RSA_verify_PKCS1_PSS_mgf1(rsak, digest_data, get_md_type(pss_params.hashAlg), get_md_type(pss_params.mgf), pss_buf, pss_params.sLen), 1, "VERIFY PSS SIGNATURE");
         }
@@ -898,6 +995,7 @@ void test_rsa_sign_pss(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK
           asrt(EVP_DigestInit_ex(md_ctx, get_md_type(mech_type), NULL), 1, "DIGEST INIT");
           asrt(EVP_DigestUpdate(md_ctx, data, pss_params.sLen), 1, "DIGEST UPDATE");
           asrt(EVP_DigestFinal_ex(md_ctx, digest_data, &digest_data_len), 1, "DIGEST FINAL");
+          EVP_MD_CTX_destroy(md_ctx);
 
           asrt(RSA_verify_PKCS1_PSS_mgf1(rsak, digest_data, get_md_type(pss_params.hashAlg), get_md_type(pss_params.mgf), pss_buf, pss_params.sLen), 1, "VERIFY PSS SIGNATURE");
         }
@@ -926,7 +1024,7 @@ static int is_data_too_large(int enc_ret) {
   unsigned long err;
   ERR_load_crypto_strings();
   err = ERR_get_error();
-  //char err_str[128];  
+  //char err_str[128] = {0};
   //ERR_error_string_n(err, err_str, 128);
   //printf("%ld    %s\n", err, err_str);
   if(err == 67534980) { // Error code for "data too large for modulus"
@@ -937,10 +1035,10 @@ static int is_data_too_large(int enc_ret) {
 
 void test_rsa_decrypt(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_OBJECT_HANDLE_PTR obj_pvtkey, 
                       CK_BYTE n_keys, RSA* rsak, CK_MECHANISM_TYPE mech_type, CK_ULONG padding) {
-  CK_ULONG  i, j;
+  CK_BYTE   i, j;
   int       data_len, enc_len;
   CK_BYTE*  data;
-  CK_BYTE   enc[512];
+  CK_BYTE   enc[512] = {0};
   CK_BYTE*  dec;
   CK_ULONG  dec_len;
 
@@ -957,14 +1055,11 @@ void test_rsa_decrypt(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_
 
   for (i = 0; i < n_keys; i++) {
     for (j = 0; j < 4; j++) {
-      // This is because the numerical value of the clear data cannot be larger than the numerical value of the RSA key modulus
-      // Adding a padding takes care of this, but with RSA_NO_PADDING, we need to deal with that manually
-      do {
-        if(RAND_bytes(data, data_len) <= 0)
-          exit(EXIT_FAILURE);
+      if(RAND_bytes(data, data_len) <= 0)
+        exit(EXIT_FAILURE);
 
-        enc_len = RSA_public_encrypt(data_len, data, enc, rsak, padding);
-      } while(is_data_too_large(enc_len) == 1);
+      data[0] &= 0x7f; // Unset high bit to ensure it's less than modulus (required for raw RSA)
+      enc_len = RSA_public_encrypt(data_len, data, enc, rsak, padding);
 
       // Decrypt
       asrt(funcs->C_DecryptInit(session, &mech, obj_pvtkey[i]), CKR_OK, "DECRYPT INIT");
@@ -1003,11 +1098,11 @@ void test_rsa_decrypt_oaep(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session
                            CK_BYTE n_keys, CK_MECHANISM_TYPE mdhash,  RSA* rsak) {
   CK_ULONG  i, j;
   int       data_len;
-  CK_BYTE   data[32];
-  CK_BYTE   padded_data[512];
+  CK_BYTE   data[32] = {0};
+  CK_BYTE   padded_data[512] = {0};
   CK_BYTE   padded_data_len;
-  CK_BYTE   enc[512];
-  CK_BYTE   dec[512];
+  CK_BYTE   enc[512] = {0};
+  CK_BYTE   dec[512] = {0};
   CK_ULONG  dec_len;
   size_t    enc_len;
   const EVP_MD *md;
@@ -1060,12 +1155,12 @@ void test_rsa_decrypt_oaep(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session
 
 void test_rsa_encrypt(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_OBJECT_HANDLE_PTR obj_pvtkey, 
                       CK_BYTE n_keys, RSA* rsak, CK_MECHANISM_TYPE mech_type, CK_ULONG padding) {
-  CK_ULONG  i,j;
-  CK_BYTE   data[32];
+  CK_BYTE   i,j;
+  CK_BYTE   data[32] = {0};
   CK_ULONG  data_len = sizeof(data);
-  CK_BYTE   enc[128];
+  CK_BYTE   enc[128] = {0};
   CK_ULONG  enc_len;
-  CK_BYTE   dec[512];
+  CK_BYTE   dec[512] = {0};
   CK_ULONG  dec_len;
 
   CK_RSA_PKCS_OAEP_PARAMS params = {0};
@@ -1081,6 +1176,8 @@ void test_rsa_encrypt(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_
       if(RAND_bytes(data, data_len) <= 0)
         exit(EXIT_FAILURE);
 
+      data[0] &= 0x7f; // Unset high bit to ensure it's less than modulus (required for raw RSA)
+
       // Encrypt
       asrt(funcs->C_EncryptInit(session, &mech, pubkey), CKR_OK, "ENCRYPT INIT CKM_RSA_PKCS");
       enc_len = 0;
@@ -1123,7 +1220,7 @@ void test_rsa_encrypt(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, CK_
 
 static void test_pubkey_basic_attributes(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, 
                                          CK_OBJECT_HANDLE pubkey, CK_ULONG key_type, CK_ULONG key_size,
-                                         const unsigned char* label, CK_BBOOL is_neo) {
+                                         const unsigned char* label) {
   CK_ULONG obj_class;
   CK_BBOOL obj_token;
   CK_BBOOL obj_private;
@@ -1136,7 +1233,7 @@ static void test_pubkey_basic_attributes(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_
   CK_BBOOL obj_derive;
   CK_ULONG obj_modulus_bits;
   CK_BBOOL obj_modifiable;
-  char obj_label[1024];
+  char obj_label[1024] = {0};
   CK_ULONG obj_label_len;
 
   CK_ATTRIBUTE template[] = {
@@ -1164,11 +1261,7 @@ static void test_pubkey_basic_attributes(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_
   asrt(obj_private, CK_FALSE, "PRIVATE");
   asrt(obj_key_type, key_type, "KEY_TYPE");
   asrt(obj_trusted, CK_FALSE, "TRUSTED");
-  if(is_neo) {
-    asrt(obj_local, CK_FALSE, "LOCAL");
-  } else {
-    asrt(obj_local, CK_TRUE, "LOCAL");
-  }
+  asrt(obj_local, CK_TRUE, "LOCAL");
   asrt(obj_encrypt, CK_TRUE, "ENCRYPT");
   asrt(obj_verify, CK_TRUE, "VERIFY");
   asrt(obj_wrap, CK_FALSE, "WRAP");
@@ -1186,17 +1279,17 @@ static void test_pubkey_basic_attributes(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_
 void test_pubkey_attributes_rsa(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, 
                                 CK_OBJECT_HANDLE pubkey, CK_ULONG key_size, 
                                 const unsigned char* label, CK_ULONG modulus_len,
-                                CK_BYTE_PTR pubexp, CK_ULONG pubexp_len, CK_BBOOL is_neo) {
+                                CK_BYTE_PTR pubexp, CK_ULONG pubexp_len) {
 
-  CK_BYTE obj_pubexp[1024];
-  CK_BYTE obj_modulus[1024];
+  CK_BYTE obj_pubexp[1024] = {0};
+  CK_BYTE obj_modulus[1024] = {0};
 
   CK_ATTRIBUTE template[] = {
     {CKA_MODULUS, obj_modulus, sizeof(obj_modulus)},
     {CKA_PUBLIC_EXPONENT, &obj_pubexp, sizeof(obj_pubexp)},
   };
 
-  test_pubkey_basic_attributes(funcs, session, pubkey, CKK_RSA, key_size, label, is_neo);
+  test_pubkey_basic_attributes(funcs, session, pubkey, CKK_RSA, key_size, label);
 
   asrt(funcs->C_GetAttributeValue(session, pubkey, template, 2), CKR_OK, "GET RSA ATTRIBUTES");
   asrt(template[0].ulValueLen, modulus_len, "MODULUS LEN");
@@ -1207,16 +1300,16 @@ void test_pubkey_attributes_rsa(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE se
 void test_pubkey_attributes_ec(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, 
                                 CK_OBJECT_HANDLE pubkey, CK_ULONG key_size, 
                                 const unsigned char* label, CK_ULONG ec_point_len,
-                                CK_BYTE_PTR ec_params, CK_ULONG ec_params_len, CK_BBOOL is_neo) {
-  CK_BYTE obj_ec_point[1024];
-  CK_BYTE obj_ec_param[1024];
+                                CK_BYTE_PTR ec_params, CK_ULONG ec_params_len) {
+  CK_BYTE obj_ec_point[1024] = {0};
+  CK_BYTE obj_ec_param[1024] = {0};
 
   CK_ATTRIBUTE template[] = {
     {CKA_EC_POINT, obj_ec_point, sizeof(obj_ec_point)},
     {CKA_EC_PARAMS, obj_ec_param, sizeof(obj_ec_param)}
   };
 
-  test_pubkey_basic_attributes(funcs, session, pubkey, CKK_EC, key_size, label, is_neo);
+  test_pubkey_basic_attributes(funcs, session, pubkey, CKK_EC, key_size, label);
 
   asrt(funcs->C_GetAttributeValue(session, pubkey, template, 2), CKR_OK, "GET EC ATTRIBUTES");
   asrt(template[0].ulValueLen, ec_point_len, "EC POINT LEN");
@@ -1226,7 +1319,7 @@ void test_pubkey_attributes_ec(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE ses
 
 static void test_privkey_basic_attributes(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, 
                                           CK_OBJECT_HANDLE privkey, CK_ULONG key_type, CK_ULONG key_size,
-                                         const unsigned char* label, CK_BBOOL always_authenticate, CK_BBOOL is_neo) {
+                                         const unsigned char* label, CK_BBOOL always_authenticate) {
   CK_ULONG obj_class;
   CK_BBOOL obj_token;
   CK_BBOOL obj_private;
@@ -1243,7 +1336,7 @@ static void test_privkey_basic_attributes(CK_FUNCTION_LIST_PTR funcs, CK_SESSION
   CK_ULONG obj_modulus_bits;
   CK_BBOOL obj_always_authenticate;
   CK_BBOOL obj_modifiable;
-  char obj_label[1024];
+  char obj_label[1024] = {0};
   CK_ULONG obj_label_len;
 
   CK_ATTRIBUTE template[] = {
@@ -1278,11 +1371,7 @@ static void test_privkey_basic_attributes(CK_FUNCTION_LIST_PTR funcs, CK_SESSION
   asrt(obj_always_sensitive, CK_TRUE, "ALWAYS_SENSITIVE");
   asrt(obj_extractable, CK_FALSE, "EXTRACTABLE");
   asrt(obj_never_extractable, CK_TRUE, "NEVER_EXTRACTABLE");
-  if(is_neo) {
-    asrt(obj_local, CK_FALSE, "LOCAL");
-  } else {
-    asrt(obj_local, CK_TRUE, "LOCAL");
-  }
+  asrt(obj_local, CK_TRUE, "LOCAL");
   asrt(obj_decrypt, CK_TRUE, "DECRYPT");
   asrt(obj_unwrap, CK_FALSE, "UNWRAP");
   asrt(obj_sign, CK_TRUE, "SIGN");
@@ -1301,17 +1390,17 @@ void test_privkey_attributes_rsa(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE s
                                 CK_OBJECT_HANDLE pubkey, CK_ULONG key_size, 
                                 const unsigned char* label, CK_ULONG modulus_len,
                                 CK_BYTE_PTR pubexp, CK_ULONG pubexp_len, 
-                                CK_BBOOL always_authenticate, CK_BBOOL is_neo) {
+                                CK_BBOOL always_authenticate) {
 
-  CK_BYTE obj_pubexp[1024];
-  CK_BYTE obj_modulus[1024];
+  CK_BYTE obj_pubexp[1024] = {0};
+  CK_BYTE obj_modulus[1024] = {0};
 
   CK_ATTRIBUTE template[] = {
     {CKA_MODULUS, obj_modulus, sizeof(obj_modulus)},
     {CKA_PUBLIC_EXPONENT, &obj_pubexp, sizeof(obj_pubexp)},
   };
 
-  test_privkey_basic_attributes(funcs, session, pubkey, CKK_RSA, key_size, label, always_authenticate, is_neo);
+  test_privkey_basic_attributes(funcs, session, pubkey, CKK_RSA, key_size, label, always_authenticate);
 
   asrt(funcs->C_GetAttributeValue(session, pubkey, template, 2), CKR_OK, "GET RSA ATTRIBUTES");
   asrt(template[0].ulValueLen, modulus_len, "MODULUS LEN");
@@ -1323,16 +1412,16 @@ void test_privkey_attributes_ec(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE se
                                 CK_OBJECT_HANDLE pubkey, CK_ULONG key_size, 
                                 const unsigned char* label, CK_ULONG ec_point_len,
                                 CK_BYTE_PTR ec_params, CK_ULONG ec_params_len, 
-                                CK_BBOOL always_authenticate, CK_BBOOL is_neo) {
-  CK_BYTE obj_ec_point[1024];
-  CK_BYTE obj_ec_param[1024];
+                                CK_BBOOL always_authenticate) {
+  CK_BYTE obj_ec_point[1024] = {0};
+  CK_BYTE obj_ec_param[1024] = {0};
 
   CK_ATTRIBUTE template[] = {
     {CKA_EC_POINT, obj_ec_point, sizeof(obj_ec_point)},
     {CKA_EC_PARAMS, obj_ec_param, sizeof(obj_ec_param)}
   };
 
-  test_privkey_basic_attributes(funcs, session, pubkey, CKK_EC, key_size, label, always_authenticate, is_neo);
+  test_privkey_basic_attributes(funcs, session, pubkey, CKK_EC, key_size, label, always_authenticate);
 
   asrt(funcs->C_GetAttributeValue(session, pubkey, template, 2), CKR_OK, "GET EC ATTRIBUTES");
   asrt(template[0].ulValueLen, ec_point_len, "EC POINT LEN");
@@ -1343,8 +1432,8 @@ void test_privkey_attributes_ec(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE se
 void test_find_objects_by_class(CK_FUNCTION_LIST_PTR funcs, CK_SESSION_HANDLE session, 
                                 CK_ULONG class, CK_BYTE ckaid,
                                 CK_ULONG n_expected, CK_OBJECT_HANDLE obj_expected) {
-  CK_BYTE i;
-  CK_OBJECT_HANDLE obj[10];
+  CK_ULONG i;
+  CK_OBJECT_HANDLE obj[10] = {0};
   CK_ULONG n = 0;
   CK_BBOOL found = CK_FALSE;
 
