@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2015-2016 Yubico AB
+ * Copyright (c) 2015-2016,2019-2020 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -36,7 +36,7 @@
 #include "openssl_utils.h"
 
 #include <stdbool.h>
-#include "../tool/util.h"
+#include "../common/util.h"
 
 #define MIN_RSA_KEY_SIZE 1024
 #define MAX_RSA_KEY_SIZE 2048
@@ -66,12 +66,14 @@ static const token_mechanism token_mechanisms[] = {
   CKM_ECDSA_SHA224, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY},
   CKM_ECDSA_SHA256, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY},
   CKM_ECDSA_SHA384, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY},
+  CKM_ECDH1_DERIVE, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_DERIVE},
   CKM_SHA_1, {0, 0, CKF_DIGEST},
   CKM_SHA256, {0, 0, CKF_DIGEST},
   CKM_SHA384, {0, 0, CKF_DIGEST},
   CKM_SHA512, {0, 0, CKF_DIGEST}
 };
 
+// The commented out objects below are either not supported (PIV_DATA_OBJ_BITGT) or requires authentication.
 static const piv_obj_id_t token_objects[] = { // TODO: is there a way to get this from the token?
     PIV_DATA_OBJ_X509_PIV_AUTH,   // PIV authentication
     PIV_DATA_OBJ_X509_DS,         // digital signature
@@ -101,15 +103,15 @@ static const piv_obj_id_t token_objects[] = { // TODO: is there a way to get thi
     PIV_DATA_OBJ_CCC,             // Card capability container
     PIV_DATA_OBJ_CHUI,            // Cardholder unique id
     //PIV_DATA_OBJ_CHF,           // Cardholder fingerprints
-    //PIV_DATA_OBJ_SEC_OBJ,       // Security object
+    PIV_DATA_OBJ_SEC_OBJ,         // Security object
     //PIV_DATA_OBJ_CHFI,          // Cardholder facial images
     //PIV_DATA_OBJ_PI,            // Cardholder printed information
     PIV_DATA_OBJ_DISCOVERY,       // Discovery object
     PIV_DATA_OBJ_HISTORY,         // History object
     //PIV_DATA_OBJ_IRIS_IMAGE,    // Cardholder iris images
     //PIV_DATA_OBJ_BITGT,         // Biometric information templates group template
-    //PIV_DATA_OBJ_SM_SIGNER,     // Secure messaging signer
-    //PIV_DATA_OBJ_PC_REF_DATA,   // Pairing code reference data
+    PIV_DATA_OBJ_SM_SIGNER,       // Secure messaging signer
+    PIV_DATA_OBJ_PC_REF_DATA,     // Pairing code reference data
 };
 
 CK_RV get_token_model(ykpiv_state *state, CK_UTF8CHAR_PTR str, CK_ULONG len) {
@@ -138,7 +140,7 @@ CK_RV get_token_model(ykpiv_state *state, CK_UTF8CHAR_PTR str, CK_ULONG len) {
 
 CK_RV get_token_version(ykpiv_state *state, CK_VERSION_PTR version) {
 
-  char buf[16];
+  char buf[16] = {0};
 
   if (version == NULL)
     return CKR_ARGUMENTS_BAD;
@@ -158,7 +160,7 @@ CK_RV get_token_version(ykpiv_state *state, CK_VERSION_PTR version) {
 CK_RV get_token_serial(ykpiv_state *state, CK_CHAR_PTR str, CK_ULONG len) {
 
   uint32_t serial;
-  char buf[64];
+  char buf[64] = {0};
   int actual;
 
   ykpiv_rc rc = ykpiv_get_serial(state, &serial);
@@ -179,7 +181,7 @@ CK_RV get_token_serial(ykpiv_state *state, CK_CHAR_PTR str, CK_ULONG len) {
 CK_RV get_token_label(ykpiv_state *state, CK_CHAR_PTR str, CK_ULONG len) {
 
   uint32_t serial;
-  char buf[64];
+  char buf[64] = {0};
   int actual;
 
   ykpiv_rc rc = ykpiv_get_serial(state, &serial);
@@ -239,21 +241,32 @@ CK_RV token_change_pin(ykpiv_state *state, CK_USER_TYPE user_type, CK_UTF8CHAR_P
 
   switch(user_type){
     case CKU_SO:{
-      unsigned char new_key[24];
+      unsigned char new_key[24] = {0};
       size_t new_key_len = sizeof(new_key);
       if(ykpiv_hex_decode((const char*)pNewPin, ulNewLen, new_key, &new_key_len) != YKPIV_OK) {
         DBG("Failed to decode new pin")
         return CKR_PIN_INVALID;
       }
+      DBG("Changing SO PIN")
       res = ykpiv_set_mgmkey(state, new_key);
+      OPENSSL_cleanse(new_key, sizeof(new_key));
       break;
     }
     case CKU_USER:
       if(ulOldLen >= 4 && strncmp((const char*)pOldPin, "puk:", 4) == 0){
-        DBG("Changing PUK pin")
-        res = ykpiv_change_puk(state, (const char*)pOldPin + 4, ulOldLen - 4, (const char*)pNewPin, ulNewLen, &tries);
+        if(ulNewLen >= 4 && strncmp((const char*)pNewPin, "pin:", 4) == 0) {
+          DBG("Unblocking PIN with PUK")
+          res = ykpiv_unblock_pin(state, (const char*)pOldPin + 4, ulOldLen - 4, (const char*)pNewPin + 4, ulNewLen - 4, &tries);
+        } else {
+          DBG("Changing PUK")
+          if(ulNewLen >= 4 && strncmp((const char*)pNewPin, "puk:", 4) == 0) {
+            res = ykpiv_change_puk(state, (const char*)pOldPin + 4, ulOldLen - 4, (const char*)pNewPin + 4, ulNewLen - 4, &tries);
+          } else {
+            res = ykpiv_change_puk(state, (const char*)pOldPin + 4, ulOldLen - 4, (const char*)pNewPin, ulNewLen, &tries);
+          }
+        }
       }else{
-        DBG("Changing USER pin")
+        DBG("Changing PIN")
         res = ykpiv_change_pin(state, (const char*)pOldPin, ulOldLen, (const char*)pNewPin, ulNewLen, &tries);
       }
       break;
@@ -278,41 +291,73 @@ CK_RV token_change_pin(ykpiv_state *state, CK_USER_TYPE user_type, CK_UTF8CHAR_P
 
 CK_RV token_login(ykpiv_state *state, CK_USER_TYPE user, CK_UTF8CHAR_PTR pin, CK_ULONG pin_len) {
 
-  int tries = 0; // TODO: this is effectively disregarded, should we add a better value in ykpiv_verify?
-  unsigned char key[24];
-  size_t key_len = sizeof(key);
-  unsigned char *term_pin;
   ykpiv_rc res;
 
-  if (user == CKU_USER) {
-    // add null termination for the pin
-    term_pin = malloc(pin_len + 1);
-    if (term_pin == NULL) {
-      return CKR_HOST_MEMORY;
-    }
+  if (pin_len >= YKPIV_MIN_PIN_LEN && pin_len <= YKPIV_MAX_PIN_LEN) {
+    char term_pin[YKPIV_MAX_PIN_LEN + 1] = {0};
+
     memcpy(term_pin, pin, pin_len);
     term_pin[pin_len] = 0;
 
-    res = ykpiv_verify(state, (char *)term_pin, &tries);
+    int tries;
+    res = ykpiv_verify(state, term_pin, &tries);
 
     OPENSSL_cleanse(term_pin, pin_len);
-    free(term_pin);
 
     if (res != YKPIV_OK) {
-      DBG("Failed to login");
-      return CKR_PIN_INCORRECT;
+      DBG("Failed to login: %s, %d tries left", ykpiv_strerror(res), tries);
+
+      if(res == YKPIV_WRONG_PIN)
+        return CKR_PIN_INCORRECT;
+
+      if(res == YKPIV_PIN_LOCKED)
+        return CKR_PIN_LOCKED;
+
+      return CKR_DEVICE_ERROR;
     }
+  } else if(pin_len != YKPIV_MGM_KEY_LEN || user != CKU_SO) {
+    DBG("PIN is wrong length");
+    return CKR_ARGUMENTS_BAD;
   }
-  else if (user == CKU_SO) {
-    if(ykpiv_hex_decode((char *)pin, pin_len, key, &key_len) != YKPIV_OK) {
-      DBG("Failed decoding key");
-      return CKR_PIN_INVALID;
+
+  if (user == CKU_SO) {
+    unsigned char key[24] = {0};
+
+    if (pin_len == YKPIV_MGM_KEY_LEN) {
+      size_t key_len = sizeof(key);
+      if(ykpiv_hex_decode((char *)pin, pin_len, key, &key_len) != YKPIV_OK) {
+        DBG("Failed decoding key");
+        OPENSSL_cleanse(key, key_len);
+        return CKR_ARGUMENTS_BAD;
+      }
+    } else {
+      ykpiv_config cfg;
+      res = ykpiv_util_get_config(state, &cfg);
+      if(res != YKPIV_OK) {
+        DBG("Failed to get device configuration: %s", ykpiv_strerror(res));
+        return CKR_DEVICE_ERROR;
+      }
+
+      if(cfg.mgm_type != YKPIV_CONFIG_MGM_PROTECTED) {
+        DBG("Device configuration invalid, no PIN-protected MGM key available");
+        return CKR_USER_PIN_NOT_INITIALIZED;
+      }
+
+      memcpy(key, cfg.mgm_key, sizeof(key));
+      OPENSSL_cleanse(cfg.mgm_key, sizeof(cfg.mgm_key));
     }
 
-    if(ykpiv_authenticate(state, key) != YKPIV_OK) {
-      DBG("Failed to authenticate");
-      return CKR_PIN_INCORRECT;
+    if((res = ykpiv_authenticate(state, key)) != YKPIV_OK) {
+      DBG("Failed to authenticate: %s", ykpiv_strerror(res));
+      OPENSSL_cleanse(key, sizeof(key));
+
+      if(res == YKPIV_AUTHENTICATION_ERROR)
+        return CKR_PIN_INCORRECT;
+
+      return CKR_DEVICE_ERROR;
     }
+
+    OPENSSL_cleanse(key, sizeof(key));
   }
 
   return CKR_OK;
@@ -320,15 +365,14 @@ CK_RV token_login(ykpiv_state *state, CK_USER_TYPE user, CK_UTF8CHAR_PTR pin, CK
 
 CK_RV token_generate_key(ykpiv_state *state, CK_BYTE algorithm, CK_BYTE key, CK_BYTE_PTR cert_data, CK_ULONG_PTR cert_len) {
   // TODO: make a function in ykpiv for this
-  unsigned char in_data[11];
+  unsigned char in_data[11] = {0};
   unsigned char *in_ptr = in_data;
-  unsigned char data[1024];
+  unsigned char data[1024] = {0};
   unsigned char templ[] = {0, YKPIV_INS_GENERATE_ASYMMETRIC, 0, 0};
   unsigned char *certptr;
-  unsigned long len, offset, recv_len = sizeof(data);
-  char version[7];
-  char label[32];
-  int len_bytes;
+  unsigned long len, len_bytes, offs, recv_len = sizeof(data);
+  char version[7] = {0};
+  char label[32] = {0};
   int sw;
 
   switch(algorithm) {
@@ -346,12 +390,10 @@ CK_RV token_generate_key(ykpiv_state *state, CK_BYTE algorithm, CK_BYTE key, CK_
         DBG("Failed to communicate.");
         return CKR_DEVICE_ERROR;
       }
-      offset = 5;
       break;
 
     case YKPIV_ALGO_ECCP256:
     case YKPIV_ALGO_ECCP384:
-      offset = 3;
       break;
 
     default:
@@ -374,18 +416,17 @@ CK_RV token_generate_key(ykpiv_state *state, CK_BYTE algorithm, CK_BYTE key, CK_
   snprintf(label, sizeof(label), "YubiKey PIV Slot %x", key);
 
   // Create a new empty certificate for the key
+  offs = 2 + get_length(data + 2, data + recv_len, &len);
+  if(offs == 2)
+    return CKR_DEVICE_ERROR;
+
   len = recv_len;
   recv_len = sizeof(data);
-  CK_RV rv = do_create_empty_cert(data + offset, len - offset, algorithm, label, data, &recv_len);
+  CK_RV rv = do_create_empty_cert(data + offs, len - offs, algorithm, label, data, &recv_len);
   if(rv != CKR_OK)
     return rv;
 
-  if (recv_len < 0x80)
-    len_bytes = 1;
-  else if (recv_len < 0xff)
-    len_bytes = 2;
-  else
-    len_bytes = 3;
+  len_bytes = get_length_size(recv_len);
 
   certptr = data;
   memmove(data + len_bytes + 1, data, recv_len);
@@ -416,7 +457,7 @@ CK_RV token_generate_key(ykpiv_state *state, CK_BYTE algorithm, CK_BYTE key, CK_
 
 CK_RV token_import_cert(ykpiv_state *state, CK_ULONG cert_id, CK_BYTE_PTR in, CK_ULONG in_len) {
 
-  unsigned char certdata[YKPIV_OBJ_MAX_SIZE + 16];
+  unsigned char certdata[YKPIV_OBJ_MAX_SIZE + 16] = {0};
   unsigned char *certptr;
   CK_ULONG cert_len;
 
