@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2016 Yubico AB
+ * Copyright (c) 2014-2017,2019-2020 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -29,9 +29,9 @@
  */
 
 #include "ykcs11.h"
-#include "ykcs11-version.h"
+#include "ykcs11-config.h"
 #include <stdlib.h>
-#include <ykpiv.h>
+#include "ykpiv.h"
 #include <string.h>
 #include "obj_types.h"
 #include "objects.h"
@@ -47,10 +47,6 @@
 #define YKCS11_MANUFACTURER "Yubico (www.yubico.com)"
 #define YKCS11_LIBDESC      "PKCS#11 PIV Library (SP-800-73)"
 
-#define PIV_MIN_PIN_LEN 6
-#define PIV_MAX_PIN_LEN 8
-#define PIV_MGM_KEY_LEN 48
-
 #define YKCS11_MAX_SLOTS       16
 #define YKCS11_MAX_SESSIONS    16
 
@@ -62,6 +58,7 @@ static ykcs11_session_t sessions[YKCS11_MAX_SESSIONS];
 static CK_C_INITIALIZE_ARGS locking;
 static void *global_mutex;
 static uint64_t pid;
+int verbose;
 
 static CK_FUNCTION_LIST function_list;
 
@@ -111,6 +108,13 @@ CK_DEFINE_FUNCTION(CK_RV, C_Initialize)(
   CK_VOID_PTR pInitArgs
 )
 {
+#if YKCS11_DBG
+  verbose = YKCS11_DBG;
+#else
+  const char *dbg = getenv("YKCS11_DBG");
+  verbose = dbg ? atoi(dbg) : 0;
+#endif
+
   DIN;
   CK_RV rv;
 
@@ -256,6 +260,12 @@ CK_DEFINE_FUNCTION(CK_RV, C_GetInfo)(
   CK_RV rv;
   DIN;
 
+  if (!pid) {
+    DBG("libykpiv is not initialized or already finalized");
+    rv = CKR_CRYPTOKI_NOT_INITIALIZED;
+    goto info_out;
+  }
+
   if (pInfo == NULL) {
     DBG("Wrong/Missing parameter");
     rv = CKR_ARGUMENTS_BAD;
@@ -306,7 +316,7 @@ CK_DEFINE_FUNCTION(CK_RV, C_GetSlotList)(
 )
 {
   DIN;
-  char readers[2048];
+  char readers[2048] = {0};
   size_t len = sizeof(readers);
   CK_RV rv;
 
@@ -323,7 +333,7 @@ CK_DEFINE_FUNCTION(CK_RV, C_GetSlotList)(
   }
 
   ykpiv_state *piv_state;
-  if (ykpiv_init(&piv_state, YKCS11_DBG) != YKPIV_OK) {
+  if (ykpiv_init(&piv_state, verbose) != YKPIV_OK) {
     DBG("Unable to initialize libykpiv");
     rv = CKR_FUNCTION_FAILED;
     goto slotlist_out;
@@ -376,7 +386,7 @@ CK_DEFINE_FUNCTION(CK_RV, C_GetSlotList)(
       if(slot == slots + n_slots) {
         DBG("Initializing slot %lu for '%s'", slot-slots, reader);
         ykpiv_rc rc;
-        if((rc = ykpiv_init(&slot->piv_state, YKCS11_DBG)) != YKPIV_OK) {
+        if((rc = ykpiv_init(&slot->piv_state, verbose)) != YKPIV_OK) {
           DBG("Unable to initialize libykpiv: %s", ykpiv_strerror(rc));
           locking.pfnUnlockMutex(global_mutex);
           rv = CKR_FUNCTION_FAILED;
@@ -385,38 +395,44 @@ CK_DEFINE_FUNCTION(CK_RV, C_GetSlotList)(
         n_slots++;
       }
 
-      char buf[sizeof(readers) + 1];
-      snprintf(buf, sizeof(buf), "@%s", reader);
-
       // Try to connect if unconnected (both new and existing slots)
-      if (!(slot->slot_info.flags & CKF_TOKEN_PRESENT) && ykpiv_connect(slot->piv_state, buf) == YKPIV_OK) {
+      if (ykpiv_validate(slot->piv_state, reader) != YKPIV_OK) {
 
-        DBG("Connected slot %lu to '%s'", slot-slots, reader);
+        slot->login_state = YKCS11_PUBLIC;
+        slot->slot_info.flags &= ~CKF_TOKEN_PRESENT;
 
-        slot->slot_info.flags |= CKF_TOKEN_PRESENT;
-        slot->token_info.flags = CKF_RNG | CKF_LOGIN_REQUIRED | CKF_USER_PIN_INITIALIZED | CKF_TOKEN_INITIALIZED;
+        char buf[sizeof(readers) + 1] = {0};
+        snprintf(buf, sizeof(buf), "@%s", reader);
 
-        slot->token_info.ulMinPinLen = PIV_MIN_PIN_LEN;
-        slot->token_info.ulMaxPinLen = PIV_MGM_KEY_LEN;
+        if (ykpiv_connect(slot->piv_state, buf) == YKPIV_OK) {
 
-        slot->token_info.ulMaxRwSessionCount = YKCS11_MAX_SESSIONS;
-        slot->token_info.ulMaxSessionCount = YKCS11_MAX_SESSIONS;
+          DBG("Connected slot %lu to '%s'", slot-slots, reader);
 
-        slot->token_info.ulTotalPublicMemory = CK_UNAVAILABLE_INFORMATION;
-        slot->token_info.ulFreePublicMemory = CK_UNAVAILABLE_INFORMATION;
-        slot->token_info.ulTotalPrivateMemory = CK_UNAVAILABLE_INFORMATION;
-        slot->token_info.ulFreePrivateMemory = CK_UNAVAILABLE_INFORMATION;
+          slot->slot_info.flags |= CKF_TOKEN_PRESENT;
+          slot->token_info.flags = CKF_RNG | CKF_LOGIN_REQUIRED | CKF_USER_PIN_INITIALIZED | CKF_TOKEN_INITIALIZED;
 
-        slot->token_info.hardwareVersion.major = 1;
-        slot->token_info.hardwareVersion.minor = 0;
+          slot->token_info.ulMinPinLen = YKPIV_MIN_PIN_LEN;
+          slot->token_info.ulMaxPinLen = YKPIV_MGM_KEY_LEN;
 
-        memstrcpy(slot->token_info.manufacturerID, sizeof(slot->token_info.manufacturerID), YKCS11_MANUFACTURER);
-        memset(slot->token_info.utcTime, ' ', sizeof(slot->token_info.utcTime));
+          slot->token_info.ulMaxRwSessionCount = YKCS11_MAX_SESSIONS;
+          slot->token_info.ulMaxSessionCount = YKCS11_MAX_SESSIONS;
 
-        get_token_model(slot->piv_state, slot->token_info.model, sizeof(slot->token_info.model));
-        get_token_serial(slot->piv_state, slot->token_info.serialNumber, sizeof(slot->token_info.serialNumber));
-        get_token_version(slot->piv_state, &slot->token_info.firmwareVersion);
-        get_token_label(slot->piv_state, slot->token_info.label, sizeof(slot->token_info.label));
+          slot->token_info.ulTotalPublicMemory = CK_UNAVAILABLE_INFORMATION;
+          slot->token_info.ulFreePublicMemory = CK_UNAVAILABLE_INFORMATION;
+          slot->token_info.ulTotalPrivateMemory = CK_UNAVAILABLE_INFORMATION;
+          slot->token_info.ulFreePrivateMemory = CK_UNAVAILABLE_INFORMATION;
+
+          slot->token_info.hardwareVersion.major = 1;
+          slot->token_info.hardwareVersion.minor = 0;
+
+          memstrcpy(slot->token_info.manufacturerID, sizeof(slot->token_info.manufacturerID), YKCS11_MANUFACTURER);
+          memset(slot->token_info.utcTime, ' ', sizeof(slot->token_info.utcTime));
+
+          get_token_model(slot->piv_state, slot->token_info.model, sizeof(slot->token_info.model));
+          get_token_serial(slot->piv_state, slot->token_info.serialNumber, sizeof(slot->token_info.serialNumber));
+          get_token_version(slot->piv_state, &slot->token_info.firmwareVersion);
+          get_token_label(slot->piv_state, slot->token_info.label, sizeof(slot->token_info.label));
+        }
       }
     }
   }
@@ -537,6 +553,23 @@ CK_DEFINE_FUNCTION(CK_RV, C_GetTokenInfo)(
 
   memcpy(pInfo, &slots[slotID].token_info, sizeof(CK_TOKEN_INFO));
 
+  int tries = YKPIV_RETRIES_MAX;
+  ykpiv_get_pin_retries(slots[slotID].piv_state, &tries);
+
+  switch(tries) {
+    case 0:
+      pInfo->flags |= CKF_USER_PIN_LOCKED;
+      break;
+    case 1:
+      pInfo->flags |= CKF_USER_PIN_FINAL_TRY;
+      break;
+    case 2:
+      pInfo->flags |= CKF_USER_PIN_COUNT_LOW;
+      break;
+    default:
+      break;
+  }
+
   for(int i = 0; i < YKCS11_MAX_SESSIONS; i++) {
     if(sessions[i].slot) {
       if(sessions[i].info.flags & CKF_RW_SESSION) {
@@ -713,7 +746,7 @@ CK_DEFINE_FUNCTION(CK_RV, C_InitToken)(
 
   locking.pfnUnlockMutex(global_mutex);
 
-  CK_BYTE mgm_key[24];
+  CK_BYTE mgm_key[24] = {0};
   size_t len = sizeof(mgm_key);
   ykpiv_rc rc;
 
@@ -738,7 +771,7 @@ CK_DEFINE_FUNCTION(CK_RV, C_InitToken)(
   if((rc = ykpiv_authenticate(slot->piv_state, mgm_key)) != YKPIV_OK) {
     DBG("ykpiv_authenticate failed %s", ykpiv_strerror(rc));
     locking.pfnUnlockMutex(slot->mutex);
-    rv = CKR_PIN_INCORRECT;
+    rv = rc == YKPIV_AUTHENTICATION_ERROR ? CKR_PIN_INCORRECT : CKR_DEVICE_ERROR;
     goto inittoken_out;
   }
 
@@ -764,7 +797,7 @@ CK_DEFINE_FUNCTION(CK_RV, C_InitToken)(
   if((rc = ykpiv_authenticate(slot->piv_state, NULL)) != YKPIV_OK) {
     DBG("ykpiv_authenticate failed %s", ykpiv_strerror(rc));
     locking.pfnUnlockMutex(slot->mutex);
-    rv = CKR_DEVICE_ERROR;
+    rv = rc == YKPIV_AUTHENTICATION_ERROR ? CKR_PIN_INCORRECT : CKR_DEVICE_ERROR;
     goto inittoken_out;
   }
 
@@ -821,6 +854,12 @@ CK_DEFINE_FUNCTION(CK_RV, C_SetPIN)(
     goto setpin_out;
   }
 
+  if((session->info.flags & CKF_RW_SESSION) == 0) {
+    DBG("User called SetPIN on read-only session");
+    rv = CKR_SESSION_READ_ONLY;
+    goto setpin_out;
+  }
+
   locking.pfnLockMutex(session->slot->mutex);
 
   CK_USER_TYPE user_type = session->slot->login_state == YKCS11_SO ? CKU_SO : CKU_USER;
@@ -905,56 +944,61 @@ CK_DEFINE_FUNCTION(CK_RV, C_OpenSession)(
     CK_ULONG num_ids;
     get_token_object_ids(&obj_ids, &num_ids);
     for(CK_ULONG i = 0; i < num_ids; i++) {
-      ykpiv_rc rcc, rc = YKPIV_KEY_ERROR;
+      ykpiv_rc rc = YKPIV_KEY_ERROR;
       CK_BYTE sub_id = get_sub_id(obj_ids[i]);
       piv_obj_id_t cert_id = find_cert_object(sub_id);
       piv_obj_id_t pubk_id = find_pubk_object(sub_id);
       piv_obj_id_t pvtk_id = find_pvtk_object(sub_id);
       piv_obj_id_t atst_id = find_atst_object(sub_id);
-      CK_ULONG slot = piv_2_ykpiv(pvtk_id);
-      CK_BYTE data[YKPIV_OBJ_MAX_SIZE];  // Max cert value for ykpiv
+      CK_BYTE data[YKPIV_OBJ_MAX_SIZE] = {0};  // Max cert value for ykpiv
       size_t len;
       if(pvtk_id != PIV_INVALID_OBJ) {
+        CK_ULONG slot = piv_2_ykpiv(pvtk_id);
         len = sizeof(data);
-        if((rc = ykpiv_get_metadata(session->slot->piv_state, slot, data, &len)) == YKPIV_OK) {
-          DBG("Read %lu bytes metadata for slot %lx", len, slot);
-          ykpiv_metadata md = {0};
-          if((rc = ykpiv_util_parse_metadata(data, len, &md)) == YKPIV_OK) {
-            if((rv = do_create_public_key(md.pubkey, md.pubkey_len, md.algorithm, &session->slot->pkeys[sub_id])) == CKR_OK) {
-              add_object(session->slot, pubk_id);
+        if((rc = ykpiv_attest(session->slot->piv_state, slot, data, &len)) == YKPIV_OK) {
+          DBG("Created attestation for object %u slot %lx", pvtk_id, slot);
+          if((rv = do_store_cert(data, len, &session->slot->atst[sub_id])) == CKR_OK) {
+            if(atst_id != PIV_INVALID_OBJ)
+              add_object(session->slot, atst_id);
+            if((rv = do_store_pubk(session->slot->atst[sub_id], &session->slot->pkeys[sub_id])) == CKR_OK) {
+              session->slot->local[sub_id] = CK_TRUE;
               add_object(session->slot, pvtk_id);
-              if(atst_id != PIV_INVALID_OBJ && md.origin == YKPIV_METADATA_ORIGIN_GENERATED) { // Attestation key doesn't have an attestation
-                len = sizeof(data);
-                rcc = ykpiv_attest(session->slot->piv_state, slot, data, &len);
-                if(rcc == YKPIV_OK) {
-                  DBG("Created attestation for slot %lx", slot);
-                  if((rv = do_store_cert(data, len, session->slot->atst + sub_id)) == CKR_OK) {
-                    add_object(session->slot, atst_id);
-                  } else {
-                    DBG("Failed to store certificate object %u in session: %lu", atst_id, rv);
-                  }
-                } else {
-                  DBG("Failed to create attestation for slot %lx: %s", slot, ykpiv_strerror(rcc));
-                }
-              }
+              add_object(session->slot, pubk_id);
             } else {
-              DBG("Failed to create public key info for slot %lx, algorithm %u from metadata: %lu", slot, md.algorithm, rv);
-              rc = YKPIV_KEY_ERROR; // Ensure we create the key from the certificate instead
+              DBG("Failed to store key objects %u and %u in session: %lu", pubk_id, pvtk_id, rv);
             }
           } else {
-            DBG("Failed to parse metadata for slot %lx: %s", slot, ykpiv_strerror(rc));
+            DBG("Failed to store attestation certificate object %u in session: %lu", atst_id, rv);
           }
         } else {
-          DBG("Failed to read metadata for slot %lx: %s", slot, ykpiv_strerror(rc));
+          DBG("Failed to create attestation for object %u slot %lx: %s", pvtk_id, slot, ykpiv_strerror(rc));
+          len = sizeof(data);
+          if((rc = ykpiv_get_metadata(session->slot->piv_state, slot, data, &len)) == YKPIV_OK) {
+            DBG("Fetched %lu bytes metadata for object %u slot %lx", len, pvtk_id, slot);
+            ykpiv_metadata md = {0};
+            if((rc = ykpiv_util_parse_metadata(data, len, &md)) == YKPIV_OK) {
+              if((rv = do_create_public_key(md.pubkey, md.pubkey_len, md.algorithm, &session->slot->pkeys[sub_id])) == CKR_OK) {
+                session->slot->local[sub_id] = md.origin == YKPIV_METADATA_ORIGIN_GENERATED ? CK_TRUE : CK_FALSE;
+                add_object(session->slot, pvtk_id);
+                add_object(session->slot, pubk_id);
+              } else {
+                DBG("Failed to create public key for slot %lx, algorithm %u from metadata: %lu", slot, md.algorithm, rv);
+              }
+            } else {
+              DBG("Failed to parse metadata for object %u slot %lx: %s", pvtk_id, slot, ykpiv_strerror(rc));
+            }
+          } else {
+            DBG("Failed to fetch metadata for object %u slot %lx: %s", pvtk_id, slot, ykpiv_strerror(rc));
+          }
         }
       }
       unsigned long ulen = sizeof(data);
-      rcc = ykpiv_fetch_object(session->slot->piv_state, piv_2_ykpiv(obj_ids[i]), data, &ulen);
+      ykpiv_rc rcc = ykpiv_fetch_object(session->slot->piv_state, piv_2_ykpiv(obj_ids[i]), data, &ulen);
       if(rcc != YKPIV_OK) {
-        DBG("Failed to read object %u slot %lx: %s", obj_ids[i], piv_2_ykpiv(obj_ids[i]), ykpiv_strerror(rcc));
+        DBG("Failed to fetch object %u slot %lx: %s", obj_ids[i], piv_2_ykpiv(obj_ids[i]), ykpiv_strerror(rcc));
         continue;
       }
-      DBG("Read %lu bytes for object %u slot %lx", ulen, obj_ids[i], piv_2_ykpiv(obj_ids[i]));
+      DBG("Fetched %lu bytes for object %u slot %lx", ulen, obj_ids[i], piv_2_ykpiv(obj_ids[i]));
       rv = store_data(session->slot, sub_id, data, ulen);
       if (rv != CKR_OK) {
         DBG("Failed to store data object %u in session: %lu", obj_ids[i], rv);
@@ -962,29 +1006,15 @@ CK_DEFINE_FUNCTION(CK_RV, C_OpenSession)(
       }
       add_object(session->slot, obj_ids[i]);
       if(cert_id != PIV_INVALID_OBJ) {
-        rv = store_cert(session->slot, sub_id, data, ulen, CK_FALSE);
+        rv = store_cert(session->slot, sub_id, data, ulen, CK_FALSE); // Will only overwrite key if not set from attestation or metadata
         if (rv != CKR_OK) {
           DBG("Failed to store certificate object %u in session: %lu", cert_id, rv);
           continue; // Bail out, can't create key objects without the public key from the cert
         }
         add_object(session->slot, cert_id);
-        if(rc != YKPIV_OK) { // Failed to get metadata, fall back to assuming we have keys for cert objects
-          add_object(session->slot, pubk_id);
+        if(rc != YKPIV_OK) { // Failed to get attestation or metadata, fall back to assuming we have keys for cert objects
           add_object(session->slot, pvtk_id);
-          if(atst_id != PIV_INVALID_OBJ) { // Attestation key doesn't have an attestation
-            len = sizeof(data);
-            rcc = ykpiv_attest(session->slot->piv_state, slot, data, &len);
-            if(rcc == YKPIV_OK) {
-              DBG("Created attestation for slot %lx", slot);
-              if((rv = do_store_cert(data, len, session->slot->atst + sub_id)) == CKR_OK) {
-                add_object(session->slot, atst_id);
-              } else {
-                DBG("Failed to store certificate object %u in session: %lu", atst_id, rv);
-              }
-            } else {
-              DBG("Failed to create attestation for slot %lx: %s", slot, ykpiv_strerror(rcc));
-            }
-          }
+          add_object(session->slot, pubk_id);
         }
       }
     }
@@ -1198,7 +1228,7 @@ CK_DEFINE_FUNCTION(CK_RV, C_Login)(
     goto login_out;
   }
 
-  DBG("user %lu, pin %s, pinlen %lu", userType, pPin, ulPinLen);
+  DBG("userType %lu, pinLen %lu", userType, ulPinLen);
 
   ykcs11_session_t* session = get_session(hSession);
 
@@ -1217,11 +1247,6 @@ CK_DEFINE_FUNCTION(CK_RV, C_Login)(
     }
     // Fall through
   case CKU_USER:
-    if (ulPinLen < PIV_MIN_PIN_LEN || ulPinLen > PIV_MAX_PIN_LEN) {
-      rv = CKR_ARGUMENTS_BAD;
-      goto login_out;
-    }
-
     locking.pfnLockMutex(session->slot->mutex);
 
     // We allow multiple logins for CKU_CONTEXT_SPECIFIC (we allow it regardless of CKA_ALWAYS_AUTHENTICATE because it's based on hardcoded tables and might be wrong)
@@ -1254,11 +1279,6 @@ CK_DEFINE_FUNCTION(CK_RV, C_Login)(
     break;
 
   case CKU_SO:
-    if (ulPinLen != PIV_MGM_KEY_LEN) {
-      rv = CKR_ARGUMENTS_BAD;
-      goto login_out;
-    }
-
     locking.pfnLockMutex(session->slot->mutex);
 
     if (session->slot->login_state == YKCS11_USER) {
@@ -1454,16 +1474,14 @@ CK_DEFINE_FUNCTION(CK_RV, C_CreateObject)(
       goto create_out;
     }
 
+    session->slot->local[id] = CK_FALSE;
+
     // Add objects that were not already present
 
-    if(!is_present(session->slot, dobj_id))
-      add_object(session->slot, dobj_id);
-    if(!is_present(session->slot, cert_id))
-      add_object(session->slot, cert_id);
-    if(!is_present(session->slot, pvtk_id))
-      add_object(session->slot, pvtk_id);
-    if(!is_present(session->slot, pubk_id))
-      add_object(session->slot, pubk_id);
+    add_object(session->slot, dobj_id);
+    add_object(session->slot, cert_id);
+    add_object(session->slot, pvtk_id);
+    add_object(session->slot, pubk_id);
 
     // No attestation can be created for imported objects
 
@@ -1539,6 +1557,14 @@ CK_DEFINE_FUNCTION(CK_RV, C_CreateObject)(
       }
     }
 
+    session->slot->local[id] = CK_FALSE;
+
+    add_object(session->slot, pvtk_id);
+
+    // No attestation can be created for imported objects
+
+    sort_objects(session->slot);
+
     locking.pfnUnlockMutex(session->slot->mutex);
     *phObject = (CK_OBJECT_HANDLE)pvtk_id;
     break;
@@ -1594,7 +1620,7 @@ CK_DEFINE_FUNCTION(CK_RV, C_DestroyObject)(
 
   // Silently ignore valid but not-present handles for compatibility with applications
   CK_BYTE id = get_sub_id(hObject);
-  if(id == 0) {
+  if(id == 0 && hObject != PIV_SECRET_OBJ) {
     DBG("Object handle is invalid");
     rv = CKR_OBJECT_HANDLE_INVALID;
     goto destroy_out;
@@ -1602,26 +1628,28 @@ CK_DEFINE_FUNCTION(CK_RV, C_DestroyObject)(
 
   locking.pfnLockMutex(session->slot->mutex);
 
-  // SO must be logged in
-  if (session->slot->login_state != YKCS11_SO) {
-    DBG("Authentication as SO required to delete objects");
-    locking.pfnUnlockMutex(session->slot->mutex);
-    rv = CKR_USER_TYPE_INVALID;
-    goto destroy_out;
-  }
+  if(id) {
+    // SO must be logged in
+    if (session->slot->login_state != YKCS11_SO) {
+      DBG("Authentication as SO required to delete objects");
+      locking.pfnUnlockMutex(session->slot->mutex);
+      rv = CKR_USER_TYPE_INVALID;
+      goto destroy_out;
+    }
 
-  DBG("Deleting object %lu", hObject);
-
-  rv = token_delete_cert(session->slot->piv_state, piv_2_ykpiv(find_data_object(id)));
-  if (rv != CKR_OK) {
-    DBG("Unable to delete object %lx from token", piv_2_ykpiv(find_data_object(id)));
-    locking.pfnUnlockMutex(session->slot->mutex);
-    goto destroy_out;
+    DBG("Deleting object %lx from token", piv_2_ykpiv(find_data_object(id)));
+ 
+    rv = token_delete_cert(session->slot->piv_state, piv_2_ykpiv(find_data_object(id)));
+    if (rv != CKR_OK) {
+      DBG("Unable to delete object %lx from token", piv_2_ykpiv(find_data_object(id)));
+      locking.pfnUnlockMutex(session->slot->mutex);
+      goto destroy_out;
+    }
   }
 
   // Remove the related objects from the session
 
-  DBG("%lu session objects before destroying object %lu", session->slot->n_objects, hObject);
+  DBG("%lu slot objects before destroying object %lu", session->slot->n_objects, hObject);
 
   CK_ULONG j = 0;
   for (CK_ULONG i = 0; i < session->slot->n_objects; i++) {
@@ -1630,18 +1658,18 @@ CK_DEFINE_FUNCTION(CK_RV, C_DestroyObject)(
   }
   session->slot->n_objects = j;
 
-  DBG("%lu session objects after destroying object %lu", session->slot->n_objects, hObject);
+  DBG("%lu slot objects after destroying object %lu", session->slot->n_objects, hObject);
 
   rv = delete_data(session->slot, id);
   if (rv != CKR_OK) {
-    DBG("Unable to delete data from session");
+    DBG("Unable to delete data from slot");
     locking.pfnUnlockMutex(session->slot->mutex);
     goto destroy_out;
   }
 
   rv = delete_cert(session->slot, id);
   if (rv != CKR_OK) {
-    DBG("Unable to delete certificate from session");
+    DBG("Unable to delete certificate from slot");
     locking.pfnUnlockMutex(session->slot->mutex);
     goto destroy_out;
   }
@@ -2315,7 +2343,7 @@ CK_DEFINE_FUNCTION(CK_RV, C_Decrypt)(
   if (pData == NULL) {
     // Just return the size of the decrypted data
     *pulDataLen = datalen;
-    DBG("The size of the signature will be %lu", *pulDataLen);
+    DBG("The size of the data will be %lu", *pulDataLen);
     DOUT;
     return CKR_OK;
   }
@@ -3380,7 +3408,7 @@ CK_DEFINE_FUNCTION(CK_RV, C_GenerateKeyPair)(
   piv_obj_id_t   pvtk_id;
   piv_obj_id_t   pubk_id;
   piv_obj_id_t   atst_id;
-  CK_BYTE        cert_data[YKPIV_OBJ_MAX_SIZE];
+  CK_BYTE        cert_data[YKPIV_OBJ_MAX_SIZE] = {0};
   CK_ULONG       cert_len;
   
   if (!pid) {
@@ -3489,29 +3517,26 @@ CK_DEFINE_FUNCTION(CK_RV, C_GenerateKeyPair)(
     goto genkp_out;
   }
 
+  session->slot->local[gen.key_id] = CK_TRUE;
+
   // Add objects that were not already present
 
-  if(!is_present(session->slot, dobj_id))
-    add_object(session->slot, dobj_id);
-  if(!is_present(session->slot, cert_id))
-    add_object(session->slot, cert_id);
-  if(!is_present(session->slot, pvtk_id))
-    add_object(session->slot, pvtk_id);
-  if(!is_present(session->slot, pubk_id))
-    add_object(session->slot, pubk_id);
+  add_object(session->slot, dobj_id);
+  add_object(session->slot, cert_id);
+  add_object(session->slot, pvtk_id);
+  add_object(session->slot, pubk_id);
 
   // Create an attestation, if appropriate and able
 
   if(atst_id != PIV_INVALID_OBJ) {
-    unsigned char data[YKPIV_OBJ_MAX_SIZE];
+    unsigned char data[YKPIV_OBJ_MAX_SIZE] = {0};
     size_t len = sizeof(data);
     ykpiv_rc rc = ykpiv_attest(session->slot->piv_state, slot, data, &len);
     if(rc == YKPIV_OK) {
       DBG("Created attestation for slot %lx", slot);
       if((rv = do_store_cert(data, len, session->slot->atst + gen.key_id)) == CKR_OK) {
         // Add attestation object if not already present
-        if(!is_present(session->slot, atst_id))
-          add_object(session->slot, atst_id);
+        add_object(session->slot, atst_id);
       } else {
         DBG("Failed to store attestation certificate %u in session: %lu", atst_id, rv);
       }
@@ -3575,9 +3600,85 @@ CK_DEFINE_FUNCTION(CK_RV, C_DeriveKey)(
 )
 {
   DIN;
-  DBG("TODO!!!");
+  ykcs11_session_t* session = get_session(hSession);
+
+  if (session == NULL || session->slot == NULL) {
+    DBG("Session is not open");
+    return CKR_SESSION_HANDLE_INVALID;
+  }
+
+  if (hBaseKey < PIV_PVTK_OBJ_PIV_AUTH || hBaseKey > PIV_PVTK_OBJ_ATTESTATION) {
+    DBG("Key handle %lu is not a private key", hBaseKey);
+    return CKR_KEY_HANDLE_INVALID;
+  }
+
+  CK_BYTE id = get_sub_id(hBaseKey);
+  CK_BYTE algo = do_get_key_algorithm(session->slot->pkeys[id]);
+  CK_ULONG size;
+
+  switch(algo) {
+    case YKPIV_ALGO_ECCP256:
+      size = 65;
+      break;
+    case YKPIV_ALGO_ECCP384:
+      size = 97;
+      break;
+    default:
+      DBG("Key handle %lu is not an ECDH private key", hBaseKey);
+      return CKR_KEY_TYPE_INCONSISTENT;
+  }
+
+  if (pMechanism->mechanism != CKM_ECDH1_DERIVE) {
+    DBG("Mechanism invalid");
+    return CKR_MECHANISM_INVALID;
+  }
+
+  if (pMechanism->pParameter == NULL || pMechanism->ulParameterLen != sizeof(CK_ECDH1_DERIVE_PARAMS)) {
+    DBG("Mechanism parameters invalid");
+    return CKR_MECHANISM_PARAM_INVALID;
+  }
+
+  CK_ECDH1_DERIVE_PARAMS *params = pMechanism->pParameter;
+
+  if (params->kdf != CKD_NULL || params->ulSharedDataLen != 0 || params->pPublicData == NULL || params->ulPublicDataLen != size) {
+    DBG("Key derivation parameters invalid");
+    return CKR_MECHANISM_PARAM_INVALID;
+  }
+
+  for(CK_ULONG i = 0; i < ulAttributeCount; i++) {
+    CK_RV rv = validate_derive_key_attribute(pTemplate[i].type, pTemplate[i].pValue);
+    if(rv != CKR_OK) {
+      DOUT;
+      return rv;
+    }
+  }
+
+  CK_ULONG slot = piv_2_ykpiv(hBaseKey);
+  unsigned char buf[128];
+  size_t len = sizeof(buf);
+
+  locking.pfnLockMutex(session->slot->mutex);
+
+  DBG("Deriving ECDH shared secret into object %u using slot %lx", PIV_SECRET_OBJ, slot);
+  ykpiv_rc rc = ykpiv_decipher_data(session->slot->piv_state, params->pPublicData, params->ulPublicDataLen, &buf, &len, algo, slot);
+
+  if(rc != YKPIV_OK) {
+    DBG("Failed to derive key in slot %lx: %s", slot, ykpiv_strerror(rc));
+    locking.pfnUnlockMutex(session->slot->mutex);
+    DOUT;
+    return CKR_FUNCTION_FAILED;
+  }
+
+  *phKey = PIV_SECRET_OBJ;
+
+  store_data(session->slot, 0, buf, len);
+  add_object(session->slot, *phKey);
+  sort_objects(session->slot);
+
+  locking.pfnUnlockMutex(session->slot->mutex);
+  
   DOUT;
-  return CKR_FUNCTION_NOT_SUPPORTED;
+  return CKR_OK;
 }
 
 /* Random number generation functions */
